# Controls AnimeContinuum pages.
class AnimeContinuumController < ApplicationController
  before_filter :login_required, :except => [:index, :list, :show, :new,
                                             :create, :search_select_merge_anime_continuums]
  after_filter :store_location, :only => [:index, :list, :show, :new, :edit,
                                          :manage, :select_merge]
  
  require_role 'anime_continuum_approver', :for_all_except => [:index, :list,
                                                               :show, :new,
                                                               :create, :search_select_merge_anime_continuums]
  require_role 'anime_continuum_moderator', :for => [:destroy, :select_merge,
                                                     :merge_into]
  
  in_place_edit_with_validation_for :anime_continuum, :name
  
  # Renders a list of AnimeContinuums.
  def index
    list
    render :action => :list
  end

  # Render a list of AnimeContinuums.
  def list
    @anime_continuums = AnimeContinuum.paginate :all, :page => params[:page], :order => 'name ASC'
    @title = 'Genericon Trivia System - Anime Continuums List'
  end

  # Render the details of an AnimeContinuum for normal users.
  def show
    @anime_continuum = AnimeContinuum.find params[:id]
    @title = "Genericon Trivia System - Anime Continuum Details - #{@anime_continuum.name}"
  end

  # Render the screen to insert a new AnimeContinuum.
  def new
    # We reference flash[] in case we've been redirected following an error.
    @anime_continuum = AnimeContinuum.new flash[:anime_continuum]
    @submission_note = flash[:submission_note]
    @title = 'Genericon Trivia System - New Anime Continuum'
  end

  # Create a new AnimeContinuum and redirect back to new.
  def create
    # Translate the alternate_names hash to an array...
    params[:anime_continuum][:alternate_names] = params[:anime_continuum][:alternate_names].keys.sort.collect { |key| params[:anime_continuum][:alternate_names][key] }
    # Remove duplicate alternate_names.
    params[:anime_continuum][:alternate_names] = params[:anime_continuum][:alternate_names].uniq
    # Remove empty alternate_names
    params[:anime_continuum][:alternate_names] = params[:anime_continuum][:alternate_names].reject { |name| name.nil? or name.empty? }
    
    # Create the AnimeContinuum.
    @anime_continuum = AnimeContinuum.new params[:anime_continuum]
    
    # Create the submission note if there was one.
    if !params[:submission_note].nil? and !params[:submission_note].empty?
      comment = @anime_continuum.comments.build :item => @anime_continuum, :text => params[:submission_note], :comment_type => :submission_note
    end

    success = true
    AnimeContinuum.transaction do
      Comment.transaction do
        success &&= @anime_continuum.save
        success &&= comment.save if !params[:submission_note].nil? and !params[:submission_note].empty?
      end
    end
    
    # Save and redirect.
    if success
      flash[:notice] = "Successfully submitted your anime continuum."
      
      # Auto approve if we're logged in as an approver.
      if current_user.has_role? 'anime_continuum_approver'
        @anime_continuum.approve!
      end
    else
      # Render the errors, and retain the values input previously for next time.
      flash[:error] = render_to_string :inline => "<%= error_messages_for 'anime_continuum' %>"
      flash[:error_is_html] = true
      flash[:anime_continuum] = params[:anime_continuum]
      flash[:submission_note] = params[:submission_note]
    end
    
    # Redirect back to the new path.
    redirect_to new_anime_continuum_path
  end

  # Shows the edit page for an AnimeContinuum.
  def edit
    @anime_continuum = AnimeContinuum.find params[:id]
    @title = "Genericon Trivia System - Edit Anime Continuum - #{@anime_continuum.name}"
  end
  
  # Adds an alternate name to an AnimeContinuum.
  def add_alternate_name
    @anime_continuum = AnimeContinuum.find params[:id]
    
    # Moderators automatically add the alternate name, while approvers add it as a pending change.
    if current_user.has_role? 'anime_continuum_moderator'
      @anime_continuum.alternate_names << params[:new_alternate_name]
    elsif current_user.has_role? 'anime_continuum_approver'
      @anime_continuum.add_pending_change_support :alternate_names, :add, params[:new_alternate_name]
    end
    if @anime_continuum.save
      flash[:notice] = "Successfully submitted the alternate name."
    else
      # Render the errors, and retain the values input previously for next time.
      flash[:error] = render_to_string :inline => "<%= error_messages_for 'anime_continuum' %>"
      flash[:error_is_html] = true
    end
    
    redirect_to edit_anime_continuum_path, params[:id]
  end
  
  # Removes an alternate_name from an AnimeContinuum.
  def remove_alternate_name
    @anime_continuum = AnimeContinuum.find params[:id]
    
    # Moderators automatically remove the alternate name, while approvers add it as a pending change.
    if @anime_continuum.alternate_names.include? params[:name]
      if current_user.has_role? 'anime_continuum_moderator'
        @anime_continuum.alternate_names.delete params[:name]
      elsif current_user.has_role? 'anime_continuum_approver'
        @anime_continuum.add_pending_change_support :alternate_names, :remove, params[:name]
      end
      if @anime_continuum.save
        flash[:notice] = "Successfully removed the alternate name."
      else
        # Render the errors, and retain the values input previously for next time.
        flash[:error] = render_to_string :inline => "<%= error_messages_for 'anime_continuum' %>"
        flash[:error_is_html] = true
      end
    end
    
    redirect_to edit_anime_continuum_path, params[:id]
  end
  
  # Approves a pending change for an AnimeContinuum.
  def approve_pending_change
    @anime_continuum = AnimeContinuum.find params[:id]
    
    # Moderators commit the change.
    if current_user.has_role? 'anime_continuum_moderator'
      @anime_continuum.apply_pending_change params[:key].to_sym, params[:change_action].to_sym, params[:value]
    # Everyone else (read: approvers) just has their support added.
    else
      @anime_continuum.add_pending_change_support params[:key].to_sym, params[:change_action].to_sym, params[:value]
    end
    if @anime_continuum.save
      flash[:notice] = "Successfully approved the pending change."
    else
      # Render the errors, and retain the values input previously for next time.
      flash[:error] = render_to_string :inline => "<%= error_messages_for 'anime_continuum' %>"
      flash[:error_is_html] = true
    end
    
    redirect_to edit_anime_continuum_path, params[:id]
  end
  
  # Rejects a pending change for an AnimeContinuum.
  def reject_pending_change
    @anime_continuum = AnimeContinuum.find params[:id]
    
    # Moderators need to delete the pending change entirely
    if current_user.has_role? 'anime_continuum_moderator'
      @anime_continuum.delete_pending_change params[:key].to_sym, params[:change_action].to_sym, params[:value]
    # Everyone else (read: approvers) just has their support (if any) removed.
    else
      @anime_continuum.remove_pending_change_support params[:key].to_sym, params[:change_action].to_sym, params[:value]
    end
    if @anime_continuum.save
      flash[:notice] = "Successfully rejected the pending change."
    else
      # Render the errors, and retain the values input previously for next time.
      flash[:error] = render_to_string :inline => "<%= error_messages_for 'anime_continuum' %>"
      flash[:error_is_html] = true
    end
    
    redirect_to edit_anime_continuum_path, params[:id]
  end

  # Destroys an AnimeContinuum.
  def destroy
    # I sure hope you added a warning.
    # TODO: block when there are dependencies.
    
    if AnimeContinuum.destroy params[:id]
      flash[:notice] = "Successfully deleted the anime continuum."
    else
      flash[:error] = "There was a problem deleting that anime continuum."
    end
    
    redirect_to manage_anime_continuum_path
  end

  # Shows the list of AnimeContinuums to be managed.
  def manage
    # TODO: Here, we order by whether or not the continuum is approved.
    @anime_continuums = AnimeContinuum.paginate :all, :page => params[:page], :select => "*, CASE state WHEN 'pending_approval' THEN 1 WHEN 'pending_changes' THEN 2 WHEN 'approved' THEN 3 END AS state_id", :order => 'state_id ASC, approved_at DESC, created_at DESC, name ASC'
    @title = 'Genericon Trivia System - Anime Continuum Management'
  end

  # Approve an AnimeContinuum.
  def approve
    @anime_continuum = AnimeContinuum.find(params[:id])
    
    if @anime_continuum.pending_approval? and @anime_continuum.approvable_changes?
      if @anime_continuum.approve!
        flash[:notice] = "Successfully approved the anime continuum."
      else
        flash[:error] = render_to_string :inline => "<%= error_messages_for 'anime_continuum' %>"
        flash[:error_is_html] = true
      end
    elsif @anime_continuum.pending_changes? and @anime_continuum.approvable_changes?
      if @anime_continuum.approve_changes!
        flash[:notice] = "Successfully approved the anime continuum changes."
      else
        flash[:error] = render_to_string :inline => "<%= error_messages_for 'anime_continuum' %>"
        flash[:error_is_html] = true
      end
    else
      flash[:error] = render_to_string :inline => "<%= error_messages_for 'anime_continuum' %>"
      flash[:error_is_html] = true
    end
    
    redirect_to edit_anime_continuum_path, params[:id]
  end
  
  # Select the destination for a merge_into operation.
  def select_merge
    @anime_continuum = AnimeContinuum.find(params[:id])
    @title = 'Genericon Trivia System - Merge Anime Continuums'
  end
  
  # Merge the given AnimeContinuum into the destination AnimeContinuum.
  def merge_into
    @src = AnimeContinuum.find(params[:id])
    @dest = AnimeContinuum.find(params[:dest])
    
    if @src.merge_into! @dest
      flash[:notice] = "Successfully merged the anime continuums."
    else
      flash[:error] = "Could not merge the anime continuums."
      flash[:error_is_html] = false
    end
    
    redirect_to manage_anime_continuum_path
  end
  
  # Return the list of anime continuums to merge into.
  def select_merge_search
    find_options = {
      :conditions => [ "LOWER(name) LIKE ?", '%' + params[:name].downcase + '%' ],
      :order => "name ASC",
      :limit => 10 }
    
    @main = AnimeContinuum.find(params[:id])
    @items = AnimeContinuum.find(:all, find_options)
    
    if params[:name].empty?
      merge_list = ''
    else
      merge_list = render_to_string :partial => 'select_merge_search'
    end
    
    render :update do |page|
      AnimeContinuum.query_cache.clear_query_cache if AnimeContinuum.method_defined? :query_cache
      page.replace_html("merge_options", merge_list)
    end
  end

end
