require 'zip/zip'
require 'ftools'
require 'mime/types'
require 'filemagic'

class ZipFileWrapper
  def initialize entry
    @entry = entry
  end
  
  def original_filename
    File.basename(@entry.name)
  end
  
  def content_type
    MIME::Types.type_for(self.original_filename)
  end
  
  def size
    @entry.size
  end
  
  def to_tempfile
    tempfile = Tempfile.new("stream")
    tempfile.binmode
    tempfile.write(@entry.get_input_stream.read)
    tempfile.rewind
    tempfile
  end
  
  def inspect
    "#<ZipFileWrapper: name=#{@entry.name}>"
  end
  
  def not_image?
    tempfile = Tempfile.new("stream")
    tempfile.binmode
    tempfile.write(@entry.get_input_stream.read)
    tempfile.rewind
    type = begin
             Paperclip.run("identify", %Q[-format "%wx%h" "#{tempfile.path}"])
           rescue Paperclip::PaperclipCommandLineError
             "identify: no decode delegate for this image format"
           end
    type.match(/^identify: no decode delegate for this image format/)
  end
end

# Controls AnimeScene pages.
class AnimeSceneController < ApplicationController
  before_filter :login_required, :except => [:index, :list, :show, :new, :new_zip,
                                             :create, :create_zip,
                                             :auto_complete_for_anime_series_name,
                                             :find_anime_genres_like,
                                             :find_similar_genres_to_anime,
                                             :play, :image, :randomize_play_window,
                                             :new_game,
					     :remove_anime_series_from_game,
       					     :set_game_history_entry,
					     :set_max_from_series]
  before_filter :clear_unused, :except => [:play, :image, :randomize_play_window,
                                           :new_game, :remove_anime_series_from_game,
					   :set_game_history_entry,
					   :set_max_from_series]
  after_filter :store_location, :only => [:index, :list, :show, :new, :new_zip, :edit,
                                          :manage, :play, :new_game]
  
  require_role 'anime_scene_approver', :for_all_except => [:index, :list,
                                                           :show, :new, :new_zip,
                                                           :create, :create_zip,
                                                           :auto_complete_for_anime_series_name,
                                                           :find_anime_genres_like,
                                                           :find_similar_genres_to_anime,
                                                           :play, :image, :randomize_play_window,
                                                           :new_game,
							   :remove_anime_series_from_game,
							   :set_game_history_entry,
							   :set_max_from_series]
  require_role 'anime_scene_moderator', :for => [:destroy]
  
  in_place_edit_for :anime_scene, :episode
  
  # Cleans the unused questions on exit.
  def clear_unused
    GameSession.clean_sessions(session[:unused]) if session[:unused]
    session[:unused] = nil
  end

  # Custom validation for setting anime series name.
  def set_anime_series_name
    @item = AnimeScene.find(params[:id])
    @anime_series = AnimeSeries.find_by_name params[:anime_series][:name]
    if @anime_series != nil
      @item.anime_series = @anime_series
      @item.save
      render :update do |page| 
        page.replace_html("anime_series_name_#{params[:id]}_in_place_editor", 
                          @item.anime_series.name)
      end
    else
      render :update do |page|
        page.alert('Must have a pre-existing anime set.')
        AnimeScene.query_cache.clear_query_cache if AnimeScene.method_defined?:query_cache
        AnimeSeries.query_cache.clear_query_cache if AnimeSeries.method_defined?:query_cache
        @item.reload
        page.replace_html("anime_series_name_#{params[:id]}_in_place_editor", 
                          @item.anime_series.name)
      end
    end
  end
  
  # Renders a new AnimeScene acceptance page.
  def index
    new
    render :action => :new
  end

  # Render a list of AnimeScenes.
  def list
    if !params[:letter].nil?
      params[:page] = (AnimeScene.count(:conditions => ["LOWER(`anime_series`.name) < LOWER(?)", params[:letter]], :joins => 'INNER JOIN `anime_series` ON `anime_scenes`.anime_series_id=`anime_series`.id') / AnimeScene.per_page + 1).to_i
    end
    params.delete :letter
    @anime_scenes = AnimeScene.paginate :all, :page => params[:page], :select => '`anime_scenes`.*, `anime_series`.name', :joins => 'INNER JOIN `anime_series` ON `anime_scenes`.anime_series_id=`anime_series`.id', :order => '`anime_series`.name ASC, `anime_scenes`.created_at ASC, `anime_scenes`.id ASC'
    @title = 'Genericon Trivia System - Name-That-Anime Scene List'
  end

  # Render the details of an AnimeScene for normal users.
  def show
    @anime_scene = AnimeScene.find params[:id]
    @anime_series = @anime_scene.anime_series
    @title = "Genericon Trivia System - Name-That-Anime Scene Details"
  end

  # Render the screen to insert a new AnimeScene.
  def new
    # We reference flash[] in case we've been redirected following an error.
    @anime_scene = AnimeScene.new flash[:anime_scene]
    @anime_series = AnimeSeries.new flash[:anime_series]
    @genre_checkbox = { }
    @genre_checkbox.default = '0'
    @genre_checkbox.merge! flash[:genre_checkbox] if !flash[:genre_checkbox].nil?
    @submission_note = flash[:submission_note]
    @title = 'Genericon Trivia System - New Name-That-Anime Scene'
  end

  # Render the screen to insert a group of new AnimeScenes.
  def new_zip
    # We reference flash[] in case we've been redirected following an error.
    @anime_scene = AnimeScene.new flash[:anime_scene]
    @anime_series = AnimeSeries.new flash[:anime_series]
    @genre_checkbox = { }
    @genre_checkbox.default = '0'
    @genre_checkbox.merge! flash[:genre_checkbox] if !flash[:genre_checkbox].nil?
    @submission_note = flash[:submission_note]
    @title = 'Genericon Trivia System - Name-That-Anime Scene Batch Upload'
  end

  # Create a new AnimeScene and redirect back to new.
  def create
    # Validate existence of at least one checked genre.  We can bypass this if
    # we WANT, but we don't want to.
    has_one_genre = false
    params[:genre_checkbox].each do |genre, value|
      if value == '1'
        has_one_genre = true
        break
      end
    end
    
    # Load the series
    create_series = false
    @anime_series = AnimeSeries.find_by_name(params[:anime_series][:name])
    # If we can't load it, create it.
    if @anime_series == nil
      @anime_series = AnimeSeries.new(params[:anime_series])
      params[:genre_checkbox].each do |genre, value|
        if value == '1'
          @anime_series.anime_genres << AnimeGenre.find_by_underscored_name(genre)
        end
      end
      create_series = true
    end
    
    # Check for changes in the genres
    AnimeGenre.find(:all).each do |genre|
      # If we have a difference between genres, note the potential change.
      if (params[:genre_checkbox][genre.underscored_name] == '1') != @anime_series.anime_genres.include?(genre)
        if @anime_series.anime_genres.include? genre
          @anime_series.add_pending_change_support :anime_genres, :remove, genre.id
        else
          @anime_series.add_pending_change_support :anime_genres, :add, genre.id
        end
      end
    end
    
    @anime_scene = AnimeScene.new params[:anime_scene]
    @anime_scene.anime_series = @anime_series

    # Create the submission note if there was one.
    if !params[:submission_note].nil? and !params[:submission_note].empty?
      comment = @anime_scene.comments.build :item => @anime_scene, :text => params[:submission_note], :comment_type => :submission_note
    end
    
    # Run validation checks to get full list of errors...
    @anime_series.valid?
    @anime_scene.valid?
    
    if has_one_genre
      success = true
      AnimeScene.transaction do
        AnimeSeries.transaction do
          Comment.transaction do
            success &&= @anime_series.save
            success &&= @anime_scene.save
            success &&= comment.save if !params[:submission_note].nil? and !params[:submission_note].empty?
          end
        end
      end
    else
      @anime_series.errors.add 'genres', '-- you must select at least one'
    end

    # Save and redirect.
    if success
      flash[:notice] = "Successfully submitted your Name-That-Anime scene."
      
      # Auto approve if we're logged in as an approver.
      if current_user.has_role? 'anime_scene_approver'
        @anime_scene.approve!
      end
      if current_user.has_role? 'anime_series_approver'
        @anime_series.approve!
      end
    else
      # Render the errors, and retain the values input previously for next time.
      flash[:error] = render_to_string :inline => "<%= error_messages_for 'anime_scene' %><%= error_messages_for 'anime_series' %>"
      flash[:error_is_html] = true
      flash[:anime_scene] = params[:anime_scene].reject {|key, value| key == :anime_scene}
      flash[:anime_series] = params[:anime_series]
      flash[:genre_checkbox] = params[:genre_checkbox]
      flash[:submission_note] = params[:submission_note]
    end
    
    # Redirect back to the new path.
    redirect_to new_anime_scene_path
  end

  # Create a batch of new AnimeScenes and redirect back to new.
  def create_zip
    begin
    # Validate existence of at least one checked genre.  We can bypass this if
    # we WANT, but we don't want to.
    has_one_genre = false
    params[:genre_checkbox].each do |genre, value|
      if value == '1'
        has_one_genre = true
        break
      end
    end
    
    # Load the series
    create_series = false
    @anime_series = AnimeSeries.find_by_name(params[:anime_series][:name])
    # If we can't load it, create it.
    if @anime_series == nil
      @anime_series = AnimeSeries.new(params[:anime_series])
      params[:genre_checkbox].each do |genre, value|
        if value == '1'
          @anime_series.anime_genres << AnimeGenre.find_by_underscored_name(genre)
        end
      end
      create_series = true
    end
    
    # Check for changes in the genres
    AnimeGenre.find(:all).each do |genre|
      # If we have a difference between genres, note the potential change.
      if (params[:genre_checkbox][genre.underscored_name] == '1') != @anime_series.anime_genres.include?(genre)
        if @anime_series.anime_genres.include? genre
          @anime_series.add_pending_change_support :anime_genres, :remove, genre.id
        else
          @anime_series.add_pending_change_support :anime_genres, :add, genre.id
        end
      end
    end
    
    zip = params[:anime_scene][:anime_scene]
    params[:anime_scene].delete :anime_scene
    @anime_scene = AnimeScene.new params[:anime_scene]

    scenes = []
    comments = []
    
    is_zip = false
    
    tempfile = Tempfile.new(zip.original_filename)
    tempfile.binmode
    tempfile.write(zip.read)
    tempfile.rewind
    tempfile.close
    
    if zip.content_type == 'application/zip'
      is_zip = true
    else
      # Some browsers don't set this type :-(
      fm = FileMagic.new(FileMagic::MAGIC_MIME)
      res = fm.file(tempfile.path)
      if res == 'application/zip'
        is_zip = true
      else
        @anime_scene.errors.add 'anime_scene', "^Image ZIP wasn't a zip file"
      end
      fm.close
    end
    
    if is_zip
      Zip::ZipFile::foreach(tempfile.path) {
        |entry|
	
        if entry.file?
          img = ZipFileWrapper.new(entry)
	  
	  if img.not_image?
	    next
    	  end
	  
	  # Yeah, this is because of some pass-by-ref issues in Production.
          anime_scene = @anime_scene.clone
          anime_scene.anime_series = @anime_series
	  anime_scene.anime_scene = img
          
          # Create the submission note if there was one.
          if !params[:submission_note].nil? and !params[:submission_note].empty?
            comment = @anime_scene.comments.build :item => @anime_scene, :text => params[:submission_note], :comment_type => :submission_note
            comments << comment
          end
	  
	  scenes << anime_scene
        end
      }
    end
    tempfile.unlink
    
    # Run validation checks to get full list of errors...
    @anime_series.valid?
    scenes.each { |scene| scene.valid? }
    
    if has_one_genre and is_zip
      success = true
      AnimeSeries.transaction do
        success &&= @anime_series.save
        scenes.each do |scene|
          AnimeScene.transaction do
            Comment.transaction do
              success &&= scene.save
              if comments.size > 0
                comment = comments.pop
                success &&= comment.save if !params[:submission_note].nil? and !params[:submission_note].empty?
              end
            end
          end
        end
      end
    elsif !has_one_genre
      @anime_series.errors.add 'genres', '-- you must select at least one'
    end

    # Save and redirect.
    if success
      flash[:notice] = "Successfully submitted your Name-That-Anime scenes."
      
      # Auto approve if we're logged in as an approver.
      if current_user.has_role? 'anime_scene_approver'
        scenes.each do |scene|
          scene.approve!
        end
      end
      if current_user.has_role? 'anime_series_approver'
        @anime_series.approve!
      end
    else
      # Render the errors, and retain the values input previously for next time.
      flash[:error] = render_to_string :inline => "<%= error_messages_for 'anime_scene' %><%= error_messages_for 'anime_series' %>"
      flash[:error_is_html] = true
      flash[:anime_scene] = params[:anime_scene].reject {|key, value| key == :anime_scene}
      flash[:anime_series] = params[:anime_series]
      flash[:genre_checkbox] = params[:genre_checkbox]
      flash[:submission_note] = params[:submission_note]
    end
    
    rescue SystemStackError
      logger.error $!
      logger.error caller[0..100]
    end

    # Redirect back to the new path.
    redirect_to new_zip_anime_scene_path
  end

  # Shows the edit page for an AnimeScene.
  def edit
    @anime_scene = AnimeScene.find params[:id]
    @anime_series = @anime_scene.anime_series || AnimeSeries.new()
    @title = "Genericon Trivia System - Edit Name-That-Anime Scene"
  end

  # Approves a pending change for an AnimeScene.
  def approve_pending_change
    @anime_scene = AnimeScene.find params[:id]
    
    # Moderators commit the change.
    if current_user.has_role? 'anime_scene_moderator'
      @anime_scene.apply_pending_change params[:key].to_sym, params[:change_action].to_sym, params[:value]
    # Everyone else (read: approvers) just has their support added.
    else
      @anime_scene.add_pending_change_support params[:key].to_sym, params[:change_action].to_sym, params[:value]
    end
    if @anime_scene.save
      flash[:notice] = "Successfully approved the pending change."
    else
      # Render the errors, and retain the values input previously for next time.
      flash[:error] = render_to_string :inline => "<%= error_messages_for 'anime_scene' %>"
      flash[:error_is_html] = true
    end
    
    redirect_to edit_anime_scene_path, params[:id]
  end
  
  # Rejects a pending change for an AnimeScene.
  def reject_pending_change
    @anime_scene = AnimeScene.find params[:id]
    
    # Moderators need to delete the pending change entirely
    if current_user.has_role? 'anime_scene_moderator'
      @anime_scene.delete_pending_change params[:key].to_sym, params[:change_action].to_sym, params[:value]
    # Everyone else (read: approvers) just has their support (if any) removed.
    else
      @anime_scene.remove_pending_change_support params[:key].to_sym, params[:change_action].to_sym, params[:value]
    end
    if @anime_scene.save
      flash[:notice] = "Successfully rejected the pending change."
    else
      # Render the errors, and retain the values input previously for next time.
      flash[:error] = render_to_string :inline => "<%= error_messages_for 'anime_scene' %>"
      flash[:error_is_html] = true
    end
    
    redirect_to edit_anime_scene_path, params[:id]
  end

  # Destroys an AnimeScene.
  def destroy
    # I sure hope you added a warning.
    # TODO: block when there are dependencies.
    
    if AnimeScene.destroy params[:id]
      flash[:notice] = "Successfully deleted the Name-That-Anime scene."
    else
      flash[:error] = "There was a problem deleting that Name-That-Anime scene."
    end
    
    redirect_to manage_anime_scene_path
  end
  
  # Shows the list of AnimeScenes to be managed.
  def manage
    order = params[:order_by] || 'state_id ASC, `anime_scenes`.approved_at DESC, `anime_scenes`.created_at DESC'
    @anime_scenes = AnimeScene.paginate :all, :page => params[:page], :select => "`anime_scenes`.*, CASE `anime_scenes`.state WHEN 'pending_approval' THEN 1 WHEN 'pending_changes' THEN 2 WHEN 'approved' THEN 3 END AS state_id, `anime_series`.name, `ratings`.rating", :joins => "INNER JOIN `anime_series` ON `anime_scenes`.anime_series_id=`anime_series`.id LEFT JOIN (SELECT item_id, SUM(rating) AS rating FROM `game_history_entries` GROUP BY item_type, item_id HAVING item_type='AnimeScene') AS `ratings` ON `anime_scenes`.id=`ratings`.item_id", :order => order
    @order = order
    @title = 'Genericon Trivia System - Name-That-Anime Scene Management'
  end
  
  # Approve an AnimeScene.
  def approve
    @anime_scene = AnimeScene.find(params[:id])
    
    if @anime_scene.pending_approval? and @anime_scene.approvable_changes?
      if @anime_scene.approve!
        flash[:notice] = "Successfully approved Name-That-Anime scene."
      else
        flash[:error] = render_to_string :inline => "<%= error_messages_for 'anime_scene' %>"
        flash[:error_is_html] = true
      end
    elsif @anime_scene.pending_changes? and @anime_scene.approvable_changes?
      if @anime_scene.approve_changes!
        flash[:notice] = "Successfully approved Name-That-Anime scene changes."
      else
        flash[:error] = render_to_string :inline => "<%= error_messages_for 'anime_scene' %>"
        flash[:error_is_html] = true
      end
    else
      flash[:error] = render_to_string :inline => "<%= error_messages_for 'anime_scene' %>"
      flash[:error_is_html] = true
    end
    
    redirect_to edit_anime_scene_path, params[:id]
  end

  # Configure a new game.
  def new_game
    # End any active game.
    if !session[:unused].nil?
      begin
        @game_session = GameSession.find(session[:unused])
        if @game_session.game_type != 'anime_scene'
          GameSession.clean_sessions(session[:unused])
          session[:unused] = nil
        end
      rescue ActiveRecord::RecordNotFound
        session[:unused] = nil
      end
    end
    
    # Setup the new game.
  end

  # Play Name-That-Anime.
  def play
    # First, check the old session to make sure it's a valid one.  If we came
    # from another game, we NEED to reset.  We also need to gracefully restart
    # from a timed-out game.
    if !session[:unused].nil?
      begin
        @game_session = GameSession.find(session[:unused])
        if @game_session.game_type != 'anime_scene'
          GameSession.clean_sessions(session[:unused])
          session[:unused] = nil
        end
      rescue ActiveRecord::RecordNotFound
        session[:unused] = nil
      end
    end
    
    # Initialize the 'unused' array of scenes.
    if session[:unused].nil?
      @unused = AnimeScene.find(:all)
      @unused = @unused.select {|q| !q.pending_approval? and !q.anime_series.pending_approval? and (q.anime_series.anime_continuum.nil? or !q.anime_series.anime_continuum.pending_approval?)}
      @unused = @unused.collect { |q| q.id }
      @window = [ ]
      @counters = { }
      @counters.default = 0
      @max_from_series = params[:max_from_series].to_i
      @max_from_series = @unused.length if @max_from_series == 0
      
      9.times do
        id = @unused.slice!(rand(@unused.length))
        @scene = AnimeScene.find(id)
        
        # Make sure to filter by max_from_series (though this is sort of
        # irrelevant when starting.
        while @counters[@scene.anime_series.id] >= @max_from_series and !@unused.empty?
          id = @unused.slice!(rand(@unused.length))
          @scene = AnimeScene.find(id)
        end
        
        if @counters[@scene.anime_series.id] < @max_from_series
          @window << id
        end
        
        if @unused.empty?
          break
        end
      end
      
      # Remember to save the current window and unused.
      @game_session = GameSession.new(:game_type => 'anime_scene')
      @game_session.data = {:unused => @unused, :window => @window, :max_from_series => @max_from_series}
      @game_session.save
      session[:unused] = @game_session.id
    else
      @game_session = GameSession.find(session[:unused])
      @unused = @game_session.data[:unused]
      @window = @game_session.data[:window]
      @counters = { }
      @counters.default = 0
      @counters.merge!(@game_session.data[:counters]) if @game_session.data[:counters]
      @max_from_series = @game_session.data[:max_from_series]
      
      # Can keep selecting until the window is empty.
      if @window.empty?
        GameSession.clean_sessions(session[:unused])
        session[:unused] = nil
        flash[:notice] = 'All scenes have been used up.  Game over.'
        redirect_to :action => 'index'
        return
      end
    end
    
    # Remove the current id from @window and @unused and replace it in @window.
    if !params[:id].nil? and !params[:id].empty?
      id = params[:id].to_i
      @unused.delete(id)
      
      # Get the index of it in @window so we know where to insert the new one.
      index = @window.index(id)
      indices = [ ]
      while !index.nil?
        indices << index
        @window.delete_at(index)
        index = @window.index(id)
      end
      
      while !indices.empty? and !@unused.empty?
        insert_at = indices.shift
        new_id = @unused.slice!(rand(@unused.length))
        new_scene = AnimeScene.find(new_id)
        
        # Make sure to filter by max_from_series
        while @counters[new_scene.anime_series.id] >= @max_from_series and !@unused.empty?
          new_id = @unused.slice!(rand(@unused.length))
          new_scene = AnimeScene.find(new_id)
        end
        
        if @counters[new_scene.anime_series.id] < @max_from_series
          @window.insert(insert_at, new_id)
        end
        
        if @unused.empty?
          break
        end
      end
      
      # Remember to save @window and @unused.
      @game_session = GameSession.find(session[:unused])
      data = @game_session.data || { }
      data[:window] = @window
      data[:unused] = @unused
      @game_session.data = data
      @game_session.save
      
      @scene = AnimeScene.find(id)
      
      # Create the (initially neutral) GameHistoryEntry.
      @game_history_entry = GameHistoryEntry.new :item => @scene, :rating => 0
      @game_history_entry.save
      
      # Increment the appropriate counter and remove anything still from
      # now-disallowed series, saving the counter.
      @counters[@scene.anime_series.id] = @counters[@scene.anime_series.id] + 1
      
      if @counters[@scene.anime_series.id] >= @max_from_series
        remove_scenes = [ ]
        add_scenes = [ ]
        @window.each do |old_id|
          if AnimeScene.find(old_id).anime_series.id == @scene.anime_series.id
            # Remember to remove this now-disallowed scene.
            remove_scenes << old_id
            
            # And get a new scene to replace it with
            if !@unused.empty?
              new_id = @unused.slice!(rand(@unused.length))
              new_scene = AnimeScene.find(new_id)
              
              # Make sure to filter by max_from_series
              while @counters[new_scene.anime_series.id] >= @max_from_series and !@unused.empty?
                new_id = @unused.slice!(rand(@unused.length))
                new_scene = AnimeScene.find(new_id)
              end
              
              if @counters[new_scene.anime_series.id] < @max_from_series
                add_scenes << new_id
              end
            end
          end
        end
        remove_scenes.each do |id|
          insert_at = @window.index(id)
          @window.delete_at(insert_at)
          @window.insert(insert_at, add_scenes.shift) if !add_scenes.empty?
        end
        
        # Remember to save @window and @unused
        @game_session = GameSession.find(session[:unused])
        data = @game_session.data || { }
        data[:window] = @window
        data[:unused] = @unused
        @game_session.data = data
        @game_session.save
      end
      
      @game_session = GameSession.find(session[:unused])
      data = @game_session.data || { }
      data[:counters] = @counters
      @game_session.data = data
      @game_session.save
    else
      @scene = nil
    end
    
    @window = @window.collect {|id| AnimeScene.find(id)}
    
    @title = "Genericon Trivia System - Name-That-Anime"
  end

  # Shows an image corresponding to a given Name-That-Anime Scene
  def image
    @scene = AnimeScene.find(params[:id])
    @title = "Genericon Trivia System - Name-That-Anime"
    @simple_render = true
  end

  # Rerandomizes @window for play.
  def randomize_play_window
    @game_session = GameSession.find(session[:unused])
    @unused = @game_session.data[:unused]
    @window = @game_session.data[:window]
    @max_from_series = @game_session.data[:max_from_series]
    @counters = { }
    @counters.default = 0
    @counters.merge!(@game_session.data[:counters]) if @game_session.data[:counters]
    
    @window.each {|id| @unused << id}
    
    @window = [ ]
    9.times do
      id = @unused.slice!(rand(@unused.length))
      @scene = AnimeScene.find(id)
      
      # Make sure to filter by max_from_series (though this is sort of
      # irrelevant when starting.
      while @counters[@scene.anime_series.id] >= @max_from_series and !@unused.empty?
        id = @unused.slice!(rand(@unused.length))
        @scene = AnimeScene.find(id)
      end
      
      if @counters[@scene.anime_series.id] < @max_from_series
        @window << id
      end
      
      if @unused.empty?
        break
      end
    end
    
    # Remember to save the current window and unused.
    @game_session = GameSession.find(session[:unused])
    data = @game_session.data || { }
    data[:window] = @window
    data[:unused] = @unused
    @game_session.data = data
    @game_session.save
    
    @window = @window.collect {|id| AnimeScene.find(id)}
    
    render :partial => 'scene_window', :object => @window
  end

  def remove_anime_series_from_game
    id = params[:id]
    
    @game_session = GameSession.find(params[:session])
    @unused = @game_session.data[:unused]
    @window = @game_session.data[:window]
    @max_from_series = @game_session.data[:max_from_series]
    @counters = { }
    @counters.default = 0
    @counters.merge!(@game_session.data[:counters]) if @game_session.data[:counters]
    
    scenes_to_remove = AnimeScene.find(:all, :conditions => ['anime_series_id = ?', id.to_i]).to_set
    scenes_to_remove.collect! {|q| q.id}
    @window.reject! {|i| scenes_to_remove.include? i}
    @unused.reject! {|i| scenes_to_remove.include? i}
    
    # NOTE: Not committed!!
    if @window.length < 9
      (9 - @window.length).times do
        if @unused.empty?
          break
        end
        id = @unused.slice!(rand(@unused.length))
        @scene = AnimeScene.find(id)
        
        # Make sure to filter by max_from_series (though this is sort of
        # irrelevant when starting.
        while @counters[@scene.anime_series.id] >= @max_from_series and !@unused.empty?
          id = @unused.slice!(rand(@unused.length))
          @scene = AnimeScene.find(id)
        end
        
        if @counters[@scene.anime_series.id] < @max_from_series
          @window << id
        end
        
        if @unused.empty?
          break
        end
      end
    end
    
    # Remember to save the current window and unused.
    data = @game_session.data || { }
    data[:window] = @window
    data[:unused] = @unused
    @game_session.data = data
    @game_session.save
    
    @window = @window.collect {|id| AnimeScene.find(id)}
    
    render :partial => 'scene_window', :object => @window
  end

  # Set the game history entry for this scene.
  def set_game_history_entry
    id = params[:id]
    history_id = params[:history_id]
    rating = params[:rating].to_i
    if rating < -1
      rating = -1
    elsif rating > 1
      rating = 1
    end
    
    if history_id.nil?
      game_history_entry = GameHistoryEntry.new :item => AnimeScene.find(id), :rating => rating
    else
      game_history_entry = GameHistoryEntry.find(history_id)
      game_history_entry.rating = rating
    end
    game_history_entry.save
    
    render :text => game_history_entry.id
  end

  # Change the maximum questions from series.
  def set_max_from_series
    @game_session = GameSession.find(params[:session])
    data = @game_session.data || { }
    if params[:max_from_series].nil? or params[:max_from_series].empty? or params[:max_from_series].to_i == 0
      data[:max_from_series] = data[:unused].length
    else
      data[:max_from_series] = params[:max_from_series].to_i
    end
    @game_session.data = data
    @game_session.save
  end

end
