# Controls AnimeSeries pages.
class AnimeSeriesController < ApplicationController
  before_filter :login_required, :except => [:index, :list, :show, :new,
                                             :create, :auto_complete_for_anime_continuum_name,
                                             :find_anime_genres_like,
                                             :find_similar_genres_to_anime,
                                             :search_select_merge_anime_series]
  after_filter :store_location, :only => [:index, :list, :show, :new, :edit,
                                          :manage, :select_merge]
  
  require_role 'anime_series_approver', :for_all_except => [:index, :list,
                                                            :show, :new,
                                                            :create, :auto_complete_for_anime_continuum_name,
                                                            :find_anime_genres_like,
                                                            :find_similar_genres_to_anime,
                                                            :search_select_merge_anime_series]
  require_role 'anime_series_moderator', :for => [:destroy, :select_merge,
                                                  :merge_into]
  
  in_place_edit_with_validation_for :anime_series, :name
  in_place_edit_with_validation_for :anime_series, :anidb_id
  in_place_edit_with_validation_for :anime_series, :years
  
  # Custom validation for setting anime continuum name.
  def set_anime_continuum_name
    @item = AnimeSeries.find(params[:id])
    @anime_continuum = AnimeContinuum.find_by_name params[:anime_continuum][:name]
    if @anime_continuum != nil
      @item.anime_continuum = @anime_continuum
      @item.save
      render :update do |page| 
        page.replace_html("anime_continuum_name_#{params[:id]}_in_place_editor", 
                          @item.anime_continuum.name)
      end
    else
      render :update do |page|
        page.alert('Must have a pre-existing anime continuum set.')
        AnimeSeries.query_cache.clear_query_cache if AnimeSeries.method_defined? :query_cache
        AnimeContinuum.query_cache.clear_query_cache if AnimeContinuum.method_defined? :query_cache
        @item.reload
        page.replace_html("anime_continuum_name_#{params[:id]}_in_place_editor", 
                          @item.anime_continuum.name)
      end
    end
  end
  
  # Renders a list of AnimeSeries.
  def index
    list
    render :action => :list
  end

  # Render a list of AnimeSeries.
  def list
    @anime_series = AnimeSeries.paginate :all, :page => params[:page], :order => 'name ASC'
    @title = 'Genericon Trivia System - Anime Series List'
  end

  # Render the details of an AnimeSeries for normal users.
  def show
    @anime_series = AnimeSeries.find params[:id]
    @title = "Genericon Trivia System - Anime Series Details - #{@anime_series.name}"
  end

  # Render the screen to insert a new AnimeSeries.
  def new
    # We reference flash[] in case we've been redirected following an error.
    @anime_series = AnimeSeries.new flash[:anime_series]
    @genre_checkbox = { }
    @genre_checkbox.default = '0'
    @genre_checkbox.merge! flash[:genre_checkbox] if !flash[:genre_checkbox].nil?
    @anime_continuum = AnimeContinuum.new flash[:anime_continuum]
    @submission_note = flash[:submission_note]
    @title = 'Genericon Trivia System - New Anime Series'
  end

  # Create a new AnimeSeries and redirect back to new.
  def create
    # Clean up the parameters, like years and alternate_names
    if params[:anime_series][:years].empty?
      params[:anime_series][:years] = nil
    end
    
    # Translate the alternate_names hash to an array...
    params[:anime_series][:alternate_names] = params[:anime_series][:alternate_names].keys.sort.collect { |key| params[:anime_series][:alternate_names][key] }
    # Remove duplicate alternate_names.
    params[:anime_series][:alternate_names] = params[:anime_series][:alternate_names].uniq
    # Remove empty alternate_names
    params[:anime_series][:alternate_names] = params[:anime_series][:alternate_names].reject { |name| name.nil? or name.empty? }
	
    # Validate existence of at least one checked genre.  We can bypass this if
    # we WANT, but we don't want to.
    has_one_genre = false
    params[:genre_checkbox].each do |genre, value|
      if value == '1'
        has_one_genre = true
        break
      end
    end
    
    # Load the continuum
    create_continuum = false
    if !params[:anime_continuum][:name].empty?
      @anime_continuum = AnimeContinuum.find_by_name(params[:anime_continuum][:name])
      # If we can't load it, create it.
      if @anime_continuum == nil
        @anime_continuum = AnimeContinuum.new(params[:anime_continuum])
        create_continuum = true
      end
    else
      @anime_continuum = nil
    end
    
    # Create the AnimeSeries.
    @anime_series = AnimeSeries.new params[:anime_series]
    @anime_series.anime_continuum = @anime_continuum

    params[:genre_checkbox].each do |genre, value|
      if value == '1'
        @anime_series.anime_genres << AnimeGenre.find_by_underscored_name(genre)
      end
    end
    
    # Create the submission note if there was one.
    if !params[:submission_note].nil? and !params[:submission_note].empty?
      comment = @anime_series.comments.build :item => @anime_series, :text => params[:submission_note], :comment_type => :submission_note
    end
    
    @anime_continuum.valid? if !@anime_continuum.nil?
    @anime_series.valid?
    
    if has_one_genre
      success = true
      AnimeSeries.transaction do
        AnimeContinuum.transaction do
          Comment.transaction do
            success &&= @anime_continuum.save if !@anime_continuum.nil?
            success &&= @anime_series.save
            success &&= comment.save if !params[:submission_note].nil? and !params[:submission_note].empty?
          end
        end
      end
    else
      @anime_series.errors.add 'genres', '-- you must select at least one'
    end
    
    # Save and redirect.
    if success
      flash[:notice] = "Successfully submitted your anime series."
      
      # Auto approve if we're logged in as an approver.
      if current_user.has_role? 'anime_series_approver'
        @anime_series.approve!
      end
      if !@anime_continuum.nil? and current_user.has_role? 'anime_continuum_approver'
        @anime_continuum.approve!
      end
    else
      # Render the errors, and retain the values input previously for next time.
      flash[:error] = render_to_string :inline => "<%= error_messages_for 'anime_series' %>"
      flash[:error_is_html] = true
      flash[:anime_series] = params[:anime_series]
      flash[:genre_checkbox] = params[:genre_checkbox]
      flash[:anime_continuum] = params[:anime_continuum]
      flash[:submission_note] = params[:submission_note]
    end
    
    # Redirect back to the new path.
    redirect_to new_anime_series_path
  end

  # Shows the edit page for an AnimeSeries.
  def edit
    @anime_series = AnimeSeries.find params[:id]
    @anime_continuum = @anime_series.anime_continuum || AnimeSeries.new()
    @title = "Genericon Trivia System - Edit Anime Series - #{@anime_series.name}"
  end
  
  # Adds an alternate name to an AnimeSeries.
  def add_alternate_name
    @anime_series = AnimeSeries.find params[:id]
    
    # Moderators automatically add the alternate name, while approvers add it as a pending change.
    if current_user.has_role? 'anime_series_moderator'
      @anime_series.alternate_names << params[:new_alternate_name]
    elsif current_user.has_role? 'anime_series_approver'
      @anime_series.add_pending_change_support :alternate_names, :add, params[:new_alternate_name]
    end
    if @anime_series.save
      flash[:notice] = "Successfully submitted the alternate name."
    else
      # Render the errors, and retain the values input previously for next time.
      flash[:error] = render_to_string :inline => "<%= error_messages_for 'anime_series' %>"
      flash[:error_is_html] = true
    end
    
    redirect_to edit_anime_series_path, params[:id]
  end
  
  # Removes an alternate name from an AnimeSeries.
  def remove_alternate_name
    @anime_series = AnimeSeries.find params[:id]
    
    # Moderators automatically remove the alternate name, while approvers add it as a pending change.
    if @anime_series.alternate_names.include? params[:name]
      if current_user.has_role? 'anime_series_moderator'
        @anime_series.alternate_names.delete params[:name]
      elsif current_user.has_role? 'anime_series_approver'
        @anime_series.add_pending_change_support :alternate_names, :remove, params[:name]
      end
      if @anime_series.save
        flash[:notice] = "Successfully removed the alternate name."
      else
        # Render the errors, and retain the values input previously for next time.
        flash[:error] = render_to_string :inline => "<%= error_messages_for 'anime_series' %>"
        flash[:error_is_html] = true
      end
    end
    
    redirect_to edit_anime_series_path, params[:id]
  end
  
  # Adds an anime genre to an AnimeSeries.
  def add_anime_genre
    if params[:new_anime_genre] == ''
      redirect_to edit_anime_series_path, params[:id]
      return
    end
    
    @anime_series = AnimeSeries.find params[:id]
    
    # Moderators automatically add the anime genre, while approvers add it as a pending change.
    if current_user.has_role? 'anime_series_moderator'
      @anime_series.anime_genres << AnimeGenre.find(params[:new_anime_genre])
    elsif current_user.has_role? 'anime_series_approver'
      @anime_series.add_pending_change_support :anime_genres, :add, params[:new_anime_genre]
    end
    if @anime_series.save
      flash[:notice] = "Successfully submitted the genre."
    else
      # Render the errors, and retain the values input previously for next time.
      flash[:error] = render_to_string :inline => "<%= error_messages_for 'anime_series' %>"
      flash[:error_is_html] = true
    end
    
    redirect_to edit_anime_series_path, params[:id]
  end
  
  # Removes an AnimeGenre from an AnimeSeries.
  def remove_anime_genre
    @anime_series = AnimeSeries.find params[:id]
    
    # Moderators automatically remove the genre, while approvers add it as a pending change.
    if @anime_series.anime_genres.include? AnimeGenre.find(params[:genre_id])
      if current_user.has_role? 'anime_series_moderator'
        @anime_series.anime_genres.delete AnimeGenre.find(params[:genre_id])
      elsif current_user.has_role? 'anime_series_approver'
        @anime_series.add_pending_change_support :anime_genres, :remove, params[:genre_id]
      end
      if @anime_series.save
        flash[:notice] = "Successfully removed the genre."
      else
        # Render the errors, and retain the values input previously for next time.
        flash[:error] = render_to_string :inline => "<%= error_messages_for 'anime_series' %>"
        flash[:error_is_html] = true
      end
    end
    
    redirect_to edit_anime_series_path, params[:id]
  end
  
  # Approves a pending change for an AnimeSeries.
  def approve_pending_change
    @anime_series = AnimeSeries.find params[:id]
    
    # Moderators commit the change.
    if current_user.has_role? 'anime_series_moderator'
      @anime_series.apply_pending_change params[:key].to_sym, params[:change_action].to_sym, params[:value]
    # Everyone else (read: approvers) just has their support added.
    else
      @anime_series.add_pending_change_support params[:key].to_sym, params[:change_action].to_sym, params[:value]
    end
    if @anime_series.save
      flash[:notice] = "Successfully approved the pending change."
    else
      # Render the errors, and retain the values input previously for next time.
      flash[:error] = render_to_string :inline => "<%= error_messages_for 'anime_series' %>"
      flash[:error_is_html] = true
    end
    
    redirect_to edit_anime_series_path, params[:id]
  end
  
  # Rejects a pending change for an AnimeSeries.
  def reject_pending_change
    @anime_series = AnimeSeries.find params[:id]
    
    # Moderators need to delete the pending change entirely
    if current_user.has_role? 'anime_series_moderator'
      @anime_series.delete_pending_change params[:key].to_sym, params[:change_action].to_sym, params[:value]
    # Everyone else (read: approvers) just has their support (if any) removed.
    else
      @anime_series.remove_pending_change_support params[:key].to_sym, params[:change_action].to_sym, params[:value]
    end
    if @anime_series.save
      flash[:notice] = "Successfully rejected the pending change."
    else
      # Render the errors, and retain the values input previously for next time.
      flash[:error] = render_to_string :inline => "<%= error_messages_for 'anime_series' %>"
      flash[:error_is_html] = true
    end
    
    redirect_to edit_anime_series_path, params[:id]
  end

  # Destroys an AnimeSeries.
  def destroy
    # I sure hope you added a warning.
    # TODO: block when there are dependencies.
    
    if AnimeSeries.destroy params[:id]
      flash[:notice] = "Successfully deleted the anime series."
    else
      flash[:error] = "There was a problem deleting that anime series."
    end
    
    redirect_to manage_anime_series_path
  end

  # Shows the list of AnimeSeries to be managed.
  def manage
    # TODO: Here, we order by whether or not the series is approved.
    @anime_series = AnimeSeries.paginate :all, :page => params[:page], :select => "*, CASE state WHEN 'pending_approval' THEN 1 WHEN 'pending_changes' THEN 2 WHEN 'approved' THEN 3 END AS state_id", :order => 'state_id ASC, approved_at DESC, created_at DESC, name ASC'
    @title = 'Genericon Trivia System - Anime Series Management'
  end

  # Approve an AnimeSeries.
  def approve
    @anime_series = AnimeSeries.find(params[:id])
    
    if @anime_series.pending_approval? and @anime_series.approvable_changes?
      if @anime_series.approve!
        flash[:notice] = "Successfully approved anime series."
      else
        flash[:error] = render_to_string :inline => "<%= error_messages_for 'anime_series' %>"
        flash[:error_is_html] = true
      end
    elsif @anime_series.pending_changes? and @anime_series.approvable_changes?
      if @anime_series.approve_changes!
        flash[:notice] = "Successfully approved anime series changes."
      else
        flash[:error] = render_to_string :inline => "<%= error_messages_for 'anime_series' %>"
        flash[:error_is_html] = true
      end
    else
      flash[:error] = render_to_string :inline => "<%= error_messages_for 'anime_series' %>"
      flash[:error_is_html] = true
    end
    
    redirect_to edit_anime_series_path, params[:id]
  end

  # Select the destination for a merge_into operation.
  def select_merge
    @anime_series = AnimeSeries.find(params[:id])
    @title = 'Genericon Trivia System - Merge Anime Series'
  end
  
  # Merge the given AnimeSeries into the destination AnimeSeries.
  def merge_into
    @src = AnimeSeries.find(params[:id])
    @dest = AnimeSeries.find(params[:dest])
    
    if @src.merge_into! @dest
      flash[:notice] = "Successfully merged the anime series."
    else
      flash[:error] = "Could not merge the anime series."
      flash[:error_is_html] = false
    end
    
    redirect_to manage_anime_series_path
  end
  
  # Return the list of anime series to merge into.
  def select_merge_search
    find_options = {
      :conditions => [ "LOWER(name) LIKE ?", '%' + params[:name].downcase + '%' ],
      :order => "name ASC",
      :limit => 10 }
    
    @main = AnimeSeries.find(params[:id])
    @items = AnimeSeries.find(:all, find_options)
    
    if params[:name].empty?
      merge_list = ''
    else
      merge_list = render_to_string :partial => 'select_merge_search'
    end
    
    render :update do |page|
      AnimeSeries.query_cache.clear_query_cache if AnimeSeries.method_defined? :query_cache
      page.replace_html("merge_options", merge_list)
    end
  end

end
