# Controls User pages for sign up and administration.
class UsersController < ApplicationController
  # Protect these actions behind an admin login
  before_filter :login_required, :only => [:suspend, :unsuspend, :destroy,
                                           :purge, :edit]
  before_filter :find_user, :only => [:suspend, :unsuspend, :destroy, :purge,
                                      :show, :edit, :update]
  
  require_role 'admin', :for_all_except => [:index, :new, :create, :activate,
                                            :list, :show, :edit,
                                            :reset_password,
                                            :send_reset_password_email, :update]

  # Renders user list.
  def index
    list
    render :action => 'list'
  end

  # Render the sign up screen.
  def new
    @user = User.new
    @title = 'Genericon Trivia System - Sign Up'
  end
 
  # Attempt to create a User by signing them up.
  def create
    logout_keeping_session!
    @user = User.new(params[:user])
    @user.register! if @user && @user.valid?
    success = @user && @user.valid?
    if success && @user.errors.empty?
      flash[:notice] = "Thanks for signing up!  We're sending you an email with your activation code."
      redirect_back_or_default '/'
    else
      flash.now[:error]  = "We couldn't set up that account, sorry.  Please try again, or contact an admin (link is above)."
      render :action => 'new'
    end
  end

  # Attempt to activate a User.
  def activate
    logout_keeping_session!
    user = User.find_by_activation_code(params[:activation_code]) unless params[:activation_code].blank?
    case
    when (!params[:activation_code].blank?) && user && !user.active?
      user.activate!
      flash[:notice] = "Signup complete! Please sign in to continue."
      redirect_to login_path
    when params[:activation_code].blank?
      flash[:error] = "The activation code was missing.  Please follow the URL from your email."
      redirect_to '/'
    else 
      flash[:error] = "We couldn't find a user with that activation code -- check your email? Or maybe you've already activated -- try signing in."
      redirect_to '/'
    end
  end
  
  # Prepare to send a new password-activation for a User.
  def reset_password
    logout_keeping_session!
    @user = User.find_by_activation_code(params[:activation_code]) unless params[:activation_code].blank?
    case
    when (!params[:activation_code].blank?) && @user && @user.active?
      @got_activation_code = true
    when params[:activation_code].blank?
      @got_activation_code = false
    else
      @got_activation_code = false
      flash.now[:error] = "We can't reset your password if you do not have an active account."
    end
    if flash[:user_error]
      @user_error = flash[:user_error]
    end
    @title = 'Genericon Trivia System - Reset Password'
  end
  
  # Attempt to send a new password-activation for a User.
  def send_reset_password_email
    user = User.find_by_email(params[:email])
    if user
      user.make_activation_code
      user.save!
      flash[:notice] = "An e-mail has been sent to you containing details on how to reset your password. Please follow the URL in the e-mail to login to change your password."
    else
      flash[:error] = "We couldn't find a user with that e-mail address."
    end
    redirect_to reset_password_path
  end

  # Suspend a User.
  def suspend
    if @user != current_user
      @user.suspend!
    else
      flash[:error] = "You can't suspend yourself!"
    end
    redirect_to users_path
  end

  # Unsuspend a User.
  def unsuspend
    @user.unsuspend! 
    redirect_to users_path
  end

  # Deletes a User without removing all records (by calling User#delete!)
  def destroy
    @user.delete!
    redirect_to users_path
  end

  # Completely removes all records of a User (by calling User#destroy)
  def purge
    @user.destroy
    redirect_to users_path
  end
  
  # Render the User list screen.
  def list
    # Only admins should see the 'Edit' and 'Suspend' columns.
    if current_user and current_user.has_role?('admin')
      @render_admin_actions = true
    else
      @render_admin_actions = false
    end
    @users = User.paginate :all, :page => params[:page], :order => 'created_at ASC'
    @title = 'Genericon Trivia System - User List'
  end
  
  # Render the User detail screen for normal users.
  def show
    @title = 'Genericon Trivia System - User Details: ' + @user.login
  end
  
  # Render the User edit screen only for admins and one's self.
  def edit
    if flash[:user_error]
      @user_error = flash[:user_error]
    end
    if @user == current_user || current_user.has_role?('admin')
      if current_user.has_role? 'admin'
        @stylesheets = ['user_edit']
      end
      @available_roles = Role.find(:all, :order => 'sort_order')
      @available_roles = @available_roles.select { |role| !@user.has_role? role }
      @title = "Genericon Trivia System - Editing User: #{@user.login}"
    else
      flash[:error] = "You aren't permitted to edit the user details of #{@user.login}"
      redirect_back_or_default index_users_path
    end
  end
  
  # Update the User's data.
  def update
    if params[:reset_password]
      @user = User.find_by_activation_code params[:activation_code]
      if @user.nil?
        flash[:error] = "We couldn't find your user login. Please report it to the administrator."
        redirect_to '/'
      else
        @user.password = params[:password]
        @user.password_confirmation = params[:password_confirmation]
        @user.activation_code = nil
        if !@user.valid?
          flash[:user_error] = @user
          redirect_to reset_password_path + '/' + params[:activation_code]
        else
          @user.save
          flash[:notice] = "Your password was successfully reset. You may now login."
          redirect_to login_path
        end
      end
    else
      if (current_user.has_role?('admin') || User.authenticate(params[:user][:login], params[:current_password])) && @user.update_attributes(params[:user])
        flash[:notice] = "User was successfully updated."
        redirect_to users_path
      else
        if @user.errors.empty?
          flash[:error] = "Couldn't update user details. Did you type the right password?"
        end
        redirect_to edit_user_path(@user.id)
      end
    end
  end
  
  # Add a role to a user.
  def add_role
    role = Role.find(params[:id].split('_')[1])
    @user = User.find(params[:user_id])
    # Remove encompassed roles and add this role.
    role.encompassed_roles.each do |encompassed_role|
      if @user.roles.include? encompassed_role
        @user.roles.delete encompassed_role
      end
    end
    @user.roles << role
    if @user.valid?
      @user.save
      @select_roles = [role.id]
      @hide_roles = role.encompassed_roles.collect { |role| role.id }
      flash.now[:notice] = "The role was successfully added."
    end
    @available_roles = Role.find(:all, :order => 'sort_order')
    @available_roles = @available_roles.select { |role| !@user.has_role? role }
    render :action => 'edit_role.js.rjs'
  end
  
  # Remove a role from a user.
  def remove_role
    role = Role.find(params[:id].split('_')[1])
    @user = User.find(params[:user_id])
    if role.name == 'admin'
      # Can't delete your own admin status!
      if @user == current_user
        flash.now[:error] = "You cannot remove admin status from yourself!"
      end
    else
      if @user.can_delegate? role
        @user.delegator_roles.delete(role)
      end
      @user.roles.delete(role)
      if @user.valid?
        @user.save
        @deselect_roles = [role.id] + role.encompassed_roles.collect { |role| role.id }
        flash.now[:notice] = "The role was successfully removed."
      end
    end
    @available_roles = Role.find(:all, :order => 'sort_order')
    @available_roles = @available_roles.select { |role| !@user.has_role? role }
    render :action => 'edit_role.js.rjs'
  end
  
  # Add a delegator role to a user.
  def add_delegator_role
    role = Role.find(params[:id].split('_')[2])
    @user = User.find(params[:user_id])
    # Remove encompassed roles and add this role.
    role.encompassed_roles.each do |encompassed_role|
      if @user.delegator_roles.include? encompassed_role
        @user.delegator_roles.delete encompassed_role
      end
    end
    @user.delegator_roles << role
    if @user.valid?
      @user.save
      flash.now[:notice] = "The role was successfully added."
    end
    @available_roles = Role.find(:all, :order => 'sort_order')
    @available_roles = @available_roles.select { |role| !@user.has_role? role }
    render :action => 'edit_role.js.rjs'
  end
  
  # Remove a delegator role from a user.
  def remove_delegator_role
    role = Role.find(params[:id].split('_')[2])
    @user = User.find(params[:user_id])
    if role.name == 'admin'
      flash.now[:error] = "You can't remove delegator status from administrators!"
    else
      @user.delegator_roles.delete(role)
      if @user.valid?
        @user.save
        flash.now[:notice] = "The delegator role was successfully removed."
      end
    end
    @available_roles = Role.find(:all, :order => 'sort_order')
    @available_roles = @available_roles.select { |role| !@user.has_role? role }
    render :action => 'edit_role.js.rjs'
  end
  
  # There's no page here to update or destroy a user.  If you add those, be
  # smart -- make sure you check that the visitor is authorized to do so, that they
  # supply their old password along with a new one to update it, etc.

protected
  # Finds the User to change status for.
  def find_user
    @user = User.find(params[:id])
  rescue ActiveRecord::RecordNotFound
    @user = nil
  end
end
