# Helper methods for the User class.
module UsersHelper
  #
  # Use this to wrap view elements that the User can't access.
  # 
  # <b>!! Note: this is an _interface_, not _security_ feature !!</b>
  #
  # You need to do all access control at the controller level.
  #
  # Examples:
  #   <%= if_authorized?(:index,   User)  do link_to('List all users', users_path) end %>
  #
  #   <%= if_authorized?(:edit,    @user) do link_to('Edit this user', edit_user_path) end %>
  #
  #   <%= if_authorized?(:destroy, @user) do link_to 'Destroy', @user, :confirm => 'Are you sure?', :method => :delete end %> 
  # 
  #
  def if_authorized?(action, resource, &block)
    if authorized? action, resource
      yield action, resource
    end
  end

  #
  # Link to user's page ('users/1')
  #
  # By default, their login is used as link text and link title (tooltip)
  #
  # Takes options
  # * <tt>:content_text</tt> => 'Content text in place of <tt>user.login</tt>', escaped with
  #   the standard <tt>h()</tt> function.
  # * <tt>:content_method</tt> => <tt>:user_instance_method_to_call_for_content_text</tt>
  # * <tt>:title_method</tt> => <tt>:user_instance_method_to_call_for_title_attribute</tt>
  # * as well as <tt>link_to()</tt>'s standard options
  #
  # Examples:
  #   link_to_user @user
  #   # => <a href="/users/3" title="barmy">barmy</a>
  #
  #   # if you've added a .name attribute:
  #   content_tag :span, :class => :vcard do
  #     (link_to_user user, :class => 'fn n', :title_method => :login, :content_method => :name) +
  #           ': ' + (content_tag :span, user.email, :class => 'email')
  #   end
  #   # => <span class="vcard"><a href="/users/3" title="barmy" class="fn n">Cyril Fotheringay-Phipps</a>: <span class="email">barmy@blandings.com</span></span>
  #
  #   link_to_user @user, :content_text => 'Your user page'
  #   # => <a href="/users/3" title="barmy" class="nickname">Your user page</a>
  #
  def link_to_user(user, options={})
    raise "Invalid user" unless user
    options.reverse_merge! :content_method => :login, :title_method => :login, :class => :nickname
    content_text      = options.delete :content_text
    content_text    ||= user.send options.delete(:content_method)
    options[:title] ||= user.send options.delete(:title_method)
    link_to h(content_text), user_path(user), options
  end

  #
  # Link to login page using remote ip address as link content
  #
  # <tt>:title</tt> (and thus, tooltip) is set to the IP address 
  # <tt>:tag</tt> will create a custom tag that is not a link.
  #
  # Examples:
  #   link_to_login_with_IP
  #   # => <a href="/login" title="169.69.69.69">169.69.69.69</a>
  #
  #   link_to_login_with_IP :content_text => 'not signed in'
  #   # => <a href="/login" title="169.69.69.69">not signed in</a>
  #
  def link_to_login_with_IP content_text=nil, options={}
    ip_addr           = request.remote_ip
    content_text    ||= ip_addr
    options.reverse_merge! :title => ip_addr
    if tag = options.delete(:tag)
      content_tag tag, h(content_text), options
    else
      link_to h(content_text), login_path, options
    end
  end

  #
  # Link to the current User's page (using link_to_user) or to the login page
  # (using link_to_login_with_IP).
  #
  def link_to_current_user(options={})
    if current_user
      link_to_user current_user, options
    else
      content_text = options.delete(:content_text) || 'not signed in'
      # kill ignored options from link_to_user
      [:content_method, :title_method].each{|opt| options.delete(opt)} 
      link_to_login_with_IP content_text, options
    end
  end
  
  # Generates the table cell that displays role overview information for a role.
  def generate_role_cell(user, role)
    role_class = ''
    role_status = ''
    role_reason = ''
    role_extra = ''
    if user.has_role? 'admin'
      role_class = 'has_role'
      role_status = 'Yes'
    elsif role.name == 'admin' && user.has_role?('super_moderator')
      role_class = 'partial_role'
      role_status = 'Super Moderator'
      if user.can_delegate? 'super_moderator'
        role_extra = 'Delegator'
      end
    elsif role.name == 'admin' && user.has_role?('super_approver')
      role_class = 'partial_role'
      role_status = 'Super Approver'
      if user.can_delegate? 'super_approver'
        role_extra = 'Delegator'
      end
    elsif user.can_delegate? role.name
      role_class = 'has_role'
      role_status = 'Yes'
      role_extra = 'Delegator'
    elsif role.name.include?('moderator') && user.has_role?(role.name)
      role_class = 'has_role'
      role_status = 'Yes'
    elsif user.has_role? role.name.gsub(/moderator/, 'approver')
      role_class = 'partial_role'
      role_status = 'Approval Only'
    elsif role.name != 'admin'
      role.encompassed_roles.each do |sub_role|
        if user.has_role? sub_role.name
          role_class = 'partial_role'
          role_status = 'Partial'
          break
        end
      end
    end
    if role_class == ''
      role_class = 'no_role'
      role_status = 'No'
    end
    details_link = ''
    #details_link = '<a href="' + show_roles_path(role.id) + '">Details</a>'
    
    if role_extra == ''
      "<td class=\"#{role_class}\"><div>#{role_status}</div></td>"
    else
      "<td class=\"#{role_class}\"><div>#{role_status} (#{role_extra})</div></td>"
    end
  end
end
