# A Comment is associated with any of the user-modifiable items, such as an
# AnimeSeries, AnimeTriviaQuestion, VideoGameSystem, etc.
#
# Comments may be threaded and thus may have a parent comment.  By default,
# all new comments have no parent (+parent_id+ of 0).
#
# Comments have an associated +type+, which is one of:
#
# * +:approval_pending_modification+ - marking a tentative approval of some item following a requested modification being made.
# * +:request_approval+ - marking a vote for the approval of some item,
# * +:request_deletion+ - marking a request for the deletion of some item,
# * +:request_modification+ - marking a request for the modification of some item (usually already approved),
# * +:request_for_comments+ - marking a request for other moderators to comment on some aspect of an item,
# * +:note+ - marking a general note to be displayed with an item,
# * +:submission_note+ - marking a note with additional information that was provided at submission time,
# * +:general_comment+ - any other type of comment that is none of the above (usually a comment in a thread that expresses no particular opinion).
class Comment < ActiveRecord::Base
  belongs_to :item, :polymorphic => true
  acts_as_tree :order => 'created_at'
  
  validates_inclusion_of :comment_type, :in => [ :approval_pending_modification,
      :request_approval, :request_deletion, :request_modification,
      :request_for_comments, :note, :submission_note, :general_comment,
      :request_handled ]
  validates_presence_of :item, :text
  
  attr_readonly :item, :item_id, :item_type
  
  stampable
  
  # Returns a hash with valid comment types as keys and their descriptions as values.
  def self.types
    { :approval_pending_modification => 'This is a tentative approval of this item, as long as a modification is made.',
      :request_approval              => 'This is a request to approve this item as it is.',
      :request_deletion              => 'This is a request to delete this item.',
      :request_modification          => 'This is a request for a modification to be made to this already-approved item.',
      :request_for_comments          => 'This is a request to get other moderators to comment on the issue before action is taken on this item.',
      :note                          => 'This is a note to be displayed with this item.',
      :submission_note               => 'This is a note made during the submission of this item.',
      :general_comment               => 'This is a general comment.',
      :request_handled               => 'This is a note to mark a request as having been handled.' }
  end
  
  # Returns comment type as a symbol.
  def comment_type
    read_attribute(:comment_type).to_sym
  end
  
  # Sets comment type as a string.
  def comment_type=(new_type)
    write_attribute :comment_type, new_type.to_s
  end
  
  # Need to override the << action on children
  alias :_children :children
  private :_children
  def children
    all_children = _children
    parent_comment = self
    all_children.metaclass.class_eval do
      # Redefine << to set the child comment's item.
      if !instance_methods.include? '_push'
        alias_method :_push, :<<
        define_method :<< do |*args|
          args.each { |arg| arg.item = parent_comment.item }
          
          _push *args
        end
        alias_method :push, :<<
        alias_method :concat, :<<
      end
    end
    all_children
  end
end
