require 'digest/sha1'

# A user of the Genericon Trivia System.
class User < ActiveRecord::Base
  include Authentication
  include Authentication::ByPassword
  include Authentication::ByCookieToken
  include Authorization::StatefulRoles
  validates_presence_of     :login
  validates_length_of       :login,    :within => 3..40
  validates_uniqueness_of   :login,    :case_sensitive => false
  validates_format_of       :login,    :with => RE_LOGIN_OK, :message => MSG_LOGIN_BAD

  validates_format_of       :name,     :with => RE_NAME_OK,  :message => MSG_NAME_BAD, :allow_nil => true
  validates_length_of       :name,     :maximum => 100

  validates_presence_of     :email
  validates_length_of       :email,    :within => 6..100 #r@a.wk
  validates_uniqueness_of   :email,    :case_sensitive => false
  validates_format_of       :email,    :with => RE_EMAIL_OK, :message => MSG_EMAIL_BAD
  
  # A dummy link class.  Use roles unless you really need to.
  has_many :roles_users
  has_many :roles, :through => :roles_users, :uniq => true
  # A dummy link class.  User delegator_roles unless you really need to.
  has_many :delegators_roles
  has_many :delegator_roles, :through => :delegators_roles, :source => :role, :uniq => true
  
  # Is the object used to stamp items.
  model_stamper
  # And can be stamped.
  stampable

  # HACK HACK HACK -- how to do attr_accessible from here?
  # prevents a user from submitting a crafted form that bypasses activation
  # anything else you want your user to change should be added here.
  attr_accessible :login, :email, :name, :password, :password_confirmation

  # Authenticates a User by their login name and unencrypted password.  Returns the user or nil.
  #
  # uff.  This is really an authorization, not authentication routine.  
  # We really need a Dispatch Chain here or something.
  # This will also let us return a human error message.
  #
  def self.authenticate(login, password)
    u = find_in_state :first, :active, :conditions => {:login => login} # need to get the salt
    u && u.authenticated?(password) ? u : nil
  end
  
  # has_role? simply needs to return true or false whether a user has a Role or not.  
  # It may be a good idea to have "admin" roles return true always
  def has_role?(role_in_question)
    role_list = self.roles.collect(&:name)
    return true if role_list.include?("admin")
    if !role_list.include?(role_in_question.to_s)
      role_list.each {|role| return true if Role.find_by_name(role).encompasses_role?(role_in_question.to_s)}
      return false
    else
      return true
    end
  end
  
  # can_delegate? simply needs to return true or false whether a user is a delegator for a Role or not.  
  # It may be a good idea to have "admin" roles return true always
  def can_delegate?(role_in_question)
    delegate_role_list = self.delegator_roles.collect(&:name)
    return true if has_role?('admin')
    if !delegate_role_list.include?(role_in_question.to_s)
      delegate_role_list.each {|role| return true if Role.find_by_name(role).encompasses_role?(role_in_question.to_s)}
      return false
    else
      return true
    end
  end
  
  # Creates a new activation code for a User.
  def make_activation_code
    self.deleted_at = nil
    self.activation_code = self.class.make_token
  end
end
