require 'yaml'

# Represents a single video game.
class VideoGame < ActiveRecord::Base
  has_many_values :alternate_names, :uniq => true, :dependent => :destroy, :as => :item
  has_many :comments, :dependent => :destroy, :as => :item
  belongs_to :approver, :class_name => 'User'
  belongs_to :video_game_continuum
  belongs_to :video_game_system
  has_and_belongs_to_many :video_game_genres, :uniq => true
#  has_many :anime_trivia_questions, :dependent => :destroy
#  has_many :anime_scenes, :dependent => :destroy
  
  validates_uniqueness_of :name, :message => 'has already been taken'
  validates_presence_of :name, :message => 'may not be empty'
  validates_presence_of :video_game_system, :message => 'may not be empty'
  validates_inclusion_of :year, :in => 1900..2100, :allow_nil => true, :message => 'must be between 1900 and 2100'
  
  before_save :render_pending_changes
  
  acts_as_state_machine :initial => :pending_approval
  
  state :pending_approval
  state :approved, :enter => :do_approval
  state :pending_changes
  
  event :approve do
    transitions :to => :approved, :from => :pending_approval
  end

  event :changed do
    transitions :to => :pending_changes, :from => :approved
  end
  
  event :approve_changes do
    transitions :to => :approved, :from => :pending_changes, :guard => :approvable_changes?
  end
  
  # Modifies the addition of a genre such that it doesn't cause an error.
  if !instance_methods.include? '_video_game_genres'
    alias_method :_video_game_genres, :video_game_genres
    define_method :video_game_genres do
      genres = _video_game_genres
      genres.metaclass.instance_eval do
        if !instance_methods.include? '_push'
          alias_method :_push, :<<
          define_method :<< do |*values|
            values = values - to_ary
            values.collect! { |value| _push(value) }
          end
          alias_method :push, :<<
          alias_method :concat, :<<
        end

        if !instance_methods.include? '_delete'
          alias_method :_delete, :delete
          define_method :delete do |*values|
            values = to_ary.select { |item| values.include? item }
            values.collect! { |value| _delete(value) }
          end
        end
      end
      genres
    end
  end

  # Renders the pending changes from the YAML in the database.
  def pending_changes
    unless @pending_changes_loaded or read_attribute(:pending_changes).nil?
      @pending_changes ||= YAML::load read_attribute(:pending_changes)
      if !@pending_changes.nil?
        # Build the proper pending changes hash mechanism before outputting our value.
        @pending_changes = (Hash.new {|h,k| h[k] = Hash.new {|h,k| h[k] = {}}}).merge @pending_changes
      end
    else
      @pending_changes ||= nil
    end
    @pending_changes_loaded = true
    @pending_changes
  end
  
  # Sets pending changes.
  def pending_changes=(value)
    if value.nil?
      @pending_changes = nil
    else
      @pending_changes = (Hash.new {|h,k| h[k] = Hash.new {|h,k| h[k] = {}}}).merge value
    end
    @pending_changes_loaded = true
    render_pending_changes
  end
  
  # Applies a pending change and removes its pending status.
  def apply_pending_change(attribute, action, value)
    # Check to see if our pending changes have changed since we last checked.
    pending_changes unless @pending_changes_loaded
    
    # Build @pending_changes as needed.
    if @pending_changes.nil?
      @pending_changes = Hash.new {|h,k| h[k] = Hash.new {|h,k| h[k] = {}}}
    end
    
    # Basically, we just do exactly what the pending change says, if it exists, and then delete it.
    if action == :add
      # Construct the hash for that action as well.
      if !@pending_changes[attribute].has_key? :add
        @pending_changes[attribute][:add] = Hash.new {|h,k| h[k] = []}
      elsif @pending_changes[attribute][:add].default_proc.nil?
        # We loaded it from somewhere and don't have the default action.
        @pending_changes[attribute][:add] = (Hash.new {|h,k| h[k] = []}).merge @pending_changes[attribute][:add]
      end
      
      # No need to process further if the pending change doesn't exist.
      if !@pending_changes[attribute][:add].has_key? value
        return
      end
      
      # Ah, special cases here anyway...
      if attribute == :alternate_names
        self.alternate_names << value
      elsif attribute == :video_game_genres
        if !self.video_game_genres.include? VideoGameGenre.find(value)
          self.video_game_genres << VideoGameGenre.find(value)
        end
      end
    elsif action == :remove
      # Construct the hash for that action as well.
      if !@pending_changes[attribute].has_key? :remove
        @pending_changes[attribute][:remove] = Hash.new {|h,k| h[k] = []}
      elsif @pending_changes[attribute][:remove].default_proc.nil?
        # We loaded it from somewhere and don't have the default action.
        @pending_changes[attribute][:remove] = (Hash.new {|h,k| h[k] = []}).merge @pending_changes[attribute][:remove]
      end
      
      # No need to process further if the pending change doesn't exist.
      if !@pending_changes[attribute][:remove].has_key? value
        return
      end
      
      # Ah, special cases here anyway...
      if attribute == :alternate_names
        self.alternate_names.delete value
      elsif attribute == :video_game_genres
        self.video_game_genres.delete(VideoGameGenre.find(value))
      end
    else
      # No need to process further if the pending change doesn't exist.
      if !@pending_changes[attribute].has_key? value
        return
      end
      
      # Otherwise, apply the change.
      if attribute == :video_game_continuum
        self.video_game_continuum = VideoGameContinuum.find(value)
      elsif attribute == :video_game_system
        self.video_game_system = VideoGameSystem.find(value)
      else
        write_attribute attribute, value
      end
    end
    
    # Delete the pending change now that it's applied.
    delete_pending_change attribute, action, value
  end
  
  # Deletes a pending change (removing its pending status).
  def delete_pending_change(attribute, action, value)
    pending_changes unless @pending_changes_loaded
    
    # No need to build @pending_changes if there is none.
    if @pending_changes.nil? or !@pending_changes.has_key? attribute
      return
    end
    
    # Some provide an action...
    if action != :change
      # No need to process if it doesn't exist.
      if !@pending_changes[attribute].has_key? action or !@pending_changes[attribute][action].has_key? value
        return
      end
      
      # Otherwise, delete the value...
      @pending_changes[attribute][action].delete value
      
      # And now we chain deletions as needed (if nothing is left in each category).
      if @pending_changes[attribute][action].empty?
        @pending_changes[attribute].delete action
      end
      if @pending_changes[attribute].empty?
        @pending_changes.delete attribute
      end
      if @pending_changes.empty?
        @pending_changes = nil
      end
    else
      # No need to process if it doesn't exist.
      if !@pending_changes[attribute].has_key? value
        return
      end
      
      # Otherwise, delete the value.
      @pending_changes[attribute].delete value
      
      # And now we chain deletions as needed (if nothing is left in each category).
      if @pending_changes[attribute].empty?
        @pending_changes.delete attribute
      end
      if @pending_changes.empty?
        @pending_changes = nil
      end
    end
    
    # Let's finish by changing the state.
    if approved? and not @pending_changes.nil?
      self.state = 'pending_changes'
    elsif pending_changes? and @pending_changes.nil?
      self.state = 'approved'
    end
    
    # And commit the changes as YAML.
    render_pending_changes
  end
  
  # Adds support for a pending change.
  def add_pending_change_support(attribute, action, value)
    # Check to see if our pending changes have changed since we last checked.
    pending_changes unless @pending_changes_loaded
    
    # Build @pending_changes as needed.
    if @pending_changes.nil?
      @pending_changes = Hash.new {|h,k| h[k] = Hash.new {|h,k| h[k] = {}}}
    end
    
    # Some require a method other than 'change'
    if action != :change
      # Don't bother adding support if we're changing to something identical!
      if action == :add and attribute == :alternate_names and alternate_names.include?(value)
        return
      elsif action == :add and attribute == :video_game_genres and video_game_genres.include?(VideoGameGenre.find(value))
        return
      elsif action == :remove and attribute == :alternate_names and !alternate_names.include?(value)
        return
      elsif action == :remove and attribute == :video_game_genres and !video_game_genres.include?(VideoGameGenre.find(value))
        return
      end
	  
      # Construct the hash for that action as well.
      if !@pending_changes[attribute].has_key? action
        @pending_changes[attribute][action] = Hash.new {|h,k| h[k] = []}
      elsif @pending_changes[attribute][action].default_proc.nil?
        # We loaded it from somewhere and don't have the default action.
        @pending_changes[attribute][action] = (Hash.new {|h,k| h[k] = []}).merge @pending_changes[attribute][action]
      end
      
      # And add the support.
      if !@pending_changes[attribute][action][value].include? User.stamper
        @pending_changes[attribute][action][value] << User.stamper
      end
    else
      # Don't bother adding support if we're changing to something identical!
      if read_attribute(attribute) == value
        return
      end
	  
      # Otherwise, we can just create the array.
      if !@pending_changes[attribute].has_key? value
        @pending_changes[attribute][value] = []
      end
      
      # And add the support.
      if !@pending_changes[attribute][value].include? User.stamper
        @pending_changes[attribute][value] << User.stamper
      end
    end
    
    # Change the state and we're good.
    if approved? and not @pending_changes.nil?
      self.state = 'pending_changes'
    elsif pending_changes? and @pending_changes.nil?
      self.state = 'approved'
    end
    
    # Commit the changes as YAML
    render_pending_changes
  end
  
  # Removes support for a pending change.
  def remove_pending_change_support(attribute, action, value)
    pending_changes unless @pending_changes_loaded
    
    # No need to build @pending_changes if there is none.
    if @pending_changes.nil? or !@pending_changes.has_key? attribute
      return
    end
    
    # Some provide an action...
    if action != :change
      # No need to process if it doesn't exist.
      if !@pending_changes[attribute].has_key? action or !@pending_changes[attribute][action].has_key? value
        return
      end
      
      # No need to process if this user hasn't supported it either.
      if !@pending_changes[attribute][action][value].include? User.stamper
        return
      end
      
      # Otherwise, remove the user from the list of supporters.
      @pending_changes[attribute][action][value].delete User.stamper
      
      # And now we chain deletions as needed (if nothing is left in each category).
      if @pending_changes[attribute][action][value].empty?
        @pending_changes[attribute][action].delete value
      end
      if @pending_changes[attribute][action].empty?
        @pending_changes[attribute].delete action
      end
      if @pending_changes[attribute].empty?
        @pending_changes.delete attribute
      end
      if @pending_changes.empty?
        @pending_changes = nil
      end
    else
      # No need to process if it doesn't exist.
      if !@pending_changes[attribute].has_key? value
        return
      end
      
      # Or if the user never supported it in the first place.
      if !@pending_changes[attribute][value].include? User.stamper
        return
      end
      
      # Otherwise, remove the user from the list of supporters.
      @pending_changes[attribute.to_sym][value].delete User.stamper
      
      # And now we chain deletions as needed (if nothing is left in each category)
      if @pending_changes[attribute][value].empty?
        @pending_changes[attribute].delete value
      end
      if @pending_changes[attribute].empty?
        @pending_changes.delete attribute
      end
      if @pending_changes.empty?
        @pending_changes = nil
      end
    end
    
    # Let's finish by changing the state.
    if approved? and not @pending_changes.nil?
      self.state = 'pending_changes'
    elsif pending_changes? and @pending_changes.nil?
      self.state = 'approved'
    end
    
    # And commit the changes as YAML.
    render_pending_changes
  end
  
  # Returns true if the item can be approved (i.e. None of the proposed changes conflict.)
  def approvable_changes?
    errors.clear
    
    # Check each of the single-valued attributes for more than one pending change, as these conflict.
    [:name, :year, :video_game_continuum, :video_game_system].each do |key|
      if !pending_changes.nil? and !pending_changes[key].nil? and pending_changes[key].size > 1
        errors.add(key, "has more than one pending change which must be resolved prior to approving all changes.")
      end
    end
    
    # Check each of the multi-valued attributes for pending changes that add AND remove the same thing, as these conflict.
    [:alternate_names, :video_game_genres].each do |key|
      if !pending_changes.nil? and !pending_changes[key].nil? and !pending_changes[key][:add].nil? and !pending_changes[key][:remove].nil? and pending_changes[key][:add].keys.length > pending_changes[key][:remove].keys.length
        pending_changes[key][:remove].keys.each do |name|
          if pending_changes[key][:add].key? name
            errors.add(key, "has both an add and remove request that must be resolved prior to approving all changes.")
            break
          end
        end
      elsif !pending_changes.nil? and !pending_changes[key].nil? and !pending_changes[key][:add].nil? and !pending_changes[key][:remove].nil?
        pending_changes[key][:add].keys.each do |name|
          if pending_changes[key][:remove].key? name
            errors.add(key, "has both an add and remove request that must be resolved prior to approving all changes.")
            break
          end
        end
      end
    end
    
    errors.empty?
  end
  
  # Returns comments directly attached to this VideoGame, ordered from oldest
  # to newest.
  def child_comments
    Comment.find :all, :conditions => ['(parent_id IS NULL OR parent_id = 0) AND item_id = ? AND item_type = ?', self.id, self.class.name], :order => 'created_at ASC'
  end

  # Merges this VideoGame and its alternate names into another VideoGame.  This is an irreversible operation (like destroy!)
  def merge_into!(other)
    # Merge the alternate names...
    other.alternate_names << self.name
    self.alternate_names.to_ary.each {|name| other.alternate_names << name if name != self.name}
    
    # Merge the genres...
    self.video_game_genres.each {|genre| other.video_game_genres << genre}
    
    # Now, merge properties, where they're nil.
    other.year = self.year if other.year.nil?
    other.video_game_continuum = self.video_game_continuum if other.video_game_continuum.nil?
    other.video_game_system = self.video_game_system if other.video_game_system.nil?
    
    # We don't merge pending changes.
    
    # TODO: Handle merging of other items...
#    for question in self.video_game_trivia_questions
#      question.video_game = other
#      question.save!
#    end
#    for scene in self.video_game_scenes
#      scene.video_game = other
#      scene.save!
#    end

    VideoGame.delete(id) unless new_record?
    other.save!
    
    freeze
    
    true
  end
  
protected
  # Calls all on-approval functions.
  def do_approval
    do_initial_approval
    commit_changes
  end
  
  # Performs the initial approval by setting +approver+ and +approved_at+
  def do_initial_approval
    if self.state == 'pending_approval'
      unless User.stamper.nil?
        self.approver_id = User.stamper
      else
        self.approver_id = nil
      end
      self.approved_at = Time.now.utc
    end
  end
  
  # Commits any pending changes.
  def commit_changes
    # Commit outstanding changes to the single-valued attributes.
    [:name, :year].each { |key| write_attribute(key, pending_changes[key].keys[0]) unless pending_changes.nil? or pending_changes[key].nil? or pending_changes[key].empty? }
    self.video_game_continuum = VideoGameContinuum.find(pending_changes[:video_game_continuum].keys[0]) unless pending_changes.nil? or pending_changes[:video_game_continuum].nil? or pending_changes[:video_game_continuum].empty?
    self.video_game_system = VideoGameSystem.find(pending_changes[:video_game_system].keys[0]) unless pending_changes.nil? or pending_changes[:video_game_system].nil? or pending_changes[:video_game_system].empty?
    
    # And then to the multi-valued attributes.
    [:alternate_names, :video_game_genres].each do |key|
      if !pending_changes.nil? and !pending_changes[key].nil?
        pending_changes[key].keys.each do |method|
          if method == :add
            if key == :alternate_names
              pending_changes[key][method].keys.each do |name|
                alternate_names << name
              end
            elsif key == :video_game_genres
              pending_changes[key][method].keys.each do |id|
                if !self.video_game_genres.include? VideoGameGenre.find(id)
                  self.video_game_genres << VideoGameGenre.find(id)
                end
              end
            end
          elsif method == :remove
            if key == :alternate_names
              pending_changes[key][method].keys.each do |name|
                alternate_names.delete(name)
              end
            elsif key == :video_game_genres
              pending_changes[key][method].keys.each do |id|
                video_game_genres.delete(VideoGameGenre.find(id))
              end
            end
          end
        end
      end
    end
    
    # Set remaining changes to nil.
    self.pending_changes = nil
  end
  
  # Renders the pending changes hash to YAML for storage in the database.
  def render_pending_changes
    pending_changes unless @pending_changes_loaded
    if @pending_changes.nil?
      write_attribute :pending_changes, nil
      
      # Auto-approve if there are no pending changes.
      if pending_changes?
        self.state = 'approved'
      end
    else
      write_attribute :pending_changes, @pending_changes.to_yaml
    end
  end
end
