require File.dirname(__FILE__) + '/../test_helper'
require 'comments_controller'

# Reraise errors caught by the controller.
class CommentsController; def rescue_action(e) raise e end; end

class CommentsControllerTest < ActionController::TestCase
  fixtures :roles
  fixtures :users
  fixtures :delegators_roles
  fixtures :roles_users
  fixtures :roles_encompassed_roles
  
  # Anyone should be able to add a new comment.
  def test_new_comment
    login_as :quentin
    
    # Try POSTing to new with a 0 id.
    post :new, :id => '0', :item_type => 'AnimeSeries', :item_id => anime_series(:kamichu).id
    assert_response :success
    assert_template 'comment_box.js.rjs'
    
    assert_equal '0', assigns(:parent_id)
    assert_equal 'AnimeSeries', assigns(:item_type)
    assert_equal anime_series(:kamichu).id.to_s, assigns(:item_id)
    
    # And with a non-0 id, as well.
    post :new, :id => comments(:brother_comment).id
    assert_response :success
    assert_template 'comment_box.js.rjs'
    
    assert_equal comments(:brother_comment).id.to_s, assigns(:parent_id)
    assert_equal comments(:brother_comment).item_type, assigns(:item_type)
    assert_equal comments(:brother_comment).item_id, assigns(:item_id)
  end
  
  # Anyone should be able to get the list of comment types for a given type.
  def test_get_type_description
    login_as :quentin
    
    # Try POSTing to new with a 0 id.
    post :get_type_description, :type => 'general_comment'
    assert_response :success
    assert_template 'comment_type_description.js.rjs'
    
    assert_equal :general_comment, assigns(:type)
  end
  
  # Anyone should be able to CREATE a comment.
  def test_create_comment
    login_as :quentin
    
    assert_difference 'Comment.count' do
      post :create, :comment_parent_id => '0', :comment_type => 'general_comment', :comment_message => "Here's a comment", :comment_item_type => 'AnimeSeries', :comment_item_id => 2
      assert_redirected_to '/'
    end
    
    # Try with a non-0 id.
    assert_difference 'Comment.count' do
      post :create, :comment_parent_id => comments(:brother_comment).id, :comment_type => 'general_comment', :comment_message => "Here's a comment"
      assert_redirected_to '/'
    end
  end
  
  # Only admins can delete.
  def test_delete_comment_normal_user
    login_as :quentin
    
    assert_no_difference 'Comment.count' do
      put :delete, :id => comments(:brother_comment).id
      assert_response 401
    end
  end
  
  def test_delete_comment_admin
    login_as :admin
    
    assert_difference 'Comment.count', -1 do
      put :delete, :id => comments(:brother_comment).id
      assert_redirected_to '/'
    end
  end
  
  # Only mods can 'make a note'.
  def test_make_note_comment_normal_user
    login_as :quentin
    
    assert_not_equal :note, Comment.find(comments(:brother_comment).id).comment_type
    
    put :make_note, :id => comments(:brother_comment).id
    assert_response 401
    
    assert_not_equal :note, Comment.find(comments(:brother_comment).id).comment_type
  end
  
  def test_make_note_comment_approver
    login_as :quentin
    
    users(:quentin).roles << roles(:anime_series_approver)
    users(:quentin).save
    
    assert_not_equal :note, Comment.find(comments(:brother_comment).id).comment_type
    
    put :make_note, :id => comments(:brother_comment).id
    assert_response 401
    
    assert_not_equal :note, Comment.find(comments(:brother_comment).id).comment_type
  end
  
  def test_make_note_comment_moderator
    login_as :quentin
    
    users(:quentin).roles << roles(:anime_series_moderator)
    users(:quentin).save
    
    assert_not_equal :note, Comment.find(comments(:brother_comment).id).comment_type
    
    put :make_note, :id => comments(:brother_comment).id
    assert_redirected_to '/'
    
    assert_equal :note, Comment.find(comments(:brother_comment).id).comment_type
  end
  
  # Only admins should edit or update comment type
  def test_edit_comment_type_normal_user
    login_as :quentin
    
    post :edit_comment_type, :id => comments(:brother_comment).id
    assert_response 401
  end
  
  def test_edit_comment_type_moderator
    login_as :quentin
    
    users(:quentin).roles << roles(:anime_series_moderator)
    users(:quentin).save
    
    post :edit_comment_type, :id => comments(:brother_comment).id
    assert_response 401
  end
  
  def test_edit_comment_type_admin
    login_as :admin
    
    post :edit_comment_type, :id => comments(:brother_comment).id
    assert_response :success
    assert_template 'comment_type_editor.js.rjs'
    
    assert_equal comments(:brother_comment), assigns(:comment)
  end
  
  def test_update_comment_type_normal_user
    login_as :quentin
    
    assert_not_equal :request_handled, Comment.find(comments(:brother_comment).id).comment_type
    
    post :update_comment_type, :id => comments(:brother_comment).id, "comment#{comments(:brother_comment).id}_type" => 'request_handled'
    assert_response 401
    
    assert_not_equal :request_handled, Comment.find(comments(:brother_comment).id).comment_type
  end
  
  def test_update_comment_type_moderator
    login_as :quentin
    
    users(:quentin).roles << roles(:anime_series_moderator)
    users(:quentin).save
    
    assert_not_equal :request_handled, Comment.find(comments(:brother_comment).id).comment_type
    
    post :update_comment_type, :id => comments(:brother_comment).id, "comment#{comments(:brother_comment).id}_type" => 'request_handled'
    assert_response 401
    
    assert_not_equal :request_handled, Comment.find(comments(:brother_comment).id).comment_type
  end
  
  def test_update_comment_type_admin
    login_as :admin
    
    assert_not_equal :request_handled, Comment.find(comments(:brother_comment).id).comment_type
    
    post :update_comment_type, :id => comments(:brother_comment).id, "comment#{comments(:brother_comment).id}_type" => 'request_handled'
    assert_redirected_to '/'
    
    assert_equal :request_handled, Comment.find(comments(:brother_comment).id).comment_type
  end
end