require File.dirname(__FILE__) + '/../test_helper'
require 'users_controller'

# Re-raise errors caught by the controller.
class UsersController; def rescue_action(e) raise e end; end

class UsersControllerTest < Test::Unit::TestCase
  fixtures :users
  fixtures :roles_users
  fixtures :delegators_roles
  fixtures :roles

  def setup
    @controller = UsersController.new
    @request    = ActionController::TestRequest.new
    @response   = ActionController::TestResponse.new
  end

  def test_should_allow_signup
    assert_difference 'User.count' do
      create_user
      assert_response :redirect
    end
  end

  def test_should_require_login_on_signup
    assert_no_difference 'User.count' do
      create_user(:login => nil)
      assert assigns(:user).errors.on(:login)
      assert_response :success
    end
  end

  def test_should_require_password_on_signup
    assert_no_difference 'User.count' do
      create_user(:password => nil)
      assert assigns(:user).errors.on(:password)
      assert_response :success
    end
  end

  def test_should_require_password_confirmation_on_signup
    assert_no_difference 'User.count' do
      create_user(:password_confirmation => nil)
      assert assigns(:user).errors.on(:password_confirmation)
      assert_response :success
    end
  end

  def test_should_require_email_on_signup
    assert_no_difference 'User.count' do
      create_user(:email => nil)
      assert assigns(:user).errors.on(:email)
      assert_response :success
    end
  end
  
  def test_should_sign_up_user_in_pending_state
    create_user
    assigns(:user).reload
    assert assigns(:user).pending?
  end

  
  def test_should_sign_up_user_with_activation_code
    create_user
    assigns(:user).reload
    assert_not_nil assigns(:user).activation_code
  end

  def test_should_activate_user
    assert_nil User.authenticate('aaron', 'test')
    get :activate, :activation_code => users(:aaron).activation_code
    assert_redirected_to '/session/new'
    assert_not_nil flash[:notice]
    assert_equal users(:aaron), User.authenticate('aaron', 'monkey')
  end
  
  def test_should_not_activate_user_without_key
    get :activate
    assert_nil flash[:notice]
  rescue ActionController::RoutingError
    # in the event your routes deny this, we'll just bow out gracefully.
  end

  def test_should_not_activate_user_with_blank_key
    get :activate, :activation_code => ''
    assert_nil flash[:notice]
  rescue ActionController::RoutingError
    # well played, sir
  end
  
  def test_should_display_user_list_without_controls
    login_as :aaron
    
    get :list
    
    assert_response :success
    assert_template 'list'
    
    assert_not_nil assigns(:users)
    assert_equal assigns(:render_admin_actions), false
  end
  
  def test_should_display_user_list_with_controls
    login_as :admin
    
    get :list
    
    assert_response :success
    assert_template 'list'
    
    assert_not_nil assigns(:users)
    assert_equal assigns(:render_admin_actions), true
  end
  
  def test_show_action
    login_as :aaron
    
    get :show, :id => users(:aaron).id
    
    assert_response :success
    assert_template 'show'
    
    assert_not_nil assigns(:user)
    assert_equal assigns(:user), users(:aaron)
  end
  
  def test_should_edit_self
    login_as :aaron
    
    get :edit, :id => users(:aaron).id
    
    assert_response :success
    assert_template 'edit'
    
    assert_not_nil assigns(:user)
    assert_equal assigns(:user), users(:aaron)
  end
  
  def test_shouldnt_edit_others
    login_as :quentin
    
    get :edit, :id => users(:admin).id
    
    assert_redirected_to '/users'
    assert_not_nil flash[:error]
  end
  
  def test_admin_should_edit_any
    login_as :admin
    
    get :edit, :id => users(:aaron).id
    
    assert_response :success
    assert_template 'edit'
    
    assert_not_nil assigns(:user)
    assert_equal assigns(:user), users(:aaron)
  end
  
  def test_update_action
    login_as :quentin
    
    post :update, :id => 1, :user => {:login => 'quentin'}, :current_password => 'monkey'
    
    assert_redirected_to :action => :index
    assert_nil flash[:error]
  end
  
  def test_reset_password_action
    get :reset_password
    
    assert_response :success
    assert_template 'reset_password'
  end
  
  def test_should_send_reset_password_email_to_user
    ActionMailer::Base.deliveries = []
    
    post :send_reset_password_email, :email => users(:quentin).email
    
    # Should generate an activation key and e-mail it.
    assert !ActionMailer::Base.deliveries.empty?
    sent = ActionMailer::Base.deliveries.first
    assert_equal [users(:quentin).email], sent.to
    assert_not_nil User.find_by_login('quentin').activation_code
    
    assert_redirected_to '/reset_password'
    assert_not_nil flash[:notice]
  end
  
  def test_shouldnt_send_reset_password_email_to_non_user
    ActionMailer::Base.deliveries = []
    
    post :send_reset_password_email, :email => 'foo@bar.com'
    
    assert ActionMailer::Base.deliveries.empty?
    
    assert_redirected_to '/reset_password'
    assert_not_nil flash[:error]
  end
  
  def test_should_reset_password
    user = User.find_by_login('quentin')
    user.activation_code = '0123456789abcdef'
    user.save!
    
    post :update, :reset_password => true, :activation_code => user.activation_code, :password => 'abcdefg', :password_confirmation => 'abcdefg'
    
    user.activation_code = nil
    
    assert_redirected_to '/session/new'
    assert_not_nil flash[:notice]
    assert_equal user, User.authenticate('quentin', 'abcdefg')
  end
  
  def test_shouldnt_reset_just_any_new_password
    post :update, :reset_password => true, :activation_code => '', :password => 'abcd', :password_confirmation => 'abcd'
    
    assert_redirected_to '/'
    assert_not_nil flash[:error]
  end
  
  def test_add_role_if_admin
    login_as :admin
    
    post :add_role, :user_id => users(:quentin).id, :id => "role_#{roles(:user_role).id}"
    
    assert_response :success
    assert_template 'edit_role.js.rjs'
    assert_not_nil flash[:notice]
  end
  
  def test_remove_role_only_if_admin
    login_as :admin
    
    post :remove_role, :user_id => users(:quentin).id, :id => "role_#{roles(:user_role).id}"
    
    assert_response :success
    assert_template 'edit_role.js.rjs'
    assert_not_nil flash[:notice]
  end
  
  def test_remove_role_cannot_remove_admin_from_self
    login_as :admin
    
    post :remove_role, :user_id => users(:admin).id, :id => "role_#{roles(:admin_role).id}"
    
    assert_response :success
    assert_template 'edit_role.js.rjs'
    assert_not_nil flash[:error]
  end
  
  def test_add_delegator_role_only_if_admin
    login_as :admin
    
    post :add_delegator_role, :user_id => users(:quentin).id, :id => "role_delegator_#{roles(:user_role).id}"
    
    assert_response :success
    assert_template 'edit_role.js.rjs'
    assert_not_nil flash[:notice]
  end
  
  def test_remove_delegator_role_only_if_admin
    login_as :admin
    
    post :remove_delegator_role, :user_id => users(:quentin).id, :id => "role_delegator_#{roles(:user_role).id}"
    
    assert_response :success
    assert_template 'edit_role.js.rjs'
    assert_not_nil flash[:notice]
  end
  
  def test_non_admins_cannot_change_roles
    login_as :quentin
    
    post :add_role, :user_id => users(:quentin).id, :id => "role_#{roles(:admin_role).id}"
    assert_response 401
    assert !users(:quentin).has_role?('admin')
    post :remove_role, :user_id => users(:admin).id, :id => "role_#{roles(:admin_role).id}"
    assert_response 401
    assert users(:admin).has_role?('admin')
    post :add_delegator_role, :user_id => users(:quentin).id, :id => "role_#{roles(:admin_role).id}"
    assert_response 401
    assert !users(:quentin).can_delegate?('admin')
    post :remove_delegator_role, :user_id => users(:admin).id, :id => "role_#{roles(:admin_role).id}"
    assert_response 401
    assert users(:admin).delegator_roles.include?(roles(:admin_role))
  end
  
protected
  def create_user(options = {})
    post :create, :user => { :login => 'quire', :email => 'quire@example.com',
      :password => 'quire69', :password_confirmation => 'quire69' }.merge(options)
  end
end
