require File.dirname(__FILE__) + '/../test_helper'
require 'video_game_controller'

# Reraise errors caught by the controller.
class VideoGameController; def rescue_action(e) raise e end; end

class VideoGameControllerTest < ActionController::TestCase
  # Anyone can see the index.
  def test_index
    login_as :quentin
    
    get :index
    assert_response :success
    assert_template 'list'
  end
  
  # Anyone can see the list.
  def test_list
    login_as :quentin
    
    get :list
    assert_response :success
    assert_template 'list'
  end
  
  # Anyone can see show.
  def test_show
    login_as :quentin
    
    get :show, :id => video_games(:sonic_the_hedgehog).id
    assert_response :success
    assert_template 'show'
    
    # And it should have @video_game set to the requested game.
    assert_equal video_games(:sonic_the_hedgehog), assigns(:video_game)
  end
  
  # Anyone can see new.
  def test_new
    login_as :quentin
    
    get :new
    assert_response :success
    assert_template 'new'
    
    # And it should have @video_game set to a new game.
    #assert_equal VideoGame.new(), assigns(:video_game)
  end
  
  # Anyone can create.
  def test_create_normal_user
    login_as :quentin
    
    # Set @video_game before POSTing create.
    assert_difference 'VideoGame.count' do
      post :create, :video_game => create_game, :video_game_continuum => {:name => ''}, :video_game_system => {:name => 'Entertainment System'}, :genre_checkbox => {'rpg' => '1'}
      assert_redirected_to new_video_game_path
    end
    
    # It should have gone successfully.
    assert_not_nil flash[:notice]
    assert_nil flash[:error]
    assert_nil flash[:video_game]
    assert_nil flash[:submission_note]
  end
  
  def test_bad_create_normal_user
    login_as :quentin
    
    # Constrast when we have a bad POST
    assert_no_difference 'VideoGame.count' do
      post :create, :video_game => create_game(:name => video_games(:sonic_the_hedgehog).name), :video_game_continuum => {:name => ''}, :video_game_system => {:name => 'NES'}, :genre_checkbox => {'rpg' => '1'}
      assert_redirected_to new_video_game_path
    end
    
    # It should NOT have gone successfully.
    assert_nil flash[:notice]
    assert_not_nil flash[:error]
    assert flash[:error_is_html]
    assert_not_nil flash[:video_game]
    assert_nil flash[:submission_note]
  end
  
  def test_complex_create_normal_user
    login_as :quentin
    
    # And now a complex success test with comments and alternate names.
    assert_difference 'VideoGame.count' do
      assert_difference 'AlternateName.count', 3 do
        assert_difference 'Comment.count' do
          post :create, :video_game => create_game(:name => '1234', :alternate_names => {1 => 'a', 2 => 'a', 3 => '', 4 => '', 5 => nil, 6 => 'b', 7 => 'c'}), :submission_note => 'A submission note.', :video_game_continuum => {:name => ''}, :video_game_system => {:name => 'Entertainment System'}, :genre_checkbox => {'rpg' => '1'}
          assert_redirected_to new_video_game_path
        end
      end
    end
    
    # It should have gone successfully.
    assert_not_nil flash[:notice]
    assert_nil flash[:error]
    assert_nil flash[:video_game]
    assert_nil flash[:submission_note]
    
    # Should get one if we wrote the test right.
    video_game = VideoGame.find_by_name('1234')
    assert_not_nil video_game
    
    # Main users should not auto-approve...
    assert video_game.pending_approval?
    assert_equal users(:quentin), video_game.creator
    
    # It should have eliminated nils, empties, and duplicates for the alternate_names.
    assert_equal 3, video_game.alternate_names.count
    assert video_game.alternate_names.to_ary.include?('a')
    assert video_game.alternate_names.to_ary.include?('b')
    assert video_game.alternate_names.to_ary.include?('c')
    
    # And we should find the comment attached.
    assert_equal 1, video_game.comments.count
    assert_equal 'A submission note.', video_game.comments[0].text
    assert_equal :submission_note, video_game.comments[0].comment_type
  end
  
  def test_complex_bad_create_normal_user
    login_as :quentin
    
    # And lastly a complex failure test with comments and alternate names.
    assert_no_difference 'VideoGame.count' do
      assert_no_difference 'AlternateName.count' do
        assert_no_difference 'Comment.count' do
          post :create, :video_game => create_game(:name => nil, :alternate_names => {1 => 'a', 2 => 'a', 3 => '', 4 => '', 5 => nil, 6 => 'b', 7 => 'c'}), :submission_note => 'A submission note.', :video_game_continuum => {:name => ''}, :video_game_system => {:name => ''}, :genre_checkbox => {'rpg' => '1'}
          assert_redirected_to new_video_game_path
        end
      end
    end
    
    # It should have gone successfully.
    assert_nil flash[:notice]
    assert_not_nil flash[:error]
    assert flash[:error_is_html]
    assert_not_nil flash[:video_game]
    assert_not_nil flash[:submission_note]
  end
  
  def test_create_approver_user
    login_as :quentin
    
    # Oh right.  Approvers and above should have their changes auto-approve.
    user = users(:quentin)
    user.roles << roles(:video_game_approver)
    user.save
    
    assert_difference 'VideoGame.count' do
      post :create, :video_game => create_game(:name => 'Dragon Quest'), :video_game_continuum => {:name => 'Dragon Quest'}, :video_game_system => {:name => 'Entertainment System'}, :genre_checkbox => {'rpg' => '1'}
      assert_redirected_to new_video_game_path
    end
    
    # Should get one if we wrote the test right.
    video_game = VideoGame.find_by_name('Dragon Quest')
    assert_not_nil video_game
    
    assert video_game.approved?
    assert_equal users(:quentin), video_game.creator
    assert_equal users(:quentin), video_game.approver
  end
  
  # TODO: test_create_with_continuum
  
  # Approvers and above can see edit.
  def test_edit_normal_user
    login_as :quentin
    
    get :edit, :id => video_games(:sonic_the_hedgehog).id
    assert_response 401
  end
  
  def test_edit_approver
    login_as :quentin
    
    user = users(:quentin)
    user.roles << roles(:video_game_approver)
    user.save
    
    get :edit, :id => video_games(:sonic_the_hedgehog).id
    assert_response :success
    assert_template 'edit'
    assert_equal video_games(:sonic_the_hedgehog), assigns(:video_game)
  end
  
  def test_edit_moderator
    login_as :quentin
    
    user = users(:quentin)
    user.roles << roles(:video_game_moderator)
    user.save
    
    get :edit, :id => video_games(:sonic_the_hedgehog).id
    assert_response :success
    assert_template 'edit'
    assert_equal video_games(:sonic_the_hedgehog), assigns(:video_game)
  end
  
  def test_edit_super_moderator
    login_as :quentin
    
    user = users(:quentin)
    user.roles << roles(:video_game_super_moderator)
    user.save
    
    get :edit, :id => video_games(:sonic_the_hedgehog).id
    assert_response :success
    assert_template 'edit'
    assert_equal video_games(:sonic_the_hedgehog), assigns(:video_game)
  end
  
  # Approvers can approve.
  def test_approve_normal_user
    login_as :quentin
    
    put :approve, :id => video_games(:okami).id
    assert_response 401
#    assert_redirected_to new_video_game_path
  end
  
  def test_approve_approver
    login_as :quentin
    
    user = users(:quentin)
    user.roles << roles(:video_game_approver)
    user.save
    
    put :approve, :id => video_games(:okami).id
    assert_redirected_to edit_video_game_path(video_games(:okami).id)
    assert_not_nil flash[:notice]
    assert VideoGame.find(video_games(:okami).id).approved?
  end
  
  def test_approve_moderator
    login_as :quentin
    
    user = users(:quentin)
    user.roles << roles(:video_game_moderator)
    user.save
    
    put :approve, :id => video_games(:okami).id
    assert_redirected_to edit_video_game_path(video_games(:okami).id)
    assert_not_nil flash[:notice]
    assert VideoGame.find(video_games(:okami).id).approved?
  end
  
  def test_approve_super_moderator
    login_as :quentin
    
    user = users(:quentin)
    user.roles << roles(:video_game_super_moderator)
    user.save
    
    put :approve, :id => video_games(:okami).id
    assert_redirected_to edit_video_game_path(video_games(:okami).id)
    assert_not_nil flash[:notice]
    assert VideoGame.find(video_games(:okami).id).approved?
    
    # Okay, but Sonic the Hedgehog SHOULDN'T approve until it's clean.
    put :approve, :id => video_games(:sonic_the_hedgehog).id
    assert_redirected_to edit_video_game_path(video_games(:sonic_the_hedgehog).id)
    assert_not_nil flash[:error]
    assert VideoGame.find(video_games(:sonic_the_hedgehog).id).pending_changes?
    
    game = video_games(:sonic_the_hedgehog)
    game.delete_pending_change :year, :change, "2006"
    game.save
    
    put :approve, :id => video_games(:sonic_the_hedgehog).id
    assert_redirected_to edit_video_game_path(video_games(:sonic_the_hedgehog).id)
    assert_not_nil flash[:notice]
    assert VideoGame.find(video_games(:sonic_the_hedgehog).id).approved?
  end
  
  # Approvers can approve.
  def test_approve_pending_change_normal_user
    login_as :quentin
    
    put :approve_pending_change, :id => video_games(:sonic_the_hedgehog).id, :change_action => 'change', :key => :name, :value => 'Sonic the Hedgehog'
    assert_response 401
#    assert_redirected_to new_video_game_path
  end
  
  def test_approve_pending_change_approver
    login_as :quentin
    
    user = users(:quentin)
    user.roles << roles(:video_game_approver)
    user.save
    
    put :approve_pending_change, :id => video_games(:sonic_the_hedgehog).id, :change_action => 'change', :key => :name, :value => 'Sonic the Hedgehog'
    assert_redirected_to edit_video_game_path(video_games(:sonic_the_hedgehog).id)
    assert_not_nil flash[:notice]
    assert_equal 'Sornic the hedgehog?', VideoGame.find(video_games(:sonic_the_hedgehog).id).name
    assert VideoGame.find(video_games(:sonic_the_hedgehog).id).pending_changes[:name]['Sonic the Hedgehog'].include?(user.id)
  end
  
  def test_approve_pending_change_moderator
    login_as :quentin
    
    user = users(:quentin)
    user.roles << roles(:video_game_moderator)
    user.save
    
    put :approve_pending_change, :id => video_games(:sonic_the_hedgehog).id, :change_action => 'change', :key => :name, :value => 'Sonic the Hedgehog'
    assert_redirected_to edit_video_game_path(video_games(:sonic_the_hedgehog).id)
    assert_not_nil flash[:notice]
    assert_equal 'Sonic the Hedgehog', VideoGame.find(video_games(:sonic_the_hedgehog).id).name
    assert !VideoGame.find(video_games(:sonic_the_hedgehog).id).pending_changes.include?(:name)
  end
  
  def test_approve_pending_change_super_moderator
    login_as :quentin
    
    user = users(:quentin)
    user.roles << roles(:video_game_super_moderator)
    user.save
    
    put :approve_pending_change, :id => video_games(:sonic_the_hedgehog).id, :change_action => 'change', :key => :name, :value => 'Sonic the Hedgehog'
    assert_redirected_to edit_video_game_path(video_games(:sonic_the_hedgehog).id)
    assert_not_nil flash[:notice]
    assert_equal 'Sonic the Hedgehog', VideoGame.find(video_games(:sonic_the_hedgehog).id).name
    assert !VideoGame.find(video_games(:sonic_the_hedgehog).id).pending_changes.include?(:name)
  end
  
  # Approvers can reject.
  def test_reject_pending_change_normal_user
    login_as :quentin
    
    put :reject_pending_change, :id => video_games(:sonic_the_hedgehog).id, :change_action => 'change', :key => :year, :value => '2006'
    assert_response 401
#    assert_redirected_to new_video_game_path
  end
  
  def test_reject_pending_change_approver
    login_as :quentin
    
    user = users(:quentin)
    user.roles << roles(:video_game_approver)
    user.save
    
    put :reject_pending_change, :id => video_games(:sonic_the_hedgehog).id, :change_action => 'change', :key => :year, :value => '2006'
    assert_redirected_to edit_video_game_path(video_games(:sonic_the_hedgehog).id)
    assert_not_nil flash[:notice]
    assert_equal 1991, VideoGame.find(video_games(:sonic_the_hedgehog).id).year
    assert !VideoGame.find(video_games(:sonic_the_hedgehog).id).pending_changes[:year].include?('2006')
    
    put :reject_pending_change, :id => video_games(:sonic_the_hedgehog).id, :change_action => 'change', :key => :year, :value => '2005'
    assert_redirected_to edit_video_game_path(video_games(:sonic_the_hedgehog).id)
    assert_not_nil flash[:notice]
    assert_equal 1991, VideoGame.find(video_games(:sonic_the_hedgehog).id).year
    assert VideoGame.find(video_games(:sonic_the_hedgehog).id).pending_changes[:year].include?('2005')
  end
  
  def test_reject_pending_change_moderator
    login_as :quentin
    
    user = users(:quentin)
    user.roles << roles(:video_game_moderator)
    user.save
    
    put :reject_pending_change, :id => video_games(:sonic_the_hedgehog).id, :change_action => 'change', :key => :year, :value => '2005'
    assert_redirected_to edit_video_game_path(video_games(:sonic_the_hedgehog).id)
    assert_not_nil flash[:notice]
    assert_equal 1991, VideoGame.find(video_games(:sonic_the_hedgehog).id).year
    assert !VideoGame.find(video_games(:sonic_the_hedgehog).id).pending_changes[:year].include?('2005')
  end
  
  def test_reject_pending_change_super_moderator
    login_as :quentin
    
    user = users(:quentin)
    user.roles << roles(:video_game_super_moderator)
    user.save
    
    put :reject_pending_change, :id => video_games(:sonic_the_hedgehog).id, :change_action => 'change', :key => :year, :value => '2005'
    assert_redirected_to edit_video_game_path(video_games(:sonic_the_hedgehog).id)
    assert_not_nil flash[:notice]
    assert_equal 1991, VideoGame.find(video_games(:sonic_the_hedgehog).id).year
    assert !VideoGame.find(video_games(:sonic_the_hedgehog).id).pending_changes[:year].include?('2005')
  end
  
  # Only moderators can add an alternate name.
  def test_add_alternate_name_normal_user
    login_as :quentin
    
    # TODO: reimplement me
    assert_no_difference 'AlternateName.count' do
      post :add_alternate_name, :id => video_games(:sonic_the_hedgehog).id, :new_alternate_name => 'Sonic 1'
      assert_response 401
    end
    
    game = VideoGame.find(video_games(:sonic_the_hedgehog).id)
    assert !game.alternate_names.include?('Sonic 1')
  end
  
  def test_add_alternate_name_approver
    login_as :quentin
    
    user = users(:quentin)
    user.roles << roles(:video_game_approver)
    user.save
    
    game = VideoGame.find(video_games(:sonic_the_hedgehog).id)
    assert game.pending_changes?
    
    assert_no_difference 'AlternateName.count' do
      put :add_alternate_name, :id => video_games(:sonic_the_hedgehog).id, :new_alternate_name => 'Sonic 1'
      assert_redirected_to edit_video_game_path(video_games(:sonic_the_hedgehog))
    end
    
    # However, approvers should have the name added to pending_changes.
    game = VideoGame.find(video_games(:sonic_the_hedgehog).id)
    assert !game.alternate_names.include?('Sonic 1')
    assert game.pending_changes?
    assert_not_nil game.pending_changes
    assert_not_nil game.pending_changes[:alternate_names]
    assert_not_nil game.pending_changes[:alternate_names][:add]
    assert_not_nil game.pending_changes[:alternate_names][:add]['Sonic 1']
    assert game.pending_changes[:alternate_names][:add]['Sonic 1'].include?(users(:quentin).id)
  end
  
  def test_add_alternate_name_moderator
    login_as :quentin
    
    # And as for moderators, it should stick immediately.
    user = users(:quentin)
    user.roles << roles(:video_game_moderator)
    user.save
    
    assert_difference 'AlternateName.count' do
      put :add_alternate_name, :id => video_games(:sonic_the_hedgehog).id, :new_alternate_name => 'Sonic 1'
      assert_redirected_to edit_video_game_path(video_games(:sonic_the_hedgehog))
    end
    
    game = VideoGame.find(video_games(:sonic_the_hedgehog).id)
    assert game.alternate_names.include?('Sonic 1')
#    assert_nil game.pending_changes[:alternate_names][:add]['My Little Goddess']
  end
  
  # Only moderators can delete an alternate name.
  def test_remove_alternate_name_normal_user
    login_as :quentin
    
    # TODO: reimplement me
    assert_no_difference 'AlternateName.count' do
      put :remove_alternate_name, :id => video_games(:super_mario_brothers).id, :name => 'SMB 1'
      assert_response 401
    end
    
    game = VideoGame.find(video_games(:super_mario_brothers).id)
    assert game.alternate_names.include?('SMB 1')
  end
    
  def test_remove_alternate_name_approver
    login_as :quentin
    
    user = users(:quentin)
    user.roles << roles(:video_game_approver)
    user.save
    
    game = VideoGame.find(video_games(:super_mario_brothers).id)
    
    assert_no_difference 'AlternateName.count' do
      put :remove_alternate_name, :id => video_games(:super_mario_brothers).id, :name => 'SMB 1'
      assert_redirected_to edit_video_game_path(video_games(:super_mario_brothers))
    end
    
    # However, approvers should have the name added to pending_changes.
    game = VideoGame.find(video_games(:super_mario_brothers).id)
    assert game.alternate_names.include?('SMB 1')
    assert game.pending_changes?
    assert_not_nil game.pending_changes
    assert_not_nil game.pending_changes[:alternate_names]
    assert_not_nil game.pending_changes[:alternate_names][:remove]
    assert_not_nil game.pending_changes[:alternate_names][:remove]['SMB 1']
    assert game.pending_changes[:alternate_names][:remove]['SMB 1'].include?(users(:quentin).id)
  end
  
  def test_remove_nonexistent_alternate_name_approver
    login_as :quentin
    
    user = users(:quentin)
    user.roles << roles(:video_game_approver)
    user.save
    
    # Deleting a non-existent name should give an error.
    game = VideoGame.find(video_games(:super_mario_brothers).id)
    
    assert_no_difference 'AlternateName.count' do
      put :remove_alternate_name, :id => video_games(:super_mario_brothers).id, :name => 'SMB1'
      assert_redirected_to edit_video_game_path(video_games(:super_mario_brothers))
    end
    
    game = VideoGame.find(video_games(:super_mario_brothers).id)
    assert !game.pending_changes?
    assert_nil game.pending_changes
  end
  
  def test_remove_alternate_name_moderator
    login_as :quentin
    
    # And as for moderators, it should stick immediately.
    user = users(:quentin)
    user.roles << roles(:video_game_moderator)
    user.save
    
    assert_difference 'AlternateName.count', -1 do
      put :remove_alternate_name, :id => video_games(:super_mario_brothers).id, :name => 'SMB 1'
      assert_redirected_to edit_video_game_path(video_games(:super_mario_brothers))
    end
    
    game = VideoGame.find(video_games(:super_mario_brothers).id)
    assert !game.alternate_names.include?('SMB 1')
    assert !game.pending_changes?
    assert_nil game.pending_changes
    
    # Deleting a non-existent name should give an error.
    game = VideoGame.find(video_games(:super_mario_brothers).id)
    
    assert_no_difference 'AlternateName.count' do
      put :remove_alternate_name, :id => video_games(:super_mario_brothers).id, :name => 'SMB1'
      assert_redirected_to edit_video_game_path(video_games(:super_mario_brothers))
    end
    
    game = VideoGame.find(video_games(:super_mario_brothers).id)
    assert !game.pending_changes?
    assert_nil game.pending_changes
  end
  
  # Only moderators can add an video game genre.
  def test_add_video_game_genre_normal_user
    login_as :quentin
    
    # TODO: reimplement me
    assert_no_difference 'VideoGame.find(video_games(:sonic_the_hedgehog).id).video_game_genres.count' do
      post :add_video_game_genre, :id => video_games(:sonic_the_hedgehog).id, :new_video_game_genre => 5
      assert_response 401
    end
    
    game = VideoGame.find(video_games(:sonic_the_hedgehog).id)
    assert !game.video_game_genres.include?(VideoGameGenre.find(5))
  end
  
  def test_add_video_game_genre_approver
    login_as :quentin
    
    user = users(:quentin)
    user.roles << roles(:video_game_approver)
    user.save
    
    game = VideoGame.find(video_games(:sonic_the_hedgehog).id)
    assert game.pending_changes?
    
    assert_no_difference 'VideoGame.find(video_games(:sonic_the_hedgehog).id).video_game_genres.count' do
      post :add_video_game_genre, :id => video_games(:sonic_the_hedgehog).id, :new_video_game_genre => 5
      assert_redirected_to edit_video_game_path(video_games(:sonic_the_hedgehog))
    end
    
    # However, approvers should have the name added to pending_changes.
    game = VideoGame.find(video_games(:sonic_the_hedgehog).id)
    assert !game.video_game_genres.include?(VideoGameGenre.find(5))
    assert game.pending_changes?
    assert_not_nil game.pending_changes
    assert_not_nil game.pending_changes[:video_game_genres]
    assert_not_nil game.pending_changes[:video_game_genres][:add]
    assert_not_nil game.pending_changes[:video_game_genres][:add]['5']
    assert game.pending_changes[:video_game_genres][:add]['5'].include?(users(:quentin).id)
  end
  
  def test_add_video_game_genre_moderator
    login_as :quentin
    
    # And as for moderators, it should stick immediately.
    user = users(:quentin)
    user.roles << roles(:video_game_moderator)
    user.save
    
    assert_difference 'VideoGame.find(video_games(:sonic_the_hedgehog).id).video_game_genres.count' do
      post :add_video_game_genre, :id => video_games(:sonic_the_hedgehog).id, :new_video_game_genre => 5
      assert_redirected_to edit_video_game_path(video_games(:sonic_the_hedgehog))
    end
    
    game = VideoGame.find(video_games(:sonic_the_hedgehog).id)
    assert game.video_game_genres.include?(VideoGameGenre.find(5))
#    assert_nil game.pending_changes[:video_game_genres][:add]['My Little Goddess']
  end
  
  # Only moderators can delete an video game genre.
  def test_remove_video_game_genre_normal_user
    login_as :quentin
    
    # TODO: reimplement me
    assert_no_difference 'VideoGame.find(video_games(:super_mario_brothers).id).video_game_genres.count' do
      put :remove_video_game_genre, :id => video_games(:super_mario_brothers).id, :genre_id => 3
      assert_response 401
    end
    
    game = VideoGame.find(video_games(:super_mario_brothers).id)
    assert game.video_game_genres.include?(VideoGameGenre.find(3))
  end
    
  def test_remove_video_game_genre_approver
    login_as :quentin
    
    user = users(:quentin)
    user.roles << roles(:video_game_approver)
    user.save
    
    game = VideoGame.find(video_games(:super_mario_brothers).id)
    
    assert_no_difference 'VideoGame.find(video_games(:super_mario_brothers).id).video_game_genres.count' do
      put :remove_video_game_genre, :id => video_games(:super_mario_brothers).id, :genre_id => 3
      assert_redirected_to edit_video_game_path(video_games(:super_mario_brothers))
    end
    
    # However, approvers should have the name added to pending_changes.
    game = VideoGame.find(video_games(:super_mario_brothers).id)
    assert game.video_game_genres.include?(VideoGameGenre.find(3))
    assert game.pending_changes?
    assert_not_nil game.pending_changes
    assert_not_nil game.pending_changes[:video_game_genres]
    assert_not_nil game.pending_changes[:video_game_genres][:remove]
    assert_not_nil game.pending_changes[:video_game_genres][:remove]['3']
    assert game.pending_changes[:video_game_genres][:remove]['3'].include?(users(:quentin).id)
  end
  
  def test_remove_nonexistent_video_game_genre_approver
    login_as :quentin
    
    user = users(:quentin)
    user.roles << roles(:video_game_approver)
    user.save
    
    # Deleting a non-existent name should give an error.
    game = VideoGame.find(video_games(:super_mario_brothers).id)
    
    assert_no_difference 'VideoGame.find(video_games(:super_mario_brothers).id).video_game_genres.count' do
      put :remove_video_game_genre, :id => video_games(:super_mario_brothers).id, :genre_id => 5
      assert_redirected_to edit_video_game_path(video_games(:super_mario_brothers))
    end
    
    game = VideoGame.find(video_games(:super_mario_brothers).id)
    assert !game.pending_changes?
    assert_nil game.pending_changes
  end
  
  def test_remove_video_game_genre_moderator
    login_as :quentin
    
    # And as for moderators, it should stick immediately.
    user = users(:quentin)
    user.roles << roles(:video_game_moderator)
    user.save
    
    assert_difference 'VideoGame.find(video_games(:super_mario_brothers).id).video_game_genres.count', -1 do
      put :remove_video_game_genre, :id => video_games(:super_mario_brothers).id, :genre_id => 3
      assert_redirected_to edit_video_game_path(video_games(:super_mario_brothers))
    end
    
    game = VideoGame.find(video_games(:super_mario_brothers).id)
    assert !game.video_game_genres.include?(VideoGameGenre.find(3))
    assert !game.pending_changes?
    assert_nil game.pending_changes
    
    # Deleting a non-existent name should give an error.
    game = VideoGame.find(video_games(:super_mario_brothers).id)
    
    assert_no_difference 'VideoGame.find(video_games(:super_mario_brothers).id).video_game_genres.count' do
      put :remove_video_game_genre, :id => video_games(:super_mario_brothers).id, :genre_id => 5
      assert_redirected_to edit_video_game_path(video_games(:super_mario_brothers))
    end
    
    game = VideoGame.find(video_games(:super_mario_brothers).id)
    assert !game.pending_changes?
    assert_nil game.pending_changes
  end
  
  # Moderators can destroy.
  def test_destroy_normal_user
    login_as :quentin
    
    delete :destroy, :id => video_games(:sonic_the_hedgehog).id
    assert_response 401
    assert_nothing_raised { VideoGame.find(video_games(:sonic_the_hedgehog).id) }
  end
  
  def test_destroy_approver
    login_as :quentin
    
    user = users(:quentin)
    user.roles << roles(:video_game_approver)
    user.save
    
    delete :destroy, :id => video_games(:sonic_the_hedgehog).id
    assert_response 401
    assert_nothing_raised { VideoGame.find(video_games(:sonic_the_hedgehog).id) }
  end
  
  def test_destroy_moderator
    login_as :quentin
    
    user = users(:quentin)
    user.roles << roles(:video_game_moderator)
    user.save
    
    delete :destroy, :id => video_games(:sonic_the_hedgehog).id
    assert_redirected_to manage_video_game_path
    assert_raise ActiveRecord::RecordNotFound do; VideoGame.find(video_games(:sonic_the_hedgehog).id); end
  end
  
  def test_destroy_super_moderator
    login_as :quentin
    
    user = users(:quentin)
    user.roles << roles(:video_game_super_moderator)
    user.save
    
    delete :destroy, :id => video_games(:super_mario_brothers).id
    assert_redirected_to manage_video_game_path
    assert_raise ActiveRecord::RecordNotFound do; VideoGame.find(video_games(:super_mario_brothers).id); end
  end
  
  # Approvers can manage.
  def test_manage_normal_user
    login_as :quentin
    
    get :manage
    assert_response 401
  end
  
  def test_manage_approver
    login_as :quentin
    
    user = users(:quentin)
    user.roles << roles(:video_game_approver)
    user.save
    
    get :manage
    assert_response :success
    assert_template 'manage'
  end
  
  def test_manage_moderator
    login_as :quentin
    
    user = users(:quentin)
    user.roles << roles(:video_game_moderator)
    user.save
    
    get :manage
    assert_response :success
    assert_template 'manage'
  end
  
  def test_manage_super_moderator
    login_as :quentin
    
    user = users(:quentin)
    user.roles << roles(:video_game_super_moderator)
    user.save
    
    get :manage
    assert_response :success
    assert_template 'manage'
  end

  def test_select_merge_normal_user
    login_as :quentin
    
    get :select_merge, :id => video_games(:sonik_der_hedgehog).id
    assert_response 401
  end
  
  def test_select_merge_approver
    login_as :quentin
    
    user = users(:quentin)
    user.roles << roles(:video_game_approver)
    user.save
    
    get :select_merge, :id => video_games(:sonik_der_hedgehog).id
    assert_response 401
  end
  
  def test_select_merge_moderator
    login_as :quentin
    
    user = users(:quentin)
    user.roles << roles(:video_game_moderator)
    user.save
    
    get :select_merge, :id => video_games(:sonik_der_hedgehog).id
    assert_response :success
    assert_template 'select_merge'
    
    # And it should have @video_game set to the requested game.
    assert_equal video_games(:sonik_der_hedgehog), assigns(:video_game)
  end
  
  def test_select_merge_super_moderator
    login_as :quentin
    
    user = users(:quentin)
    user.roles << roles(:video_game_super_moderator)
    user.save
    
    get :select_merge, :id => video_games(:sonik_der_hedgehog).id
    assert_response :success
    assert_template 'select_merge'
    
    # And it should have @video_game set to the requested game.
    assert_equal video_games(:sonik_der_hedgehog), assigns(:video_game)
  end
  
  # TODO: search_select_merge
  
  def test_merge_into_normal_user
    login_as :quentin
    
    put :merge_into, :id => video_games(:sonik_der_hedgehog).id, :dest => video_games(:sonic_the_hedgehog).id
    assert_response 401
    assert_nothing_raised { VideoGame.find(video_games(:sonik_der_hedgehog).id) }
    assert !VideoGame.find(video_games(:sonic_the_hedgehog).id).alternate_names.include?('Sonic der Hedgehog')
  end
  
  def test_merge_into_approver
    login_as :quentin
    
    user = users(:quentin)
    user.roles << roles(:video_game_approver)
    user.save
    
    put :merge_into, :id => video_games(:sonik_der_hedgehog).id, :dest => video_games(:sonic_the_hedgehog).id
    assert_response 401
    assert_nothing_raised { VideoGame.find(video_games(:sonik_der_hedgehog).id) }
    assert !VideoGame.find(video_games(:sonic_the_hedgehog).id).alternate_names.include?('Sonic der Hedgehog')
  end
  
  def test_merge_into_moderator
    login_as :quentin
    
    user = users(:quentin)
    user.roles << roles(:video_game_moderator)
    user.save
    
    put :merge_into, :id => video_games(:sonik_der_hedgehog).id, :dest => video_games(:sonic_the_hedgehog).id
    assert_redirected_to manage_video_game_path
	
    assert_raise ActiveRecord::RecordNotFound do; VideoGame.find(video_games(:sonik_der_hedgehog).id); end
    assert VideoGame.find(video_games(:sonic_the_hedgehog).id).alternate_names.include?('Sonic der Hedgehog')
  end
  
  def test_merge_into_super_moderator
    login_as :quentin
    
    user = users(:quentin)
    user.roles << roles(:video_game_super_moderator)
    user.save
    
    put :merge_into, :id => video_games(:sonik_der_hedgehog).id, :dest => video_games(:sonic_the_hedgehog).id
    assert_redirected_to manage_video_game_path
	
    assert_raise ActiveRecord::RecordNotFound do; VideoGame.find(video_games(:sonik_der_hedgehog).id); end
    assert VideoGame.find(video_games(:sonic_the_hedgehog).id).alternate_names.include?('Sonic der Hedgehog')
  end

protected
  def create_game(options = {})
    { :name => 'Zack and Wiki', :year => '2006', :alternate_names => {} }.merge(options)
  end
end
