require File.dirname(__FILE__) + '/../test_helper'
require 'video_game_system_controller'

# Reraise errors caught by the controller.
class VideoGameSystemController; def rescue_action(e) raise e end; end

class VideoGameSystemControllerTest < ActionController::TestCase
  # Anyone can see the index.
  def test_index
    login_as :quentin
    
    get :index
    assert_response :success
    assert_template 'list'
  end
  
  # Anyone can see the list.
  def test_list
    login_as :quentin
    
    get :list
    assert_response :success
    assert_template 'list'
  end
  
  # Anyone can see show.
  def test_show
    login_as :quentin
    
    get :show, :id => video_game_systems(:nes).id
    assert_response :success
    assert_template 'show'
    
    # And it should have @video_game_system set to the requested system.
    assert_equal video_game_systems(:nes), assigns(:video_game_system)
  end
  
  # Anyone can see new.
  def test_new
    login_as :quentin
    
    get :new
    assert_response :success
    assert_template 'new'
    
    # And it should have @video_game_system set to a new system.
    #assert_equal VideoGameSystem.new(), assigns(:video_game_system)
  end
  
  # Anyone can create.
  def test_create_normal_user
    login_as :quentin
    
    # Set @video_game_system before POSTing create.
    assert_difference 'VideoGameSystem.count' do
      post :create, :video_game_system => create_system
      assert_redirected_to new_video_game_system_path
    end
    
    # It should have gone successfully.
    assert_not_nil flash[:notice]
    assert_nil flash[:error]
    assert_nil flash[:video_game_system]
    assert_nil flash[:submission_note]
  end
  
  def test_bad_create_normal_user
    login_as :quentin
    
    # Constrast when we have a bad POST
    assert_no_difference 'VideoGameSystem.count' do
      post :create, :video_game_system => create_system(:name => video_game_systems(:nes).name)
      assert_redirected_to new_video_game_system_path
    end
    
    # It should NOT have gone successfully.
    assert_nil flash[:notice]
    assert_not_nil flash[:error]
    assert flash[:error_is_html]
    assert_not_nil flash[:video_game_system]
    assert_nil flash[:submission_note]
  end
  
  def test_complex_create_normal_user
    login_as :quentin
    
    # And now a complex success test with comments and alternate names.
    assert_difference 'VideoGameSystem.count' do
      assert_difference 'AlternateName.count', 3 do
        assert_difference 'Comment.count' do
          post :create, :video_game_system => create_system(:name => '1234', :manufacturer => '5678', :alternate_names => {1 => 'a', 2 => 'a', 3 => '', 4 => '', 5 => nil, 6 => 'b', 7 => 'c'}), :submission_note => 'A submission note.'
          assert_redirected_to new_video_game_system_path
        end
      end
    end
    
    # It should have gone successfully.
    assert_not_nil flash[:notice]
    assert_nil flash[:error]
    assert_nil flash[:video_game_system]
    assert_nil flash[:submission_note]
    
    # Should get one if we wrote the test right.
    video_game_system = VideoGameSystem.find_by_name('1234')
    assert_not_nil video_game_system
    
    # Main users should not auto-approve...
    assert video_game_system.pending_approval?
    assert_equal users(:quentin), video_game_system.creator
    assert_equal '5678', video_game_system.manufacturer
    
    # It should have eliminated nils, empties, and duplicates for the alternate_names.
    assert_equal 3, video_game_system.alternate_names.count
    assert video_game_system.alternate_names.to_ary.include?('a')
    assert video_game_system.alternate_names.to_ary.include?('b')
    assert video_game_system.alternate_names.to_ary.include?('c')
    
    # And we should find the comment attached.
    assert_equal 1, video_game_system.comments.count
    assert_equal 'A submission note.', video_game_system.comments[0].text
    assert_equal :submission_note, video_game_system.comments[0].comment_type
  end
  
  def test_complex_bad_create_normal_user
    login_as :quentin
    
    # And lastly a complex failure test with comments and alternate names.
    assert_no_difference 'VideoGameSystem.count' do
      assert_no_difference 'AlternateName.count' do
        assert_no_difference 'Comment.count' do
          post :create, :video_game_system => create_system(:name => nil, :manufacturer => nil, :alternate_names => {1 => 'a', 2 => 'a', 3 => '', 4 => '', 5 => nil, 6 => 'b', 7 => 'c'}), :submission_note => 'A submission note.'
          assert_redirected_to new_video_game_system_path
        end
      end
    end
    
    # It should have gone successfully.
    assert_nil flash[:notice]
    assert_not_nil flash[:error]
    assert flash[:error_is_html]
    assert_not_nil flash[:video_game_system]
    assert_not_nil flash[:submission_note]
  end
  
  def test_create_approver_user
    login_as :quentin
    
    # Oh right.  Approvers and above should have their changes auto-approve.
    user = users(:quentin)
    user.roles << roles(:video_game_system_approver)
    user.save
    
    assert_difference 'VideoGameSystem.count' do
      post :create, :video_game_system => create_system(:name => 'Nintendo 64', :manufacturer => 'Nintendo')
      assert_redirected_to new_video_game_system_path
    end
    
    # Should get one if we wrote the test right.
    video_game_system = VideoGameSystem.find_by_name('Nintendo 64')
    assert_not_nil video_game_system
    
    assert video_game_system.approved?
    assert_equal users(:quentin), video_game_system.creator
    assert_equal users(:quentin), video_game_system.approver
  end
  
  # Approvers and above can see edit.
  def test_edit_normal_user
    login_as :quentin
    
    get :edit, :id => video_game_systems(:nes).id
    assert_response 401
  end
  
  def test_edit_approver
    login_as :quentin
    
    user = users(:quentin)
    user.roles << roles(:video_game_system_approver)
    user.save
    
    get :edit, :id => video_game_systems(:nes).id
    assert_response :success
    assert_template 'edit'
    assert_equal video_game_systems(:nes), assigns(:video_game_system)
  end
  
  def test_edit_moderator
    login_as :quentin
    
    user = users(:quentin)
    user.roles << roles(:video_game_system_moderator)
    user.save
    
    get :edit, :id => video_game_systems(:nes).id
    assert_response :success
    assert_template 'edit'
    assert_equal video_game_systems(:nes), assigns(:video_game_system)
  end
  
  def test_edit_super_moderator
    login_as :quentin
    
    user = users(:quentin)
    user.roles << roles(:video_game_super_moderator)
    user.save
    
    get :edit, :id => video_game_systems(:nes).id
    assert_response :success
    assert_template 'edit'
    assert_equal video_game_systems(:nes), assigns(:video_game_system)
  end
  
  # Approvers can approve.
  def test_approve_normal_user
    login_as :quentin
    
    put :approve, :id => video_game_systems(:genesis).id
    assert_response 401
#    assert_redirected_to new_video_game_system_path
  end
  
  def test_approve_approver
    login_as :quentin
    
    user = users(:quentin)
    user.roles << roles(:video_game_system_approver)
    user.save
    
    put :approve, :id => video_game_systems(:genesis).id
    assert_redirected_to edit_video_game_system_path(video_game_systems(:genesis).id)
    assert_not_nil flash[:notice]
    assert VideoGameSystem.find(video_game_systems(:genesis).id).approved?
  end
  
  def test_approve_moderator
    login_as :quentin
    
    user = users(:quentin)
    user.roles << roles(:video_game_system_moderator)
    user.save
    
    put :approve, :id => video_game_systems(:genesis).id
    assert_redirected_to edit_video_game_system_path(video_game_systems(:genesis).id)
    assert_not_nil flash[:notice]
    assert VideoGameSystem.find(video_game_systems(:genesis).id).approved?
  end
  
  def test_approve_super_moderator
    login_as :quentin
    
    user = users(:quentin)
    user.roles << roles(:video_game_super_moderator)
    user.save
    
    put :approve, :id => video_game_systems(:genesis).id
    assert_redirected_to edit_video_game_system_path(video_game_systems(:genesis).id)
    assert_not_nil flash[:notice]
    assert VideoGameSystem.find(video_game_systems(:genesis).id).approved?
    
    # Okay, but NES SHOULDN'T approve until it's clean.
    put :approve, :id => video_game_systems(:nes).id
    assert_redirected_to edit_video_game_system_path(video_game_systems(:nes).id)
    assert_not_nil flash[:error]
    assert VideoGameSystem.find(video_game_systems(:nes).id).pending_changes?
    
    system = video_game_systems(:nes)
    system.delete_pending_change :name, :change, "NES"
    system.save
    
    put :approve, :id => video_game_systems(:nes).id
    assert_redirected_to edit_video_game_system_path(video_game_systems(:nes).id)
    assert_not_nil flash[:notice]
    assert VideoGameSystem.find(video_game_systems(:nes).id).approved?
  end
  
  # Approvers can approve.
  def test_approve_pending_change_normal_user
    login_as :quentin
    
    put :approve_pending_change, :id => video_game_systems(:nes).id, :change_action => 'change', :key => :name, :value => 'Nintendo Entertainment System'
    assert_response 401
#    assert_redirected_to new_video_game_system_path
  end
  
  def test_approve_pending_change_approver
    login_as :quentin
    
    user = users(:quentin)
    user.roles << roles(:video_game_system_approver)
    user.save
    
    put :approve_pending_change, :id => video_game_systems(:nes).id, :change_action => 'change', :key => :name, :value => 'Nintendo Entertainment System'
    assert_redirected_to edit_video_game_system_path(video_game_systems(:nes).id)
    assert_not_nil flash[:notice]
    assert_equal 'Entertainment System', VideoGameSystem.find(video_game_systems(:nes).id).name
    assert VideoGameSystem.find(video_game_systems(:nes).id).pending_changes[:name]['Nintendo Entertainment System'].include?(user.id)
  end
  
  def test_approve_pending_change_moderator
    login_as :quentin
    
    user = users(:quentin)
    user.roles << roles(:video_game_system_moderator)
    user.save
    
    put :approve_pending_change, :id => video_game_systems(:nes).id, :change_action => 'change', :key => :name, :value => 'Nintendo Entertainment System'
    assert_redirected_to edit_video_game_system_path(video_game_systems(:nes).id)
    assert_not_nil flash[:notice]
    assert_equal 'Nintendo Entertainment System', VideoGameSystem.find(video_game_systems(:nes).id).name
    assert !VideoGameSystem.find(video_game_systems(:nes).id).pending_changes[:name].include?('Nintendo Entertainment System')
  end
  
  def test_approve_pending_change_super_moderator
    login_as :quentin
    
    user = users(:quentin)
    user.roles << roles(:video_game_super_moderator)
    user.save
    
    put :approve_pending_change, :id => video_game_systems(:nes).id, :change_action => 'change', :key => :name, :value => 'Nintendo Entertainment System'
    assert_redirected_to edit_video_game_system_path(video_game_systems(:nes).id)
    assert_not_nil flash[:notice]
    assert_equal 'Nintendo Entertainment System', VideoGameSystem.find(video_game_systems(:nes).id).name
    assert !VideoGameSystem.find(video_game_systems(:nes).id).pending_changes[:name].include?('Nintendo Entertainment System')
  end
  
  # Approvers can reject.
  def test_reject_pending_change_normal_user
    login_as :quentin
    
    put :reject_pending_change, :id => video_game_systems(:nes).id, :change_action => 'change', :key => :years, :value => '2006'
    assert_response 401
#    assert_redirected_to new_video_game_system_path
  end
  
  def test_reject_pending_change_approver
    login_as :quentin
    
    user = users(:quentin)
    user.roles << roles(:video_game_system_approver)
    user.save
    
    put :reject_pending_change, :id => video_game_systems(:nes).id, :change_action => 'change', :key => :name, :value => 'NES'
    assert_redirected_to edit_video_game_system_path(video_game_systems(:nes).id)
    assert_not_nil flash[:notice]
    assert_equal 'Entertainment System', VideoGameSystem.find(video_game_systems(:nes).id).name
    assert !VideoGameSystem.find(video_game_systems(:nes).id).pending_changes[:name].include?('NES')
    
    put :reject_pending_change, :id => video_game_systems(:nes).id, :change_action => 'change', :key => :name, :value => 'NES'
    assert_redirected_to edit_video_game_system_path(video_game_systems(:nes).id)
    assert_not_nil flash[:notice]
    assert_equal 'Entertainment System', VideoGameSystem.find(video_game_systems(:nes).id).name
    assert !VideoGameSystem.find(video_game_systems(:nes).id).pending_changes[:name].include?('NES')
  end
  
  def test_reject_pending_change_moderator
    login_as :quentin
    
    user = users(:quentin)
    user.roles << roles(:video_game_system_moderator)
    user.save
    
    put :reject_pending_change, :id => video_game_systems(:nes).id, :change_action => 'change', :key => :name, :value => 'NES'
    assert_redirected_to edit_video_game_system_path(video_game_systems(:nes).id)
    assert_not_nil flash[:notice]
    assert_equal 'Entertainment System', VideoGameSystem.find(video_game_systems(:nes).id).name
    assert !VideoGameSystem.find(video_game_systems(:nes).id).pending_changes[:name].include?('NES')
  end
  
  def test_reject_pending_change_super_moderator
    login_as :quentin
    
    user = users(:quentin)
    user.roles << roles(:video_game_super_moderator)
    user.save
    
    put :reject_pending_change, :id => video_game_systems(:nes).id, :change_action => 'change', :key => :name, :value => 'NES'
    assert_redirected_to edit_video_game_system_path(video_game_systems(:nes).id)
    assert_not_nil flash[:notice]
    assert_equal 'Entertainment System', VideoGameSystem.find(video_game_systems(:nes).id).name
    assert !VideoGameSystem.find(video_game_systems(:nes).id).pending_changes[:name].include?('NES')
  end
  
  # Only moderators can add an alternate name.
  def test_add_alternate_name_normal_user
    login_as :quentin
    
    # TODO: reimplement me
    assert_no_difference 'AlternateName.count' do
      post :add_alternate_name, :id => video_game_systems(:nes).id, :new_alternate_name => 'Famicom'
      assert_response 401
    end
    
    system = VideoGameSystem.find(video_game_systems(:nes).id)
    assert !system.alternate_names.include?('Famicom')
  end
  
  def test_add_alternate_name_approver
    login_as :quentin
    
    user = users(:quentin)
    user.roles << roles(:video_game_system_approver)
    user.save
    
    system = VideoGameSystem.find(video_game_systems(:nes).id)
    assert system.pending_changes?
    
    assert_no_difference 'AlternateName.count' do
      put :add_alternate_name, :id => video_game_systems(:nes).id, :new_alternate_name => 'Famicom'
      assert_redirected_to edit_video_game_system_path(video_game_systems(:nes))
    end
    
    # However, approvers should have the name added to pending_changes.
    system = VideoGameSystem.find(video_game_systems(:nes).id)
    assert !system.alternate_names.include?('Famicom')
    assert system.pending_changes?
    assert_not_nil system.pending_changes
    assert_not_nil system.pending_changes[:alternate_names]
    assert_not_nil system.pending_changes[:alternate_names][:add]
    assert_not_nil system.pending_changes[:alternate_names][:add]['Famicom']
    assert system.pending_changes[:alternate_names][:add]['Famicom'].include?(users(:quentin).id)
  end
  
  def test_add_alternate_name_moderator
    login_as :quentin
    
    # And as for moderators, it should stick immediately.
    user = users(:quentin)
    user.roles << roles(:video_game_system_moderator)
    user.save
    
    assert_difference 'AlternateName.count' do
      put :add_alternate_name, :id => video_game_systems(:nes).id, :new_alternate_name => 'Famicom'
      assert_redirected_to edit_video_game_system_path(video_game_systems(:nes))
    end
    
    system = VideoGameSystem.find(video_game_systems(:nes).id)
    assert system.alternate_names.include?('Famicom')
#    assert_nil system.pending_changes[:alternate_names][:add]['My Little Goddess']
  end
  
  # Only moderators can delete an alternate name.
  def test_remove_alternate_name_normal_user
    login_as :quentin
    
    # TODO: reimplement me
    assert_no_difference 'AlternateName.count' do
      put :remove_alternate_name, :id => video_game_systems(:genesis).id, :name => 'Megadrive'
      assert_response 401
    end
    
    system = VideoGameSystem.find(video_game_systems(:genesis).id)
    assert system.alternate_names.include?('Megadrive')
  end
    
  def test_remove_alternate_name_approver
    login_as :quentin
    
    user = users(:quentin)
    user.roles << roles(:video_game_system_approver)
    user.save
    
    system = VideoGameSystem.find(video_game_systems(:genesis).id)
    
    assert_no_difference 'AlternateName.count' do
      put :remove_alternate_name, :id => video_game_systems(:genesis).id, :name => 'Megadrive'
      assert_redirected_to edit_video_game_system_path(video_game_systems(:genesis))
    end
    
    # However, approvers should have the name added to pending_changes.
    system = VideoGameSystem.find(video_game_systems(:genesis).id)
    assert system.alternate_names.include?('Megadrive')
    assert system.pending_approval?
    assert_not_nil system.pending_changes
    assert_not_nil system.pending_changes[:alternate_names]
    assert_not_nil system.pending_changes[:alternate_names][:remove]
    assert_not_nil system.pending_changes[:alternate_names][:remove]['Megadrive']
    assert system.pending_changes[:alternate_names][:remove]['Megadrive'].include?(users(:quentin).id)
  end
  
  def test_remove_nonexistent_alternate_name_approver
    login_as :quentin
    
    user = users(:quentin)
    user.roles << roles(:video_game_system_approver)
    user.save
    
    # Deleting a non-existent name should give an error.
    system = VideoGameSystem.find(video_game_systems(:genesis).id)
    
    assert_no_difference 'AlternateName.count' do
      put :remove_alternate_name, :id => video_game_systems(:genesis).id, :name => 'Magadrive'
      assert_redirected_to edit_video_game_system_path(video_game_systems(:genesis))
    end
    
    system = VideoGameSystem.find(video_game_systems(:genesis).id)
    assert !system.pending_changes?
    assert_nil system.pending_changes
  end
  
  def test_remove_alternate_name_moderator
    login_as :quentin
    
    # And as for moderators, it should stick immediately.
    user = users(:quentin)
    user.roles << roles(:video_game_system_moderator)
    user.save
    
    assert_difference 'AlternateName.count', -1 do
      put :remove_alternate_name, :id => video_game_systems(:genesis).id, :name => 'Megadrive'
      assert_redirected_to edit_video_game_system_path(video_game_systems(:genesis))
    end
    
    system = VideoGameSystem.find(video_game_systems(:genesis).id)
    assert !system.alternate_names.include?('Megadrive')
    assert !system.pending_changes?
    assert_nil system.pending_changes
    
    # Deleting a non-existent name should give an error.
    system = VideoGameSystem.find(video_game_systems(:genesis).id)
    
    assert_no_difference 'AlternateName.count' do
      put :remove_alternate_name, :id => video_game_systems(:genesis).id, :name => 'Magadrive'
      assert_redirected_to edit_video_game_system_path(video_game_systems(:genesis))
    end
    
    system = VideoGameSystem.find(video_game_systems(:genesis).id)
    assert !system.pending_changes?
    assert_nil system.pending_changes
  end
  
  # Moderators can destroy.
  def test_destroy_normal_user
    login_as :quentin
    
    delete :destroy, :id => video_game_systems(:nes).id
    assert_response 401
    assert_nothing_raised { VideoGameSystem.find(video_game_systems(:nes).id) }
  end
  
  def test_destroy_approver
    login_as :quentin
    
    user = users(:quentin)
    user.roles << roles(:video_game_system_approver)
    user.save
    
    delete :destroy, :id => video_game_systems(:nes).id
    assert_response 401
    assert_nothing_raised { VideoGameSystem.find(video_game_systems(:nes).id) }
  end
  
  def test_destroy_moderator
    login_as :quentin
    
    user = users(:quentin)
    user.roles << roles(:video_game_system_moderator)
    user.save
    
    delete :destroy, :id => video_game_systems(:nes).id
    assert_redirected_to manage_video_game_system_path
    assert_raise ActiveRecord::RecordNotFound do; VideoGameSystem.find(video_game_systems(:nes).id); end
  end
  
  def test_destroy_super_moderator
    login_as :quentin
    
    user = users(:quentin)
    user.roles << roles(:video_game_super_moderator)
    user.save
    
    delete :destroy, :id => video_game_systems(:genesis).id
    assert_redirected_to manage_video_game_system_path
    assert_raise ActiveRecord::RecordNotFound do; VideoGameSystem.find(video_game_systems(:genesis).id); end
  end
  
  # Approvers can manage.
  def test_manage_normal_user
    login_as :quentin
    
    get :manage
    assert_response 401
  end
  
  def test_manage_approver
    login_as :quentin
    
    user = users(:quentin)
    user.roles << roles(:video_game_system_approver)
    user.save
    
    get :manage
    assert_response :success
    assert_template 'manage'
  end
  
  def test_manage_moderator
    login_as :quentin
    
    user = users(:quentin)
    user.roles << roles(:video_game_system_moderator)
    user.save
    
    get :manage
    assert_response :success
    assert_template 'manage'
  end
  
  def test_manage_super_moderator
    login_as :quentin
    
    user = users(:quentin)
    user.roles << roles(:video_game_super_moderator)
    user.save
    
    get :manage
    assert_response :success
    assert_template 'manage'
  end
  
  def test_select_merge_normal_user
    login_as :quentin
    
    get :select_merge, :id => video_game_systems(:famicon).id
    assert_response 401
  end
  
  def test_select_merge_approver
    login_as :quentin
    
    user = users(:quentin)
    user.roles << roles(:video_game_system_approver)
    user.save
    
    get :select_merge, :id => video_game_systems(:famicon).id
    assert_response 401
  end
  
  def test_select_merge_moderator
    login_as :quentin
    
    user = users(:quentin)
    user.roles << roles(:video_game_system_moderator)
    user.save
    
    get :select_merge, :id => video_game_systems(:famicon).id
    assert_response :success
    assert_template 'select_merge'
    
    # And it should have @video_game_system set to the requested system.
    assert_equal video_game_systems(:famicon), assigns(:video_game_system)
  end
  
  def test_select_merge_super_moderator
    login_as :quentin
    
    user = users(:quentin)
    user.roles << roles(:video_game_super_moderator)
    user.save
    
    get :select_merge, :id => video_game_systems(:famicon).id
    assert_response :success
    assert_template 'select_merge'
    
    # And it should have @video_game_system set to the requested system.
    assert_equal video_game_systems(:famicon), assigns(:video_game_system)
  end
  
  # TODO: search_select_merge
  
  def test_merge_into_normal_user
    login_as :quentin
    
    put :merge_into, :id => video_game_systems(:famicon).id, :dest => video_game_systems(:nes).id
    assert_response 401
    assert_nothing_raised { VideoGameSystem.find(video_game_systems(:famicon).id) }
	assert !VideoGameSystem.find(video_game_systems(:nes).id).alternate_names.include?('NES')
  end
  
  def test_merge_into_approver
    login_as :quentin
    
    user = users(:quentin)
    user.roles << roles(:video_game_system_approver)
    user.save
    
    put :merge_into, :id => video_game_systems(:famicon).id, :dest => video_game_systems(:nes).id
    assert_response 401
    assert_nothing_raised { VideoGameSystem.find(video_game_systems(:famicon).id) }
    assert !VideoGameSystem.find(video_game_systems(:nes).id).alternate_names.include?('Famicon')
  end
  
  def test_merge_into_moderator
    login_as :quentin
    
    user = users(:quentin)
    user.roles << roles(:video_game_system_moderator)
    user.save
    
    put :merge_into, :id => video_game_systems(:famicon).id, :dest => video_game_systems(:nes).id
    assert_redirected_to manage_video_game_system_path
	
    assert_raise ActiveRecord::RecordNotFound do; VideoGameSystem.find(video_game_systems(:famicon).id); end
    assert VideoGameSystem.find(video_game_systems(:nes).id).alternate_names.include?('Famicon')
  end
  
  def test_merge_into_super_moderator
    login_as :quentin
    
    user = users(:quentin)
    user.roles << roles(:video_game_super_moderator)
    user.save
    
    put :merge_into, :id => video_game_systems(:famicon).id, :dest => video_game_systems(:nes).id
    assert_redirected_to manage_video_game_system_path
	
    assert_raise ActiveRecord::RecordNotFound do; VideoGameSystem.find(video_game_systems(:famicon).id); end
    assert VideoGameSystem.find(video_game_systems(:nes).id).alternate_names.include?('Famicon')
  end

protected
  def create_system(options = {})
    { :name => 'Dreamcast', :manufacturer => 'Sega', :alternate_names => {} }.merge(options)
  end
end
