require File.dirname(__FILE__) + '/../test_helper'

class AnimeSeriesTest < ActiveSupport::TestCase
  Error = ''
  fixtures :anime_series
  fixtures :comments
  fixtures :users
  
  def test_validations
    assert !create_series(:name => nil).valid?
    assert !create_series(:name => '').valid?
    # TODO: Associate and test genres.
#    assert !create_series(:genres => []).valid?
    assert !create_series(:anidb_id => 630).valid?
    
    assert_valid create_series()
    assert_nil create_series().anidb_id
    assert_valid create_series(:anidb_id => 3651)
    assert_equal 3651, create_series(:anidb_id => 3651).anidb_id
    
    assert_nothing_raised { create_series(:alternate_names => [1, :symbol, Object.new]) }
    assert_nothing_raised { create_series(:alternate_names => ['a', 'b', 'c']) }
    assert_equal 630, create_series(:anidb_id => 630).anidb_id
    assert_equal ['a', 'b', 'c'], create_series(:alternate_names => ['a', 'b', 'c']).alternate_names.to_ary
    assert_valid create_series(:alternate_names => ['a', 'b', 'c'])
  end
  
  def test_comments
    # Test adding a comment.
    comment = Comment.new(:text => 'Hey there.  I like this series.')
    child = Comment.new(:text => 'So do I.')
    series = anime_series(:kinos_journey)
    old_comments = series.comments.clone
    
    assert !series.comments.empty?
    assert_difference 'series.comments.size' do
      series.comments << comment
    end
    assert_valid series
    series.save
    assert series.comments.include?(comment)
    assert_equal comment, series.comments[-1]
    assert_equal old_comments + [comment], series.comments
    assert_nil comment.parent
    assert_equal series, comment.item
    assert_nothing_raised { Comment.find(comment.id) }
    
    # Test adding a child comment.
    # NOTE: Can't save series and have ALL children saved.
    assert_no_difference 'series.comments.size' do
      child = comment.children.create(:text => 'So do I.')
    end
    child.save
    assert_equal old_comments + [comment], series.comments
    # Uh, no way to make sure we can find it...
    #assert_nothing_raised { Comment.find(child.id) }
    #assert_equal child, Comment.find(child.id)
    
    # Test deleting a comment
    assert_difference 'series.comments.size', -1 do
      series.comments.delete(comment)
    end
    assert_equal old_comments, series.comments
    assert_raise ActiveRecord::RecordNotFound do; Comment.find(comment.id); end
    assert_raise ActiveRecord::RecordNotFound do; Comment.find(child.id); end
    
    # Test deleting all comments
    assert_nothing_raised { series.comments.delete_all }
    assert series.comments.empty?
    assert_equal [], series.comments
  end
  
  # TODO: Test genres.
  
  # TODO: Test AnimeContinuum
  
  def test_alternate_names
    series = anime_series(:kinos_journey)
    
    assert_kind_of Array, series.alternate_names
    assert !series.alternate_names.empty?
    assert_equal 2, series.alternate_names.size
    assert series.alternate_names.include?("Kino's Journey ~the beautiful world~")
    assert series.alternate_names.include?("Kino no Tabi ~the beautiful world~")
    
    # Try adding
    assert_difference 'series.alternate_names.count' do
      series.alternate_names << 'Kinou no Tabii'
      series.save
    end
    assert series.alternate_names.include?('Kinou no Tabii')
    
    # And deleting
    assert_difference 'series.alternate_names.count', -1 do
      series.alternate_names.delete('Kinou no Tabii')
    end
    series.save
    assert !series.alternate_names.include?('Kinou no Tabii')
    
    # And then massive deletion
    assert_nothing_raised { series.alternate_names.delete('Nonsense name!') }
    
    assert_nothing_raised { series.alternate_names.delete_all }
    series.save
    assert series.alternate_names.empty?
  end
  
  def test_years_validation
    assert !create_series(:years => '').valid?
    assert !create_series(:years => 'a2@#%$aioe.').valid?
    assert !create_series(:years => '19a8').valid?
    assert !create_series(:years => '1').valid?
    assert !create_series(:years => '1899').valid?
    assert !create_series(:years => '2101').valid?
    assert !create_series(:years => '2005-1999').valid?
    assert !create_series(:years => '2005,1999').valid?
    assert !create_series(:years => '1995,2006-1999,2008').valid?
    assert !create_series(:years => '1999,1998-1999,2001').valid?
    assert !create_series(:years => '1999,1999-2001').valid?
    
    assert_valid create_series(:years => nil)
    assert_valid create_series(:years => '1900')
    assert_valid create_series(:years => '2006')
    assert_valid create_series(:years => '2100')
    assert_valid create_series(:years => '2006-2007')
    assert_valid create_series(:years => '2006,2007')
    assert_valid create_series(:years => '2005-2006,2008')
    assert_valid create_series(:years => '1999,2005-2006,2008')
  end
  
  def test_state_transitions
    # Here we test a legitimate lifecycle of several series.
    set_current_user :admin
    
    # Can't modify something that is pending_changes, but should -> approve -> changed
    series = anime_series(:kamichu)
    old_series = series.clone
    
    assert series.pending_changes?
    series.changed!
    assert series.pending_changes?
    series.approve!
    assert series.pending_changes?
    
    # TODO: assert change in genres.
    assert_equal '2005-2006', series.years
    assert_equal 'Kamichur?', series.name
    assert_not_nil series.pending_changes
    
    # Should throw an error because series.years has a conflict.
    assert series.errors.empty?
    assert !series.approvable_changes?
    assert !series.errors.empty?
    assert_equal "has more than one pending change which must be resolved prior to approving all changes.", series.errors[:years]
    series.errors.clear
    assert series.errors.empty?
    series.approve_changes!
    assert !series.errors.empty?
    assert_equal "has more than one pending change which must be resolved prior to approving all changes.", series.errors[:years]
    assert_equal '2005-2006', series.years
    assert_equal 'Kamichur?', series.name
    assert_not_nil series.pending_changes
    
    # Removing the extra year should resolve the conflict.
    series.delete_pending_change(:years, :change, "2006")
    assert series.approvable_changes?
    series.approve_changes!
    assert series.approved?
    assert_equal '2005', series.years
    assert_equal 'Kamichu!', series.name
    assert_nil series.pending_changes
    # TODO: assert change in genres.
    
    # This test is now irrelevant. We 'change' without a transition.
    #series.changed!
    #assert series.pending_changes?
    
    assert_equal old_series.creator, series.creator
    assert_equal old_series.created_at, series.created_at
    assert_equal old_series.approver, series.approver
    assert_equal old_series.approved_at, series.approved_at
    assert_not_equal old_series.updater, series.updater
    assert_not_equal old_series.updated_at, series.updated_at
    
    # Can't approve something that is approved, but should -> changed -> approve
    series = anime_series(:kinos_journey)
    old_series = series.clone
    assert series.approved?
    series.approve!
    assert series.approved?
    series.approve_changes!
    assert series.approved?
    # This test now irrelevant.  We 'change' without a transition.
    #series.changed!
    #assert series.pending_changes?
    #series.approve_changes!
    #assert series.approved?
    
    assert_equal old_series.creator, series.creator
    assert_equal old_series.created_at, series.created_at
    assert_equal old_series.approver, series.approver
    assert_equal old_series.approved_at, series.approved_at
    # These two are irrelevant, as we now have never changed state.
    #assert_not_equal old_series.updater, series.updater
    #assert_not_equal old_series.updated_at, series.updated_at
    
    # Can't change something that is pending_approval, but should -> approve -> changed
    series = anime_series(:stellvia)
    old_series = series.clone
    assert series.pending_approval?
    series.changed!
    assert series.pending_approval?
    series.approve_changes!
    assert series.pending_approval?
    series.approve!
    assert series.approved?
    # This test now irrelevant.  We 'change' without a transition.
    #series.changed!
    #assert series.pending_changes?
    #series.approve_changes!
    #assert series.approved?
    
    assert_equal old_series.creator, series.creator
    assert_equal old_series.created_at, series.created_at
    assert_not_equal old_series.approver, series.approver
    assert_not_equal old_series.approved_at, series.approved_at
    assert_not_equal old_series.updater, series.updater
    assert_not_equal old_series.updated_at, series.updated_at
    
    # And new series should be pending_approval.
    series = create_series()
    series.save
    old_series = series.clone
    assert_equal users(:admin), series.creator
    
    set_current_user :quentin
    
    assert series.pending_approval?
    series.changed!
    assert series.pending_approval?
    series.approve_changes!
    assert series.pending_approval?
    series.approve!
    assert series.approved?
    # This test now irrelevant.  We 'change' without a transition.
    #series.changed!
    #assert series.pending_changes?
    #series.approve_changes!
    #assert series.approved?
    
    assert_equal old_series.creator, series.creator
    assert_equal old_series.created_at, series.created_at
    assert_not_equal old_series.approver, series.approver
    assert_not_equal old_series.approved_at, series.approved_at
    assert_not_equal old_series.updater, series.updater
    assert_not_equal old_series.updated_at, series.updated_at
    
    assert_equal users(:quentin), series.approver
    
    # TODO: Test that approver_id, approved_at are set
  end
  
  def test_pending_changes_non_moderator
    set_current_user :quentin
    
    series = anime_series(:kinos_journey)
    
    assert_equal "Kino's Journey", series.name
    assert_nil series.pending_changes
    assert series.approved?
    
    series.add_pending_change_support :name, :change, 'Sommat else'
    
    assert series.pending_changes?
    assert_equal "Kino's Journey", series.name
    assert_not_nil series.pending_changes
    assert_not_nil series.pending_changes[:name]
    assert_not_nil series.pending_changes[:name]['Sommat else']
    assert series.pending_changes[:name]['Sommat else'].include?(users(:quentin).id)
    
    series.pending_changes = nil
    
    assert_nil series.pending_changes
    assert series.approved?
    
    assert_equal "2003", series.years
    
    series.add_pending_change_support :years, :change, "2004"
    
    assert_equal "2003", series.years
    assert_not_nil series.pending_changes
    assert_not_nil series.pending_changes[:years]
    assert_not_nil series.pending_changes[:years]["2004"]
    assert series.pending_changes[:years]["2004"].include?(users(:quentin).id)
    
    assert_equal 575, series.anidb_id
    
    series.add_pending_change_support :anidb_id, :change, 576
    
    assert_equal 575, series.anidb_id
    assert_not_nil series.pending_changes
    assert_not_nil series.pending_changes[:years]
    assert_not_nil series.pending_changes[:years]["2004"]
    assert series.pending_changes[:years]["2004"].include?(users(:quentin).id)
    assert_not_nil series.pending_changes[:anidb_id]
    assert_not_nil series.pending_changes[:anidb_id][576]
    assert series.pending_changes[:anidb_id][576].include?(users(:quentin).id)
    
    # Also, test that adding pending changes to something pending approval does
    # nothing to the state.
    
    series = anime_series(:stellvia)
    
    assert_nil series.pending_changes
    assert series.pending_approval?
    assert_equal "Stellvia of the Universe", series.name
    
    series.add_pending_change_support :name, :change, "Stellvia"
    
    assert series.pending_approval?
    assert_equal "Stellvia of the Universe", series.name
    assert_not_nil series.pending_changes
    assert_not_nil series.pending_changes[:name]
    assert_not_nil series.pending_changes[:name]["Stellvia"]
    assert series.pending_changes[:name]["Stellvia"].include?(users(:quentin).id)
    
    # TODO: Alternate names, genres
    # TODO: Test remove pending changes.
  end
  
  def test_pending_changes_moderator
    set_current_user :quentin
    
    user = users(:quentin)
    user.roles << roles(:anime_series_moderator)
    user.save
    
    series = anime_series(:kinos_journey)
    
    assert_equal "Kino's Journey", series.name
    assert_nil series.pending_changes
    assert series.approved?
    
    series.name = 'Sommat else'
    
    assert series.approved?
    assert_equal 'Sommat else', series.name
    assert_nil series.pending_changes
    
    assert_equal "2003", series.years
    
    series.years = "2004"
    
    assert_equal "2004", series.years
    assert_nil series.pending_changes
    
    assert_equal 575, series.anidb_id
    
    series.anidb_id = 576
    
    assert_equal 576, series.anidb_id
    assert_nil series.pending_changes
    
    # Also, test that adding pending changes to something pending approval does
    # nothing to the state.
    
    series = anime_series(:stellvia)
    
    assert_nil series.pending_changes
    assert series.pending_approval?
    assert_equal "Stellvia of the Universe", series.name
    
    series.name = "Stellvia"
    
    assert series.pending_approval?
    assert_equal "Stellvia", series.name
    assert_nil series.pending_changes
    
    # TODO: Alternate names, genres
    # TODO: Test remove pending changes.
  end

  def test_merge_into
    # Merging into should delete the anime series, merge its names into the
    # other, and reassign all other anime series tied to it.
    series = create_series(:name => 'Uchuu no Stellvia')
    series.alternate_names << 'Stellvia'
    series.save
    
    series.anime_genres << AnimeGenre.find(2)
    series.save
    
    assert !AnimeSeries.find_by_name('Uchuu no Stellvia').nil?
    assert AnimeSeries.find_by_name('Uchuu no Stellvia').anime_genres.include?(AnimeGenre.find(2))
    assert_equal 0, AnimeSeries.find_by_name('Stellvia of the Universe').alternate_names.length
    assert_equal 1, AnimeSeries.find_by_name('Stellvia of the Universe').anime_genres.length
    
    # Let's try merging.
    series.merge_into!(AnimeSeries.find_by_name('Stellvia of the Universe'))
    
    assert_nil AnimeSeries.find_by_name('Uchuu no Stellvia')
    
    series = AnimeSeries.find_by_name('Stellvia of the Universe')
    assert_equal 2, series.alternate_names.length
    assert series.alternate_names.include?('Uchuu no Stellvia')
    assert series.alternate_names.include?('Stellvia')
    assert_equal 2, series.anime_genres.length
    assert series.anime_genres.include?(AnimeGenre.find(2))
  end

protected
  def create_series(options = {})
    AnimeSeries.new({ :name => 'The Melancholy of Haruhi Suzumiya', :years => '2006' }.merge(options))
  end
  
  def set_current_user(user_key)
    User.stamper = users(user_key)
  end
end
