require File.dirname(__FILE__) + '/../test_helper'

class CommentTest < ActiveSupport::TestCase
  fixtures :comments
  fixtures :users
  
  def test_types
    # Make sure the types are limited
    Comment.types.each_key do |type|
      assert_valid create_comment(:comment_type => type)
    end
    assert !create_comment(:comment_type => :crazy_type).valid?
    
    # And that they return as symbols
    assert_equal :approval_pending_modification, comments(:root_comment).comment_type
    assert_equal :general_comment, comments(:child_comment).comment_type
    assert_equal :request_for_comments, comments(:brother_comment).comment_type
  end
  
  def test_parent
    assert_equal nil, comments(:root_comment).parent
    assert_equal comments(:root_comment), comments(:child_comment).parent
    assert_equal nil, create_comment(:parent_id => 0).parent
    assert_equal comments(:root_comment), create_comment(:parent_id => comments(:root_comment).id).parent
  end
  
  def test_children
    assert_equal 2, comments(:root_comment).children.size
    assert comments(:root_comment).children.include?(comments(:child_comment))
    assert comments(:root_comment).children.include?(comments(:brother_comment))
  end
  
  def test_new_comment
    set_current_user :quentin
    
    comment = nil
    assert_difference 'Comment.find(comments(:brother_comment).id).children.size' do
      comment = create_comment()
      assert_equal :general_comment, comment.comment_type
      assert_equal nil, comment.parent
      assert_equal [], comment.children
      assert_valid comment
      comment.comment_type = :request_approval
      comment.parent = comments(:brother_comment)
      assert_valid comment
      comment.save
      assert_equal users(:quentin), comment.creator
    end
    assert_not_nil comment
    assert Comment.find(comments(:brother_comment).id).children.include?(comment)
    
    new_comment = nil
    assert_difference 'Comment.find(comments(:brother_comment).id).children.size' do
      # If item is nil, it should auto-fill when adding to a children.
      new_comment = create_comment(:parent => comments(:brother_comment), :item => nil)
      new_comment.comment_type = :general_comment
      comments(:brother_comment).children << new_comment
      assert comments(:brother_comment).valid?
      comments(:brother_comment).save
    end
    assert_not_nil new_comment
    assert Comment.find(comments(:brother_comment).id).children.include?(new_comment)
    
    assert !create_comment(:item => nil).valid?
    assert !create_comment(:text => nil).valid?
    assert !create_comment(:text => '').valid?
  end
  
  def test_destroy_comment_should_chain
    comment_id = 0
    new_comment_id = 0
    
    assert_difference 'Comment.find(comments(:brother_comment).id).children.size', 2 do
      comment = create_comment()
      comment.parent = comments(:brother_comment)
      comment.save
      comment_id = comment.id
      new_comment = create_comment()
      new_comment.parent = comments(:brother_comment)
      new_comment.save
      new_comment_id = new_comment.id
    end
    
    assert_not_equal 0, comment_id
    assert_not_equal 0, new_comment_id
    
    assert_difference 'Comment.find(comments(:root_comment).id).children.size', -1 do
      assert_nothing_raised do; Comment.find(comments(:brother_comment).id); end
      assert_nothing_raised do; Comment.find(comment_id); end
      assert_nothing_raised do; Comment.find(new_comment_id); end
      comments(:brother_comment).destroy
      assert_raise ActiveRecord::RecordNotFound do; Comment.find(comments(:brother_comment).id); end
      assert_raise ActiveRecord::RecordNotFound do; Comment.find(comment_id); end
      assert_raise ActiveRecord::RecordNotFound do; Comment.find(new_comment_id); end
    end
  end

protected
  def create_comment(options = {})
    Comment.new({ :item_id => 1, :item_type => 'AnimeSeries', :text => 'This is a new comment.' }.merge(options))
  end
  
  def set_current_user(user_key)
    User.stamper = users(user_key)
  end
end
