require 'test_helper'

class VideoGameContinuumTest < ActiveSupport::TestCase
  Error = ''
  fixtures :video_game_continuums
  fixtures :comments
  fixtures :users
  
  def test_validations
    assert !create_continuum(:name => nil).valid?
    assert !create_continuum(:name => '').valid?
    
    assert_valid create_continuum()
    
    assert_nothing_raised { create_continuum(:alternate_names => [1, :symbol, Object.new]) }
    assert_nothing_raised { create_continuum(:alternate_names => ['a', 'b', 'c']) }
    assert_equal ['a', 'b', 'c'], create_continuum(:alternate_names => ['a', 'b', 'c']).alternate_names.to_ary
    assert_valid create_continuum(:alternate_names => ['a', 'b', 'c'])
  end
  
  def test_comments
    # Test adding a comment.
    comment = Comment.new(:text => 'Hey there.  I like this continuum.')
    child = Comment.new(:text => 'So do I.')
    continuum = video_game_continuums(:sonic)
    old_comments = continuum.comments.clone
    
    assert continuum.comments.empty?
    assert_difference 'continuum.comments.size' do
      continuum.comments << comment
    end
    assert_valid continuum
    continuum.save
    assert continuum.comments.include?(comment)
    assert_equal comment, continuum.comments[-1]
    assert_equal old_comments + [comment], continuum.comments
    assert_nil comment.parent
    assert_equal continuum, comment.item
    assert_nothing_raised { Comment.find(comment.id) }
    
    # Test adding a child comment.
    # NOTE: Can't save continuum and have ALL children saved.
    assert_no_difference 'continuum.comments.size' do
      child = comment.children.create(:text => 'So do I.')
    end
    child.save
    assert_equal old_comments + [comment], continuum.comments
    # Uh, no way to make sure we can find it...
    #assert_nothing_raised { Comment.find(child.id) }
    #assert_equal child, Comment.find(child.id)
    
    # Test deleting a comment
    assert_difference 'continuum.comments.size', -1 do
      continuum.comments.delete(comment)
    end
    assert_equal old_comments, continuum.comments
    assert_raise ActiveRecord::RecordNotFound do; Comment.find(comment.id); end
    assert_raise ActiveRecord::RecordNotFound do; Comment.find(child.id); end
    
    # Test deleting all comments
    assert_nothing_raised { continuum.comments.delete_all }
    assert continuum.comments.empty?
    assert_equal [], continuum.comments
  end
  
  def test_alternate_names
    continuum = video_game_continuums(:sonic)
    
    assert_kind_of Array, continuum.alternate_names
    assert !continuum.alternate_names.empty?
    assert_equal 2, continuum.alternate_names.size
    assert continuum.alternate_names.include?("Sonic")
    assert continuum.alternate_names.include?("Shadow the Hedgehog")
    
    # Try adding
    assert_difference 'continuum.alternate_names.count' do
      continuum.alternate_names << 'Shadow'
      continuum.save
    end
    assert continuum.alternate_names.include?('Shadow')
    
    # And deleting
    assert_difference 'continuum.alternate_names.count', -1 do
      continuum.alternate_names.delete('Shadow')
    end
    continuum.save
    assert !continuum.alternate_names.include?('Shadow')
    
    # And then massive deletion
    assert_nothing_raised { continuum.alternate_names.delete('Nonsense name!') }
    
    assert_nothing_raised { continuum.alternate_names.delete_all }
    continuum.save
    assert continuum.alternate_names.empty?
  end
  
  def test_state_transitions
    # Here we test a legitimate lifecycle of several continuums.
    set_current_user :admin
    
    # Can't modify something that is pending_changes, but should -> approve -> changed
    continuum = video_game_continuums(:dragon_quest)
    old_continuum = continuum.clone
    
    assert continuum.pending_changes?
    continuum.changed!
    assert continuum.pending_changes?
    continuum.approve!
    assert continuum.pending_changes?
    
    assert_equal 'Dragon Quest', continuum.name
    assert_not_nil continuum.pending_changes
    
    # Should throw an error because continuum.name has a conflict.
    assert continuum.errors.empty?
    assert !continuum.approvable_changes?
    assert !continuum.errors.empty?
    assert_equal "has more than one pending change which must be resolved prior to approving all changes.", continuum.errors[:name]
    continuum.errors.clear
    assert continuum.errors.empty?
    continuum.approve_changes!
    assert !continuum.errors.empty?
    assert_equal "has more than one pending change which must be resolved prior to approving all changes.", continuum.errors[:name]
    assert_equal 'Dragon Quest', continuum.name
    assert_not_nil continuum.pending_changes
    
    # Removing the extra name should resolve the conflict.
    continuum.delete_pending_change :name, :change, "Dragon Warrior"
    assert continuum.approvable_changes?
    continuum.approve_changes!
    assert continuum.approved?
    assert_equal 'Dragon Quest/Dragon Warrior', continuum.name
    assert_nil continuum.pending_changes
    
    # This test is now irrelevant. We 'change' without a transition.
    #continuum.changed!
    #assert continuum.pending_changes?
    
    assert_equal old_continuum.creator, continuum.creator
    assert_equal old_continuum.created_at, continuum.created_at
    assert_equal old_continuum.approver, continuum.approver
    assert_equal old_continuum.approved_at, continuum.approved_at
    assert_not_equal old_continuum.updater, continuum.updater
    assert_not_equal old_continuum.updated_at, continuum.updated_at
    
    # Can't approve something that is approved, but should -> changed -> approve
    continuum = video_game_continuums(:mario)
    old_continuum = continuum.clone
    assert continuum.approved?
    continuum.approve!
    assert continuum.approved?
    continuum.approve_changes!
    assert continuum.approved?
    # This test now irrelevant.  We 'change' without a transition.
    #continuum.changed!
    #assert continuum.pending_changes?
    #continuum.approve_changes!
    #assert continuum.approved?
    
    assert_equal old_continuum.creator, continuum.creator
    assert_equal old_continuum.created_at, continuum.created_at
    assert_equal old_continuum.approver, continuum.approver
    assert_equal old_continuum.approved_at, continuum.approved_at
    # These two are irrelevant, as we now have never changed state.
    #assert_not_equal old_continuum.updater, continuum.updater
    #assert_not_equal old_continuum.updated_at, continuum.updated_at
    
    # Can't change something that is pending_approval, but should -> approve -> changed
    continuum = video_game_continuums(:sonic)
    old_continuum = continuum.clone
    assert continuum.pending_approval?
    continuum.changed!
    assert continuum.pending_approval?
    continuum.approve_changes!
    assert continuum.pending_approval?
    continuum.approve!
    assert continuum.approved?
    # This test now irrelevant.  We 'change' without a transition.
    #continuum.changed!
    #assert continuum.pending_changes?
    #continuum.approve_changes!
    #assert continuum.approved?
    
    assert_equal old_continuum.creator, continuum.creator
    assert_equal old_continuum.created_at, continuum.created_at
    assert_not_equal old_continuum.approver, continuum.approver
    assert_not_equal old_continuum.approved_at, continuum.approved_at
    assert_not_equal old_continuum.updater, continuum.updater
    assert_not_equal old_continuum.updated_at, continuum.updated_at
    
    # And new continuums should be pending_approval.
    continuum = create_continuum()
    continuum.save
    old_continuum = continuum.clone
    assert_equal users(:admin), continuum.creator
    
    set_current_user :quentin
    
    assert continuum.pending_approval?
    continuum.changed!
    assert continuum.pending_approval?
    continuum.approve_changes!
    assert continuum.pending_approval?
    continuum.approve!
    assert continuum.approved?
    # This test now irrelevant.  We 'change' without a transition.
    #continuum.changed!
    #assert continuum.pending_changes?
    #continuum.approve_changes!
    #assert continuum.approved?
    
    assert_equal old_continuum.creator, continuum.creator
    assert_equal old_continuum.created_at, continuum.created_at
    assert_not_equal old_continuum.approver, continuum.approver
    assert_not_equal old_continuum.approved_at, continuum.approved_at
    assert_not_equal old_continuum.updater, continuum.updater
    assert_not_equal old_continuum.updated_at, continuum.updated_at
    
    assert_equal users(:quentin), continuum.approver
    
    # TODO: Test that approver_id, approved_at are set
  end
  
  def test_pending_changes_non_moderator
    set_current_user :quentin
    
    continuum = video_game_continuums(:mario)
    
    assert_equal "Super Mario Brothers", continuum.name
    assert_nil continuum.pending_changes
    assert continuum.approved?
    
    continuum.add_pending_change_support :name, :change, 'Sommat else'
    
    assert continuum.pending_changes?
    assert_equal "Super Mario Brothers", continuum.name
    assert_not_nil continuum.pending_changes
    assert_not_nil continuum.pending_changes[:name]
    assert_not_nil continuum.pending_changes[:name]['Sommat else']
    assert continuum.pending_changes[:name]['Sommat else'].include?(users(:quentin).id)
    
    continuum.pending_changes = nil
    
    assert_nil continuum.pending_changes
    assert continuum.approved?
    
    # Also, test that adding pending changes to something pending approval does
    # nothing to the state.
    
    continuum = video_game_continuums(:sonic)
    
    assert_nil continuum.pending_changes
    assert continuum.pending_approval?
    assert_equal "Sonic the Hedgehog", continuum.name
    
    continuum.add_pending_change_support :name, :change, "Sonic"
    
    assert continuum.pending_approval?
    assert_equal "Sonic the Hedgehog", continuum.name
    assert_not_nil continuum.pending_changes
    assert_not_nil continuum.pending_changes[:name]
    assert_not_nil continuum.pending_changes[:name]["Sonic"]
    assert continuum.pending_changes[:name]["Sonic"].include?(users(:quentin).id)
    
    # TODO: Alternate names, genres
    # TODO: Test remove pending changes.
  end
  
  def test_pending_changes_moderator
    set_current_user :quentin
    
    user = users(:quentin)
    user.roles << roles(:video_game_continuum_moderator)
    user.save
    
    continuum = video_game_continuums(:mario)
    
    assert_equal "Super Mario Brothers", continuum.name
    assert_nil continuum.pending_changes
    assert continuum.approved?
    
    continuum.name = 'Sommat else'
    
    assert continuum.approved?
    assert_equal 'Sommat else', continuum.name
    assert_nil continuum.pending_changes
    
    # Also, test that adding pending changes to something pending approval does
    # nothing to the state.
    
    continuum = video_game_continuums(:sonic)
    
    assert_nil continuum.pending_changes
    assert continuum.pending_approval?
    assert_equal "Sonic the Hedgehog", continuum.name
    
    continuum.name = "Sonic"
    
    assert continuum.pending_approval?
    assert_equal "Sonic", continuum.name
    assert_nil continuum.pending_changes
    
    # TODO: Alternate names, genres
    # TODO: Test remove pending changes.
  end
  
  def test_merge_into
    # Merging into should delete the video game continuum, merge its names into
    # the other, and reassign all other video games tied to it.
    continuum = create_continuum(:name => 'Shadow the Hedgehog')
    continuum.alternate_names << 'Shadow'
    continuum.alternate_names << 'Sonic the Hedgehog'
    continuum.save
    
#    series = VideoGameSeries.new({:name => 'Magical Girl Lyrical Nanoha'})
#    series.video_game_continuum = continuum
#    series.save
    
    assert !VideoGameContinuum.find_by_name('Shadow the Hedgehog').nil?
#    assert VideoGameContinuum.find_by_name('Mahou Shoujo Lyrical Nanoha').video_game_series.include?(series)
    assert_equal 2, VideoGameContinuum.find_by_name('Sonic the Hedgehog').alternate_names.length
    assert !VideoGameContinuum.find_by_name('Sonic the Hedgehog').alternate_names.include?('Shadow')
#    assert !VideoGameContinuum.find_by_name('Magical Girl Lyrical Nanoha').video_game_series.include?(series)
    
    # Let's try merging.
    continuum.merge_into!(VideoGameContinuum.find_by_name('Sonic the Hedgehog'))
    
    assert_nil VideoGameContinuum.find_by_name('Shadow the Hedgehog')
    
    continuum = VideoGameContinuum.find_by_name('Sonic the Hedgehog')
    assert_equal 3, continuum.alternate_names.length
    assert continuum.alternate_names.include?('Shadow')
#    assert continuum.video_game_series.include?(series)
#    assert_equal continuum, VideoGameSeries.find_by_name('Magical Girl Lyrical Nanoha').video_game_continuum
  end

protected
  def create_continuum(options = {})
    VideoGameContinuum.new({ :name => 'Megaman' }.merge(options))
  end
  
  def set_current_user(user_key)
    User.stamper = users(user_key)
  end
end
