require 'test_helper'

class VideoGameSystemTest < ActiveSupport::TestCase
  Error = ''
  fixtures :video_game_systems
  fixtures :comments
  fixtures :users
  
  def test_validations
    assert !create_system(:name => nil).valid?
    assert !create_system(:name => '').valid?
    
    assert_valid create_system()
    
    assert_nothing_raised { create_system(:alternate_names => [1, :symbol, Object.new]) }
    assert_nothing_raised { create_system(:alternate_names => ['a', 'b', 'c']) }
    assert_equal ['a', 'b', 'c'], create_system(:alternate_names => ['a', 'b', 'c']).alternate_names.to_ary
    assert_valid create_system(:alternate_names => ['a', 'b', 'c'])
  end
  
  def test_comments
    # Test adding a comment.
    comment = Comment.new(:text => 'Hey there.  I like this system.')
    child = Comment.new(:text => 'So do I.')
    system = video_game_systems(:genesis)
    old_comments = system.comments.clone
    
    assert system.comments.empty?
    assert_difference 'system.comments.size' do
      system.comments << comment
    end
    assert_valid system
    system.save
    assert system.comments.include?(comment)
    assert_equal comment, system.comments[-1]
    assert_equal old_comments + [comment], system.comments
    assert_nil comment.parent
    assert_equal system, comment.item
    assert_nothing_raised { Comment.find(comment.id) }
    
    # Test adding a child comment.
    # NOTE: Can't save system and have ALL children saved.
    assert_no_difference 'system.comments.size' do
      child = comment.children.create(:text => 'So do I.')
    end
    child.save
    assert_equal old_comments + [comment], system.comments
    # Uh, no way to make sure we can find it...
    #assert_nothing_raised { Comment.find(child.id) }
    #assert_equal child, Comment.find(child.id)
    
    # Test deleting a comment
    assert_difference 'system.comments.size', -1 do
      system.comments.delete(comment)
    end
    assert_equal old_comments, system.comments
    assert_raise ActiveRecord::RecordNotFound do; Comment.find(comment.id); end
    assert_raise ActiveRecord::RecordNotFound do; Comment.find(child.id); end
    
    # Test deleting all comments
    assert_nothing_raised { system.comments.delete_all }
    assert system.comments.empty?
    assert_equal [], system.comments
  end
  
  def test_alternate_names
    system = video_game_systems(:genesis)
    
    assert_kind_of Array, system.alternate_names
    assert !system.alternate_names.empty?
    assert_equal 2, system.alternate_names.size
    assert system.alternate_names.include?("Megadrive")
    assert system.alternate_names.include?("Sega Megadrive")
    
    # Try adding
    assert_difference 'system.alternate_names.count' do
      system.alternate_names << 'Sega Genesis'
      system.save
    end
    assert system.alternate_names.include?('Sega Genesis')
    
    # And deleting
    assert_difference 'system.alternate_names.count', -1 do
      system.alternate_names.delete('Sega Genesis')
    end
    system.save
    assert !system.alternate_names.include?('Sega Genesis')
    
    # And then massive deletion
    assert_nothing_raised { system.alternate_names.delete('Nonsense name!') }
    
    assert_nothing_raised { system.alternate_names.delete_all }
    system.save
    assert system.alternate_names.empty?
  end
  
  def test_state_transitions
    # Here we test a legitimate lifecycle of several systems.
    set_current_user :admin
    
    # Can't modify something that is pending_changes, but should -> approve -> changed
    system = video_game_systems(:nes)
    old_system = system.clone
    
    assert system.pending_changes?
    system.changed!
    assert system.pending_changes?
    system.approve!
    assert system.pending_changes?
    
    assert_equal 'Entertainment System', system.name
    assert_not_nil system.pending_changes
    
    # Should throw an error because system.name has a conflict.
    assert system.errors.empty?
    assert !system.approvable_changes?
    assert !system.errors.empty?
    assert_equal "has more than one pending change which must be resolved prior to approving all changes.", system.errors[:name]
    system.errors.clear
    assert system.errors.empty?
    system.approve_changes!
    assert !system.errors.empty?
    assert_equal "has more than one pending change which must be resolved prior to approving all changes.", system.errors[:name]
    assert_equal 'Entertainment System', system.name
    assert_not_nil system.pending_changes
    
    # Removing the extra name should resolve the conflict.
    system.delete_pending_change :name, :change, "NES"
    assert system.approvable_changes?
    system.approve_changes!
    assert system.approved?
    assert_equal 'Nintendo Entertainment System', system.name
    assert_nil system.pending_changes
    
    # This test is now irrelevant. We 'change' without a transition.
    #system.changed!
    #assert system.pending_changes?
    
    assert_equal old_system.creator, system.creator
    assert_equal old_system.created_at, system.created_at
    assert_equal old_system.approver, system.approver
    assert_equal old_system.approved_at, system.approved_at
    assert_not_equal old_system.updater, system.updater
    assert_not_equal old_system.updated_at, system.updated_at
    
    # Can't approve something that is approved, but should -> changed -> approve
    system = video_game_systems(:playstation)
    old_system = system.clone
    assert system.approved?
    system.approve!
    assert system.approved?
    system.approve_changes!
    assert system.approved?
    # This test now irrelevant.  We 'change' without a transition.
    #system.changed!
    #assert system.pending_changes?
    #system.approve_changes!
    #assert system.approved?
    
    assert_equal old_system.creator, system.creator
    assert_equal old_system.created_at, system.created_at
    assert_equal old_system.approver, system.approver
    assert_equal old_system.approved_at, system.approved_at
    # These two are irrelevant, as we now have never changed state.
    #assert_not_equal old_system.updater, system.updater
    #assert_not_equal old_system.updated_at, system.updated_at
    
    # Can't change something that is pending_approval, but should -> approve -> changed
    system = video_game_systems(:genesis)
    old_system = system.clone
    assert system.pending_approval?
    system.changed!
    assert system.pending_approval?
    system.approve_changes!
    assert system.pending_approval?
    system.approve!
    assert system.approved?
    # This test now irrelevant.  We 'change' without a transition.
    #system.changed!
    #assert system.pending_changes?
    #system.approve_changes!
    #assert system.approved?
    
    assert_equal old_system.creator, system.creator
    assert_equal old_system.created_at, system.created_at
    assert_not_equal old_system.approver, system.approver
    assert_not_equal old_system.approved_at, system.approved_at
    assert_not_equal old_system.updater, system.updater
    assert_not_equal old_system.updated_at, system.updated_at
    
    # And new systems should be pending_approval.
    system = create_system()
    system.save
    old_system = system.clone
    assert_equal users(:admin), system.creator
    
    set_current_user :quentin
    
    assert system.pending_approval?
    system.changed!
    assert system.pending_approval?
    system.approve_changes!
    assert system.pending_approval?
    system.approve!
    assert system.approved?
    # This test now irrelevant.  We 'change' without a transition.
    #system.changed!
    #assert system.pending_changes?
    #system.approve_changes!
    #assert system.approved?
    
    assert_equal old_system.creator, system.creator
    assert_equal old_system.created_at, system.created_at
    assert_not_equal old_system.approver, system.approver
    assert_not_equal old_system.approved_at, system.approved_at
    assert_not_equal old_system.updater, system.updater
    assert_not_equal old_system.updated_at, system.updated_at
    
    assert_equal users(:quentin), system.approver
    
    # TODO: Test that approver_id, approved_at are set
  end
  
  def test_pending_changes_non_moderator
    set_current_user :quentin
    
    system = video_game_systems(:playstation)
    
    assert_equal "Sony Playstation", system.name
    assert_nil system.pending_changes
    assert system.approved?
    
    system.add_pending_change_support :name, :change, 'Sommat else'
    
    assert system.pending_changes?
    assert_equal "Sony Playstation", system.name
    assert_not_nil system.pending_changes
    assert_not_nil system.pending_changes[:name]
    assert_not_nil system.pending_changes[:name]['Sommat else']
    assert system.pending_changes[:name]['Sommat else'].include?(users(:quentin).id)
    
    system.pending_changes = nil
    
    assert_nil system.pending_changes
    assert system.approved?
    
    # Also, test that adding pending changes to something pending approval does
    # nothing to the state.
    
    system = video_game_systems(:genesis)
    
    assert_nil system.pending_changes
    assert system.pending_approval?
    assert_equal "Sega Genesis", system.name
    
    system.add_pending_change_support :name, :change, "Genesis"
    
    assert system.pending_approval?
    assert_equal "Sega Genesis", system.name
    assert_not_nil system.pending_changes
    assert_not_nil system.pending_changes[:name]
    assert_not_nil system.pending_changes[:name]["Genesis"]
    assert system.pending_changes[:name]["Genesis"].include?(users(:quentin).id)
    
    # TODO: Alternate names, genres
    # TODO: Test remove pending changes.
  end
  
  def test_pending_changes_moderator
    set_current_user :quentin
    
    user = users(:quentin)
    user.roles << roles(:video_game_system_moderator)
    user.save
    
    system = video_game_systems(:playstation)
    
    assert_equal "Sony Playstation", system.name
    assert_nil system.pending_changes
    assert system.approved?
    
    system.name = 'Sommat else'
    
    assert system.approved?
    assert_equal 'Sommat else', system.name
    assert_nil system.pending_changes
    
    # Also, test that adding pending changes to something pending approval does
    # nothing to the state.
    
    system = video_game_systems(:genesis)
    
    assert_nil system.pending_changes
    assert system.pending_approval?
    assert_equal "Sega Genesis", system.name
    
    system.name = "Genesis"
    
    assert system.pending_approval?
    assert_equal "Genesis", system.name
    assert_nil system.pending_changes
    
    # TODO: Alternate names, genres
    # TODO: Test remove pending changes.
  end
  
  def test_merge_into
    # Merging into should delete the anime system, merge its names into the
    # other, and reassign all other anime series tied to it.
    system = create_system(:name => 'Nintendo Famicom')
    system.alternate_names << 'Famicon'
    system.save
    
#    series = AnimeSeries.new({:name => 'Magical Girl Lyrical Nanoha'})
#    series.video_game_system = system
#    series.save
    
    assert !VideoGameSystem.find_by_name('Nintendo Famicom').nil?
#    assert VideoGameSystem.find_by_name('Mahou Shoujo Lyrical Nanoha').anime_series.include?(series)
    assert_equal 1, VideoGameSystem.find_by_name('Nintendo Famicom').alternate_names.length
    assert !VideoGameSystem.find_by_name('Entertainment System').alternate_names.include?('Nintendo Famicom')
    assert !VideoGameSystem.find_by_name('Entertainment System').alternate_names.include?('Famicon')
#    assert !VideoGameSystem.find_by_name('Magical Girl Lyrical Nanoha').anime_series.include?(series)
    
    # Let's try merging.
    system.merge_into!(VideoGameSystem.find_by_name('Entertainment System'))
    
    assert_nil VideoGameSystem.find_by_name('Nintendo Famicom')
    
    system = VideoGameSystem.find_by_name('Entertainment System')
    assert_equal 2, system.alternate_names.length
    assert system.alternate_names.include?('Nintendo Famicom')
    assert system.alternate_names.include?('Famicon')
#    assert system.anime_series.include?(series)
#    assert_equal system, AnimeSeries.find_by_name('Magical Girl Lyrical Nanoha').video_game_system
  end

protected
  def create_system(options = {})
    VideoGameSystem.new({ :name => 'Dreamcast', :manufacturer => 'Sega' }.merge(options))
  end
  
  def set_current_user(user_key)
    User.stamper = users(user_key)
  end
end
