require File.dirname(__FILE__) + '/../test_helper'

class VideoGameTest < ActiveSupport::TestCase
  Error = ''
  fixtures :video_games
  fixtures :comments
  fixtures :users
  
  def test_validations
    assert !create_game(:name => nil).valid?
    assert !create_game(:name => '').valid?
    # TODO: Associate and test genres.
#    assert !create_game(:genres => []).valid?
    assert !create_game(:video_game_system => nil).valid?
    
    assert_valid create_game()
#    assert_nil create_game().anidb_id
#    assert_valid create_game(:anidb_id => 3651)
#    assert_equal 3651, create_game(:anidb_id => 3651).anidb_id
    
    assert_nothing_raised { create_game(:alternate_names => [1, :symbol, Object.new]) }
    assert_nothing_raised { create_game(:alternate_names => ['a', 'b', 'c']) }
#    assert_equal 630, create_game(:anidb_id => 630).anidb_id
    assert_equal ['a', 'b', 'c'], create_game(:alternate_names => ['a', 'b', 'c']).alternate_names.to_ary
    assert_valid create_game(:alternate_names => ['a', 'b', 'c'])
  end
  
  def test_comments
    # Test adding a comment.
    comment = Comment.new(:text => 'Hey there.  I like this game.')
    child = Comment.new(:text => 'So do I.')
    game = video_games(:super_mario_brothers)
    old_comments = game.comments.clone
    
    assert !game.comments.empty?
    assert_difference 'game.comments.size' do
      game.comments << comment
    end
    assert_valid game
    game.save
    assert game.comments.include?(comment)
    assert_equal comment, game.comments[-1]
    assert_equal old_comments + [comment], game.comments
    assert_nil comment.parent
    assert_equal game, comment.item
    assert_nothing_raised { Comment.find(comment.id) }
    
    # Test adding a child comment.
    # NOTE: Can't save game and have ALL children saved.
    assert_no_difference 'game.comments.size' do
      child = comment.children.create(:text => 'So do I.')
    end
    child.save
    assert_equal old_comments + [comment], game.comments
    # Uh, no way to make sure we can find it...
    #assert_nothing_raised { Comment.find(child.id) }
    #assert_equal child, Comment.find(child.id)
    
    # Test deleting a comment
    assert_difference 'game.comments.size', -1 do
      game.comments.delete(comment)
    end
    assert_equal old_comments, game.comments
    assert_raise ActiveRecord::RecordNotFound do; Comment.find(comment.id); end
    assert_raise ActiveRecord::RecordNotFound do; Comment.find(child.id); end
    
    # Test deleting all comments
    assert_nothing_raised { game.comments.delete_all }
    assert game.comments.empty?
    assert_equal [], game.comments
  end
  
  # TODO: Test genres.
  
  # TODO: Test AnimeContinuum
  
  def test_alternate_names
    game = video_games(:super_mario_brothers)
    
    assert_kind_of Array, game.alternate_names
    assert !game.alternate_names.empty?
    assert_equal 2, game.alternate_names.size
    assert game.alternate_names.include?("Super Mario Bros.")
    assert game.alternate_names.include?("SMB 1")
    
    # Try adding
    assert_difference 'game.alternate_names.count' do
      game.alternate_names << 'SMB1'
      game.save
    end
    assert game.alternate_names.include?('SMB1')
    
    # And deleting
    assert_difference 'game.alternate_names.count', -1 do
      game.alternate_names.delete('SMB1')
    end
    game.save
    assert !game.alternate_names.include?('SMB1')
    
    # And then massive deletion
    assert_nothing_raised { game.alternate_names.delete('Nonsense name!') }
    
    assert_nothing_raised { game.alternate_names.delete_all }
    game.save
    assert game.alternate_names.empty?
  end
  
  def test_state_transitions
    # Here we test a legitimate lifecycle of several game.
    set_current_user :admin
    
    # Can't modify something that is pending_changes, but should -> approve -> changed
    game = video_games(:sonic_the_hedgehog)
    old_game = game.clone
    
    assert game.pending_changes?
    game.changed!
    assert game.pending_changes?
    game.approve!
    assert game.pending_changes?
    
    # TODO: assert change in genres.
    assert_equal 1991, game.year
    assert_equal 'Sornic the hedgehog?', game.name
    assert_not_nil game.pending_changes
    
    # Should throw an error because game.years has a conflict.
    assert game.errors.empty?
    assert !game.approvable_changes?
    assert !game.errors.empty?
    assert_equal "has more than one pending change which must be resolved prior to approving all changes.", game.errors[:year]
    game.errors.clear
    assert game.errors.empty?
    game.approve_changes!
    assert !game.errors.empty?
    assert_equal "has more than one pending change which must be resolved prior to approving all changes.", game.errors[:year]
    assert_equal 1991, game.year
    assert_equal 'Sornic the hedgehog?', game.name
    assert_not_nil game.pending_changes
    
    # Removing the extra year should resolve the conflict.
    game.delete_pending_change(:year, :change, "2006")
    assert game.approvable_changes?
    game.approve_changes!
    assert game.approved?
    assert_equal 2005, game.year
    assert_equal 'Sonic the Hedgehog', game.name
    assert_nil game.pending_changes
    # TODO: assert change in genres.
    
    # This test is now irrelevant. We 'change' without a transition.
    #game.changed!
    #assert game.pending_changes?
    
    assert_equal old_game.creator, game.creator
    assert_equal old_game.created_at, game.created_at
    assert_equal old_game.approver, game.approver
    assert_equal old_game.approved_at, game.approved_at
    assert_not_equal old_game.updater, game.updater
    assert_not_equal old_game.updated_at, game.updated_at
    
    # Can't approve something that is approved, but should -> changed -> approve
    game = video_games(:super_mario_brothers)
    old_game = game.clone
    assert game.approved?
    game.approve!
    assert game.approved?
    game.approve_changes!
    assert game.approved?
    # This test now irrelevant.  We 'change' without a transition.
    #game.changed!
    #assert game.pending_changes?
    #game.approve_changes!
    #assert game.approved?
    
    assert_equal old_game.creator, game.creator
    assert_equal old_game.created_at, game.created_at
    assert_equal old_game.approver, game.approver
    assert_equal old_game.approved_at, game.approved_at
    # These two are irrelevant, as we now have never changed state.
    #assert_not_equal old_game.updater, game.updater
    #assert_not_equal old_game.updated_at, game.updated_at
    
    # Can't change something that is pending_approval, but should -> approve -> changed
    game = video_games(:okami)
    old_game = game.clone
    assert game.pending_approval?
    game.changed!
    assert game.pending_approval?
    game.approve_changes!
    assert game.pending_approval?
    game.approve!
    assert game.approved?
    # This test now irrelevant.  We 'change' without a transition.
    #game.changed!
    #assert game.pending_changes?
    #game.approve_changes!
    #assert game.approved?
    
    assert_equal old_game.creator, game.creator
    assert_equal old_game.created_at, game.created_at
    assert_not_equal old_game.approver, game.approver
    assert_not_equal old_game.approved_at, game.approved_at
    assert_not_equal old_game.updater, game.updater
    assert_not_equal old_game.updated_at, game.updated_at
    
    # And new game should be pending_approval.
    game = create_game()
    game.save
    old_game = game.clone
    assert_equal users(:admin), game.creator
    
    set_current_user :quentin
    
    assert game.pending_approval?
    game.changed!
    assert game.pending_approval?
    game.approve_changes!
    assert game.pending_approval?
    game.approve!
    assert game.approved?
    # This test now irrelevant.  We 'change' without a transition.
    #game.changed!
    #assert game.pending_changes?
    #game.approve_changes!
    #assert game.approved?
    
    assert_equal old_game.creator, game.creator
    assert_equal old_game.created_at, game.created_at
    assert_not_equal old_game.approver, game.approver
    assert_not_equal old_game.approved_at, game.approved_at
    assert_not_equal old_game.updater, game.updater
    assert_not_equal old_game.updated_at, game.updated_at
    
    assert_equal users(:quentin), game.approver
    
    # TODO: Test that approver_id, approved_at are set
  end
  
  def test_pending_changes_non_moderator
    set_current_user :quentin
    
    game = video_games(:super_mario_brothers)
    
    assert_equal "Super Mario Brothers", game.name
    assert_nil game.pending_changes
    assert game.approved?
    
    game.add_pending_change_support :name, :change, 'Sommat else'
    
    assert game.pending_changes?
    assert_equal "Super Mario Brothers", game.name
    assert_not_nil game.pending_changes
    assert_not_nil game.pending_changes[:name]
    assert_not_nil game.pending_changes[:name]['Sommat else']
    assert game.pending_changes[:name]['Sommat else'].include?(users(:quentin).id)
    
    game.pending_changes = nil
    
    assert_nil game.pending_changes
    assert game.approved?
    
    assert_equal 1985, game.year
    
    game.add_pending_change_support :year, :change, "2004"
    
    assert_equal 1985, game.year
    assert_not_nil game.pending_changes
    assert_not_nil game.pending_changes[:year]
    assert_not_nil game.pending_changes[:year]["2004"]
    assert game.pending_changes[:year]["2004"].include?(users(:quentin).id)
    
    # Also, test that adding pending changes to something pending approval does
    # nothing to the state.
    
    game = video_games(:okami)
    
    assert_nil game.pending_changes
    assert game.pending_approval?
    assert_equal "Okami", game.name
    
    game.add_pending_change_support :name, :change, "Okami!"
    
    assert game.pending_approval?
    assert_equal "Okami", game.name
    assert_not_nil game.pending_changes
    assert_not_nil game.pending_changes[:name]
    assert_not_nil game.pending_changes[:name]["Okami!"]
    assert game.pending_changes[:name]["Okami!"].include?(users(:quentin).id)
    
    # TODO: Alternate names, genres
    # TODO: Test remove pending changes.
  end
  
  def test_pending_changes_moderator
    set_current_user :quentin
    
    user = users(:quentin)
    user.roles << roles(:video_game_moderator)
    user.save
    
    game = video_games(:super_mario_brothers)
    
    assert_equal "Super Mario Brothers", game.name
    assert_nil game.pending_changes
    assert game.approved?
    
    game.name = 'Sommat else'
    
    assert game.approved?
    assert_equal 'Sommat else', game.name
    assert_nil game.pending_changes
    
    assert_equal 1985, game.year
    
    game.year = "1986"
    
    assert_equal 1986, game.year
    assert_nil game.pending_changes
    
    # Also, test that adding pending changes to something pending approval does
    # nothing to the state.
    
    game = video_games(:okami)
    
    assert_nil game.pending_changes
    assert game.pending_approval?
    assert_equal "Okami", game.name
    
    game.name = "Okami!"
    
    assert game.pending_approval?
    assert_equal "Okami!", game.name
    assert_nil game.pending_changes
    
    # TODO: Alternate names, genres
    # TODO: Test remove pending changes.
  end

  def test_merge_into
    # Merging into should delete the anime game, merge its names into the
    # other, and reassign all other anime game tied to it.
    game = create_game(:name => 'Okami!')
    game.alternate_names << 'Oukami'
    game.save
    
    game.video_game_genres << VideoGameGenre.find(2)
    game.save
    
    assert !VideoGame.find_by_name('Okami!').nil?
    assert VideoGame.find_by_name('Okami!').video_game_genres.include?(VideoGameGenre.find(2))
    assert_equal 0, VideoGame.find_by_name('Okami').alternate_names.length
    assert_equal 1, VideoGame.find_by_name('Okami').video_game_genres.length
    
    # Let's try merging.
    game.merge_into!(VideoGame.find_by_name('Okami'))
    
    assert_nil VideoGame.find_by_name('Okami!')
    
    game = VideoGame.find_by_name('Okami')
    assert_equal 2, game.alternate_names.length
    assert game.alternate_names.include?('Okami!')
    assert game.alternate_names.include?('Oukami')
    assert_equal 2, game.video_game_genres.length
    assert game.video_game_genres.include?(VideoGameGenre.find(2))
  end

protected
  def create_game(options = {})
    VideoGame.new({ :name => 'Dragon Quest', :year => '2006', :video_game_system => VideoGameSystem.find(1) }.merge(options))
  end
  
  def set_current_user(user_key)
    User.stamper = users(user_key)
  end
end
