require File.dirname(__FILE__) + '/../test_helper'

class VideoGameTriviaQuestionTest < ActiveSupport::TestCase
  Error = ''
  fixtures :video_game_trivia_questions
  fixtures :video_games
  fixtures :users
  
  def test_validations
    assert !create_trivia_question(:question => nil).valid?
    assert !create_trivia_question(:question => '').valid?
    assert !create_trivia_question(:answer => nil).valid?
    assert !create_trivia_question(:answer => '').valid?
    assert !create_trivia_question(:points => nil).valid?
    assert !create_trivia_question(:points => '').valid?
    
    assert_valid create_trivia_question()
    assert !create_trivia_question(:points => 0).valid?
    assert_valid create_trivia_question(:points => 1)
    assert_valid create_trivia_question(:points => 2)
    assert_valid create_trivia_question(:points => 3)
    assert !create_trivia_question(:points => 4).valid?
  end
  
  def test_comments
    # Test adding a comment.
    comment = Comment.new(:text => 'Hey there.  I like this series.')
    child = Comment.new(:text => 'So do I.')
    trivia_question = video_game_trivia_questions(:zones)
    old_comments = trivia_question.comments.clone
    
    assert trivia_question.comments.empty?
    assert_difference 'trivia_question.comments.size' do
      trivia_question.comments << comment
    end
    assert_valid trivia_question
    trivia_question.save
    assert trivia_question.comments.include?(comment)
    assert_equal comment, trivia_question.comments[-1]
    assert_equal old_comments + [comment], trivia_question.comments
    assert_nil comment.parent
    assert_equal trivia_question, comment.item
    assert_nothing_raised { Comment.find(comment.id) }
    
    # Test adding a child comment.
    # NOTE: Can't save series and have ALL children saved.
    assert_no_difference 'trivia_question.comments.size' do
      child = comment.children.create(:text => 'So do I.')
    end
    child.save
    assert_equal old_comments + [comment], trivia_question.comments
    # Uh, no way to make sure we can find it...
    #assert_nothing_raised { Comment.find(child.id) }
    #assert_equal child, Comment.find(child.id)
    
    # Test deleting a comment
    assert_difference 'trivia_question.comments.size', -1 do
      trivia_question.comments.delete(comment)
    end
    assert_equal old_comments, trivia_question.comments
    assert_raise ActiveRecord::RecordNotFound do; Comment.find(comment.id); end
    assert_raise ActiveRecord::RecordNotFound do; Comment.find(child.id); end
    
    # Test deleting all comments
    assert_nothing_raised { trivia_question.comments.delete_all }
    assert trivia_question.comments.empty?
    assert_equal [], trivia_question.comments
  end
  
  # TODO: Test VideoGame
  
  def test_state_transitions
    # Here we test a legitimate lifecycle of several trivia questions.
    set_current_user :admin
    
    # Can't modify something that is pending_changes, but should -> approve -> changed
    trivia_question = video_game_trivia_questions(:kaguya)
    old_trivia_question = trivia_question.clone
    
    assert trivia_question.pending_changes?
    trivia_question.changed!
    assert trivia_question.pending_changes?
    trivia_question.approve!
    assert trivia_question.pending_changes?
    
    # TODO: assert change in genres.
    assert_equal 'Name the character wearing a bamboo suit.', trivia_question.question
    assert_equal 'Kaguya', trivia_question.answer
    assert_not_nil trivia_question.pending_changes
    
    # Should throw an error because trivia_question.answer has a conflict.
    assert trivia_question.errors.empty?
    assert !trivia_question.approvable_changes?
    assert !trivia_question.errors.empty?
    assert_equal "has more than one pending change which must be resolved prior to approving all changes.", trivia_question.errors[:answer]
    trivia_question.errors.clear
    assert trivia_question.errors.empty?
    trivia_question.approve_changes!
    assert !trivia_question.errors.empty?
    assert_equal "has more than one pending change which must be resolved prior to approving all changes.", trivia_question.errors[:answer]
    assert_equal 'Name the character wearing a bamboo suit.', trivia_question.question
    assert_equal 'Kaguya', trivia_question.answer
    assert_not_nil trivia_question.pending_changes
    assert_not_nil trivia_question.pending_changes
    
    # Removing the extra year should resolve the conflict.
    trivia_question.delete_pending_change(:answer, :change, "Moon Princess Kaguya")
    assert trivia_question.approvable_changes?
    trivia_question.approve_changes!
    assert trivia_question.approved?
    assert_equal 'Name the character wearing a bamboo suit.', trivia_question.question
    assert_equal 'Princess Kaguya', trivia_question.answer
    assert_nil trivia_question.pending_changes
    
    # This test is now irrelevant. We 'change' without a transition.
    #series.changed!
    #assert series.pending_changes?
    
    assert_equal old_trivia_question.creator, trivia_question.creator
    assert_equal old_trivia_question.created_at, trivia_question.created_at
    assert_equal old_trivia_question.approver, trivia_question.approver
    assert_equal old_trivia_question.approved_at, trivia_question.approved_at
    assert_not_equal old_trivia_question.updater, trivia_question.updater
    assert_not_equal old_trivia_question.updated_at, trivia_question.updated_at
    
    # Can't approve something that is approved, but should -> changed -> approve
    trivia_question = video_game_trivia_questions(:zones)
    old_trivia_question = trivia_question.clone
    assert trivia_question.approved?
    trivia_question.approve!
    assert trivia_question.approved?
    trivia_question.approve_changes!
    assert trivia_question.approved?
    # This test now irrelevant.  We 'change' without a transition.
    #trivia_question.changed!
    #assert trivia_question.pending_changes?
    #trivia_question.approve_changes!
    #assert trivia_question.approved?
    
    assert_equal old_trivia_question.creator, trivia_question.creator
    assert_equal old_trivia_question.created_at, trivia_question.created_at
    assert_equal old_trivia_question.approver, trivia_question.approver
    assert_equal old_trivia_question.approved_at, trivia_question.approved_at
    # These two are irrelevant, as we now have never changed state.
    #assert_not_equal old_trivia_question.updater, trivia_question.updater
    #assert_not_equal old_trivia_question.updated_at, trivia_question.updated_at
    
    # Can't change something that is pending_approval, but should -> approve -> changed
    trivia_question = video_game_trivia_questions(:worlds)
    old_trivia_question = trivia_question.clone
    assert trivia_question.pending_approval?
    trivia_question.changed!
    assert trivia_question.pending_approval?
    trivia_question.approve_changes!
    assert trivia_question.pending_approval?
    trivia_question.approve!
    assert trivia_question.approved?
    # This test now irrelevant.  We 'change' without a transition.
    #trivia_question.changed!
    #assert trivia_question.pending_changes?
    #trivia_question.approve_changes!
    #assert trivia_question.approved?
    
    assert_equal old_trivia_question.creator, trivia_question.creator
    assert_equal old_trivia_question.created_at, trivia_question.created_at
    assert_not_equal old_trivia_question.approver, trivia_question.approver
    assert_not_equal old_trivia_question.approved_at, trivia_question.approved_at
    assert_not_equal old_trivia_question.updater, trivia_question.updater
    assert_not_equal old_trivia_question.updated_at, trivia_question.updated_at
    
    # And new trivia questions should be pending_approval.
    trivia_question = create_trivia_question()
    trivia_question.save
    old_trivia_question = trivia_question.clone
    assert_equal users(:admin), trivia_question.creator
    
    set_current_user :quentin
    
    assert trivia_question.pending_approval?
    trivia_question.changed!
    assert trivia_question.pending_approval?
    trivia_question.approve_changes!
    assert trivia_question.pending_approval?
    trivia_question.approve!
    assert trivia_question.approved?
    # This test now irrelevant.  We 'change' without a transition.
    #trivia_question.changed!
    #assert trivia_question.pending_changes?
    #trivia_question.approve_changes!
    #assert trivia_question.approved?
    
    assert_equal old_trivia_question.creator, trivia_question.creator
    assert_equal old_trivia_question.created_at, trivia_question.created_at
    assert_not_equal old_trivia_question.approver, trivia_question.approver
    assert_not_equal old_trivia_question.approved_at, trivia_question.approved_at
    assert_not_equal old_trivia_question.updater, trivia_question.updater
    assert_not_equal old_trivia_question.updated_at, trivia_question.updated_at
    
    assert_equal users(:quentin), trivia_question.approver
    
    # TODO: Test that approver_id, approved_at are set
  end
  
  def test_pending_changes_non_moderator
    set_current_user :quentin
    
    trivia_question = video_game_trivia_questions(:zones)
    
    assert_equal "A, B, C, D, E, F, G", trivia_question.answer
    assert_nil trivia_question.pending_changes
    assert trivia_question.approved?
    
    trivia_question.add_pending_change_support :answer, :change, 'Sommat else'
    
    assert trivia_question.pending_changes?
    assert_equal "A, B, C, D, E, F, G", trivia_question.answer
    assert_not_nil trivia_question.pending_changes
    assert_not_nil trivia_question.pending_changes[:answer]
    assert_not_nil trivia_question.pending_changes[:answer]['Sommat else']
    assert trivia_question.pending_changes[:answer]['Sommat else'].include?(users(:quentin).id)
    
    trivia_question.pending_changes = nil
    
    assert_nil trivia_question.pending_changes
    assert trivia_question.approved?
    
    # Also, test that adding pending changes to something pending approval does
    # nothing to the state.
    
    trivia_question = video_game_trivia_questions(:worlds)
    
    assert_nil trivia_question.pending_changes
    assert trivia_question.pending_approval?
    assert_equal "eight", trivia_question.answer
    
    trivia_question.add_pending_change_support :answer, :change, "8"
    
    assert trivia_question.pending_approval?
    assert_equal "eight", trivia_question.answer
    assert_not_nil trivia_question.pending_changes
    assert_not_nil trivia_question.pending_changes[:answer]
    assert_not_nil trivia_question.pending_changes[:answer]["8"]
    assert trivia_question.pending_changes[:answer]["8"].include?(users(:quentin).id)
    
    # TODO: Test remove pending changes.
  end
  
  def test_pending_changes_moderator
    set_current_user :quentin
    
    user = users(:quentin)
    user.roles << roles(:video_game_trivia_moderator)
    user.save
    
    trivia_question = video_game_trivia_questions(:zones)
    
    assert_equal "A, B, C, D, E, F, G", trivia_question.answer
    assert_nil trivia_question.pending_changes
    assert trivia_question.approved?
    
    trivia_question.answer = 'Sommat else'
    
    assert trivia_question.approved?
    assert_equal 'Sommat else', trivia_question.answer
    assert_nil trivia_question.pending_changes
    
    # Also, test that adding pending changes to something pending approval does
    # nothing to the state.
    
    trivia_question = video_game_trivia_questions(:worlds)
    
    assert_nil trivia_question.pending_changes
    assert trivia_question.pending_approval?
    assert_equal "eight", trivia_question.answer
    
    trivia_question.answer = "8"
    
    assert trivia_question.pending_approval?
    assert_equal "8", trivia_question.answer
    assert_nil trivia_question.pending_changes
    
    # TODO: Alternate names, genres
    # TODO: Test remove pending changes.
  end

protected
  def create_trivia_question(options = {})
    VideoGameTriviaQuestion.new({ :video_game => VideoGame.find_by_name('Okami'), :question => 'Who is the main character?', :answer => 'Amaterasu', :points => 1 }.merge(options))
  end
  
  def set_current_user(user_key)
    User.stamper = users(user_key)
  end
end
