# HasManyValues
module ActiveRecord
  module Associations
    module ClassMethods
      # Similar to +has_many+, except that it allows for values that are not
      # ActiveRecord objects, wrapping them and putting them into a table
      # anyway. Essentially, this allows for collection-like attributes that
      # do NOT require inserting objects that are subclassed from ActiveRecord,
      # as it will do the subclassing for you.
      #
      # It requires you to make a dummy ActiveRecord class with the same name as
      # the argument, requiring a foreign key on that table to reference the
      # current class and a second column named 'value'.
      #
      # Example:
      #
      # The following classes might be defined to hold a collection of strings
      # in gizmo.gizmo_names: 
      #
      # <tt>class Gizmo < ActiveRecord::Base
      #   has_many_values :gizmo_names
      # end
      #
      # class GizmoName < ActiveRecord::Base
      # end</tt>
      #
      # The following DB migration would be needed to define GizmoName:
      #
      # <tt>class CreateGizmoNames < ActiveRecord::Migration
      #   def self.up
      #     create_table :gizmo_names do |t|
      #       t.column :gizmo_id, :integer
      #       t.column :value, :string
      #     end
      #   end
      # 
      #   def self.down
      #     drop_table :gizmo_names
      #   end
      # end</tt>
      #
      # And gizmo_names can now be used as an array of strings, like so:
      # 
      # <tt>doodad = Gizmo.new
      # doodad.gizmo_names << 'Frazbot Doodad (FB-D103)'</tt>
      def has_many_values(association_id, option = {}, &extension)
        has_many association_id, option, &extension
        
        # Now we override what collection_accessor_methods made.
        
        # Get the reflection it built the methods with...
        reflection = reflect_on_association(association_id)
        
        # Rename the old reflection.name getter method and make it private...
        define_method("_#{reflection.name}", instance_method(reflection.name))
        private "_#{reflection.name}"
        
        # And now redefine the reflection.name getter method to return a
        # 'HasManyValuesAssociation' wrapping the original HasManyAssociation.
        define_method(reflection.name) do |*params|
          association = method("_#{reflection.name}").call(*params)
          
          # Technically it isn't a real class...  The behavior of
          # a HasManyValuesAssociation is really just implemented as a
          # metaclass on top of the HasManyAssociation.
          class << association
            # First we need to undefine find (It's kinda useless for an array of
            # values.)
            # Can't undefine find because of find_target.
            #alias_method :_find, :find
            #undef_method :find
            
            # We need to redefine to_ary.
            if !instance_methods.include? '_to_ary'
              alias_method :_to_ary, :to_ary
              define_method :to_ary do
                load_target
                ary = @target
                ary = Array(@target) if not ary.is_a?(Array)
                ary.collect { |item| item.value }
              end
            end
            
            # Don't need to redefine reset.
            
            # Need to undefine build.
            if !instance_methods.include? '__build'
              alias_method :__build, :build
              undef_method :build
              define_method :_build do |options|
                __build({:item => proxy_owner}.merge(options))
              end
              # Required for Rails 2.3??
              define_method :build do
                # Do nothing
              end
            end
            
            # Need to redefine <<, push, and concat.
            if !instance_methods.include? '_push'
              alias_method :_push, :<<
              define_method :<< do |*values|
                if @reflection.options[:uniq]
                  values = values - to_ary
                end
                # Note: We ignore :as => :item.  It's assumed.
                values.collect! { |value| _build(:value => value) }
              end
              alias_method :push, :<<
              alias_method :concat, :<<
            end
            
            # Don't need to redefine delete_all.
            
            # What to do with sum?
            
            # Need to redefine delete.
            if !instance_methods.include? '_delete'
              alias_method :_delete, :delete
              define_method :delete do |*values|
                values.collect! { |value| find(:all, :conditions => [ "value = ?", value ]) }
                values.flatten!
                values.reject! { |value| value == nil }
                _delete(*values)
              end
            end
            
            # And delete_all, because we redefined delete.
            if !instance_methods.include? '_delete_all'
              alias_method :_delete_all, :delete_all
              define_method :delete_all do
                load_target
                _delete(@target)
                reset_target!
              end
            end
            
            # But not clear or destroy_all.
            
            # Undefine create and create!
            if instance_methods.include? 'create'
              undef_method :create
            end
            
            if instance_methods.include? 'create!'
              undef_method :create!
            end
            
            # But leave size and empty?
            
            # How to redefine any?
            
            # Leave uniq, but redefine replace
            if !instance_methods.include? '_replace'
              alias_method :_replace, :replace
              define_method :replace do |other_array|
                other_array.collect! do |value|
                  subarray = find(:all, :conditions => [ "value = ?", value ])
                  subarray << _build(:value => value) if subarray.empty?
                  subarray
                end
                # These seem unnecessary.  Huh.
#                other_array.flatten!
#                _replace(other_array)
              end
            end
            
            # And lastly, redefine include?
            if !instance_methods.include? '_include?'
              alias_method :_include?, :include?
              define_method :include? do |value|
                return find(:first, :conditions => [ "value = ?", value ]) != nil
              end
            end
          end
          
          association
        end
        
        # Rename the old "#{reflection.name}=" method and make it private...
        define_method("_#{reflection.name}=", instance_method("#{reflection.name}="))
        private "_#{reflection.name}="
        
        # And now redefine the "#{reflection.name}=" method to use the
        # 'HasManyValuesAssociation' returned by the reflection.name method.
        define_method("#{reflection.name}=") do |new_value|
          association = method("#{reflection.name}").call
          association.replace(new_value)
          association
        end
        
        # Lastly, just undefine the old
        # "#{reflection.name.to_s.singularize}ids=" method.
        undef_method("#{reflection.name.to_s.singularize}_ids=")
      end
    end
  end
end
